#include <iostream>

#include "AkentiInit.hpp"
#include "AkentiEngine.hpp"
#include "AkentiPrincipal.hpp"
#include "AkentiMessage.hpp"
#include "DistinguishedName.hpp"
#include "CapabilityCertificate.hpp"
#include "CertificateChain.hpp"
#include "CertificateFactory.hpp"
#include "AkEngInf.h"
#include "Exception.hpp"
#include "Error.hpp"

using namespace Akenti;

void 
doStatus(const string& in, char** out) {
   if (out) {
      if (*out) {
          free(*out);
      }

      *out = strdup(in.c_str());
   }
}

int 
akentiInit(const char *confFile, char **status) {
  string outStatus;
  int retval = -1;

  try {
    retval = AkentiInit::init(confFile, outStatus) ? 0 : 1;
  } catch(Exception& ex) {
    outStatus = ex.getMessage();
  } catch(Error& error) {
    outStatus = error.getMessage();
  }

  doStatus(outStatus, status);
  return retval;
}

int 
checkAccess(const char *subjectDN, 
            const char *subjectCA, 
            const char *resource,
            CapCert* outCert,
            AKMessage* outMsg) {
   delete(*outCert);
   delete(*outMsg);
   *outCert = NULL;
   *outMsg = NULL;

   int retval = -1;

   try {
     CapabilityCertificate capCert;
     AkentiMessage msg;
     AkentiEngine eng;
     AkentiPrincipal principal(subjectDN, subjectCA);

     if (eng.checkAccess(resource, principal, capCert, msg)) {
       *outCert = new CapabilityCertificate(capCert); 
       retval = 0;
     } else {
       retval = 1;
     }

     *outMsg = new AkentiMessage(msg.getCode(), msg.getMessage());
   } catch (Exception& ex) {
     *outMsg = new AkentiMessage(901, ex.getMessage());
   } catch (Error& err) {
     *outMsg = new AkentiMessage(901, err.getMessage());
   }

   return retval;
}

int
checkAccessByCertFile(const char *subjectCertFile,
                       const char *resource,
                       CapCert* outCert,
                       AKMessage* outMsg) {
   delete(*outCert);
   delete(*outMsg);
   *outCert = NULL;
   *outMsg = NULL;

   int retval = -1;

   try {
     AkentiMessage msg;
     AkentiEngine eng;
     CapabilityCertificate capCert;
     CertificateFactory fac = CertificateFactory::getInstance();
     CertificateChain chain = fac->initCertificateChainFromFile(
        "PEM",
        subjectCertFile
     );

     if (eng.checkAccess(resource, chain, capCert, msg)) {
       *outCert = new CapabilityCertificate(capCert); 
       retval = 0;
     } 

     *outMsg = new AkentiMessage(msg.getCode(), msg.getMessage());
   } catch (Exception& ex) {
     *outMsg = new AkentiMessage(901, ex.getMessage());
   } catch (Error& err) {
     *outMsg = new AkentiMessage(901, err.getMessage());
   }

   return retval;
}

int
checkAccessByCertString (const char *subjectCert,
                       const char *resource,
                       CapCert* outCert,
                       AKMessage* outMsg) {
   delete(*outCert);
   delete(*outMsg);
   *outCert = NULL;
   *outMsg = NULL;

   int retval = -1;

   try {
     AkentiMessage msg;
     AkentiEngine eng;
     CapabilityCertificate capCert;
     CertificateFactory fac = CertificateFactory::getInstance();
     CertificateChain chain = fac->initCertificateChain(
        "PEM",
        subjectCert
     );

     if (eng.checkAccess(resource, chain, capCert, msg)) {
       *outCert = new CapabilityCertificate(capCert); 
       retval = 0;
     } 

     *outMsg = new AkentiMessage(msg.getCode(), msg.getMessage());
   } catch (Exception& ex) {
     *outMsg = new AkentiMessage(901, ex.getMessage());
   } catch (Error& err) {
     *outMsg = new AkentiMessage(901, err.getMessage());
   }

   return retval;
}
