#include <sstream>
#include <algorithm>
#include "AkentiCertificateImpl.hpp"
#include "AkentiInputStream.hpp"
#include "AkentiOutputStream.hpp"
#include "PublicKey.hpp"
#include "PrivateKey.hpp"
#include "Provider.hpp"
#include "MessageDigest.hpp"
#include "Signature.hpp"
#include "Base64.hpp"
#include "Exception.hpp"
#include "Error.hpp"
#include "MessageDigest.hpp"
#include "StringUtil.hpp"
#include "Util.hpp"
#include "StringTokenizer.hpp"
#include "Xerces-Includes.hpp"
#include "XMLUtil.hpp"
#include "Base64Util.hpp"
#include <iostream>

using std::ostringstream;
using std::endl;

namespace Akenti {

typedef AkentiCertificateImpl ACI;

ACI::AkentiCertificateImpl(const string& type) 
  : type(type),
    version(2), 
    id("NO_ID"), 
    issuer(),
    notBefore(),
    notAfter(),
    canonAlg("Ak1CanAlg"),
    signatureAlg("RSA-MD5"),
    signature("NO_SIG") {
  if (!type.length()) {
    throw Error(
      "AkentiCertificateImpl:constructor:type = estring"
    );
  }
}

ACI::~AkentiCertificateImpl() {
}

string
ACI::hashCode() const {
  // XXX
  ACI * impl = (ACI*) this;

  if (impl->hashcode.size() == 0) {
    impl->hashcode = MessageDigest::digest("MD5", getHashString());
  }

  return hashcode;
}

const string&
ACI::getType() const {
  return type;
}

bool 
ACI::hasType(const string& type) const {
  return this->type == type;
}

void 
ACI::setVersionNumber(int version) {
  switch(version) {
    case 1:
    case 2:
      this->version = version;
      break;
    default:
      throw Exception(
        "AkentiCertificateImpl:setVersion:bad version number"
      );
  }
}

int
ACI::getVersionNumber() const {
  return version;
}

void 
ACI::setUniqueID(const string& id) {
  if (!id.length()) {
    throw Error(
      "AkentiCertificateImpl:setID:illegal argument(estring)"
    );
  }

  this->id = id;
}

const string&
ACI::getUniqueID() const {
  return id;
}

void
ACI::setIssuer(const AkentiIssuer& issuer) {
  this->issuer = issuer;
}

const AkentiIssuer&
ACI::getIssuer() const {
  return issuer;
}

void
ACI::setNotBefore(const UtcTime& notBefore) {
  this->notBefore = notBefore;
}

const UtcTime&
ACI::getNotBefore() const {
  return notBefore;
}

void
ACI::setNotAfter(const UtcTime& notAfter) {
  this->notAfter = notAfter;
}

const UtcTime&
ACI::getNotAfter() const {
  return notAfter;
}

void
ACI::setCanonAlg(const string& canonAlg) { 
  this->canonAlg = canonAlg; 
}

const string&
ACI::getCanonAlg() const {
  return canonAlg;
}

const string&
ACI::getSignatureAlg() const {
  return signatureAlg;
}


bool 
ACI::checkValidity() const {
  UtcTime utcTime;

  return checkValidity(utcTime);
}


bool 
ACI::checkValidity(const UtcTime& utcTime) const {
  return utcTime >= notBefore && utcTime <= notAfter;
}

string 
ACI::getSignable() const {
  ostringstream os;
  AkentiOutputStream aos(os);
  
  aos << getType() << ' ';
  writeObject(aos, false);
  return os.str();
}

void
ACI::sign(const string& signatureAlg, const PrivateKey& privateKey) {
  Signature sig = Signature::getInstance(signatureAlg);
  
  sig->initSign(privateKey);
  this->signatureAlg = signatureAlg; 
  signature = sig->sign(getSignable());
}

const string&
ACI::getSignature() const {
  return signature;
}

bool 
ACI::verify(const PublicKey& publicKey) const {
  Signature sig = Signature::getInstance(signatureAlg);

  sig->initVerify(publicKey);
  sig->update(getSignable());
  return sig->verify(signature);
}

void
ACI::writeObject(AkentiOutputStream &drain) const {
  writeObject(drain, true);
}

void
ACI::writeObject(AkentiOutputStream &drain, bool includeSig) const {
  drain << 'V' << version
        << ' ' << id 
        << ' ' << issuer
        << ' ' << notBefore
        << ' ' << notAfter;
  
  if (version == 2) {
    drain << ' ' << canonAlg;
  }

  drain << ' ' << signatureAlg << ' ';

  doWrite(drain);
  
  if (includeSig) {
    int len = signature.length();

    drain << ' ' << len << ' ';
    drain.write(signature.data(), len);
  }
}

void
ACI::readObject(AkentiInputStream &source) {
  if (source.get() != 'V') {
    throw Exception("reading version failed");
  }

  int version;

  source >> version;

  setVersionNumber(version);

  source >> id 
         >> issuer
         >> notBefore
         >> notAfter;

  if (version == 2) {
    source >> canonAlg;
  }

  source >> signatureAlg;

  doRead(source);

  int len;

  source >> len;

  if (len <= 0) {
    throw Exception("AkentiCertificateImpl:readObject: len < 0");
  }
  
  if (source.get() != ' ') {
    throw Exception("AkentiCertificateImpl:readObject: expecting blank space");
  }

  signature = string(len, '\0');
  source.read((char*) signature.data(), len);
}

void 
ACI::toXML(DOMElement* cert) const {
  DOMNode *temp = cert->getOwnerDocument();
  DOMDocument *doc = (DOMDocument*) temp;
  DOMElement *header = doc->createElement(XMLUtil::convert("Header"));

  header->setAttribute(XMLUtil::convert("Type"), XMLUtil::convert(getType().c_str()));
  header->setAttribute(XMLUtil::convert("SignatureDigestAlg"), XMLUtil::convert(getSignatureAlg().c_str()));
  header->setAttribute(XMLUtil::convert("CanonAlg"), XMLUtil::convert(getCanonAlg().c_str()));
  
  ostringstream os;

  os << version;
  
  header->setAttribute(XMLUtil::convert("Version"), XMLUtil::convert(os.str().c_str()));

  DOMElement* element = doc->createElement(XMLUtil::convert("UID"));
  
  element->appendChild(doc->createTextNode(XMLUtil::convert(getUniqueID().c_str())));
  header->appendChild(element);
  element = doc->createElement(XMLUtil::convert("Issuer"));
  issuer.toXML(element);
  header->appendChild(element);
  element = doc->createElement(XMLUtil::convert("ValidityPeriod"));
  
  string utc = getNotBefore().toString("YYYY-MM-DDTHH:MM:SSZ");
  element->setAttribute(XMLUtil::convert("Begin"), XMLUtil::convert(utc.c_str()));

  utc = getNotAfter().toString("YYYY-MM-DDTHH:MM:SSZ");
  element->setAttribute(XMLUtil::convert("End"), XMLUtil::convert(utc.c_str()));

  header->appendChild(element);

  DOMElement *signable = doc->createElement(XMLUtil::convert("SignablePart"));

  signable->appendChild(header);
  
  string certType = getType() + "Cert";

  element = doc->createElement(XMLUtil::convert(certType.c_str()));
  doToXML(element);
  signable->appendChild(element);
  cert->appendChild(signable);
  element = doc->createElement(XMLUtil::convert("Signature"));
  
  string b64EncodedSig = Base64::encode(getSignature());

  b64EncodedSig = Base64Util::format(b64EncodedSig);

  element->appendChild(doc->createTextNode(XMLUtil::convert(b64EncodedSig.c_str())));
  cert->appendChild(element);
}

void 
ACI::fromXML(const DOMElement* cert) {
  DOMNode* child = cert->getFirstChild();

  if (child == NULL || child->getNodeType() != DOMNode::ELEMENT_NODE) {
    throw Exception("AkentiCertificateImpl:fromXML");
  }

  DOMElement* signable = (DOMElement*) child;
  DOMNode* temp = signable->getFirstChild();

  if (temp == NULL || temp->getNodeType() != DOMNode::ELEMENT_NODE) {
    throw Exception("AkentiCertificateImpl:fromXML");
  }

  DOMElement* header = (DOMElement*) temp;
  string rep = XMLUtil::convert(header->getAttribute(XMLUtil::convert("Type")));

  if (rep != type) {
    throw Exception("AkentiCertificateImpl:fromXML");
  } 

  signatureAlg = XMLUtil::convert(header->getAttribute(XMLUtil::convert("SignatureDigestAlg")));
  canonAlg = XMLUtil::convert(header->getAttribute(XMLUtil::convert("CanonAlg")));
  rep = XMLUtil::convert(header->getAttribute(XMLUtil::convert("Version")));
  std::istringstream is(rep.c_str());

  if (!(is >> version)) {
    throw Exception("AkentiCertificateImpl:fromXML");
  }

  temp = header->getFirstChild();

  if (temp == NULL) {
    throw Exception("AkentiCertificateImpl:fromXML");
  } else if ((temp = temp->getFirstChild()) == NULL) {
    throw Exception("AkentiCertificateImpl:fromXML");
  } 

  // XXX
  id = XMLUtil::convert(temp->getNodeValue());
  temp = header->getFirstChild()->getNextSibling();

  if (temp == NULL || temp->getNodeType() != DOMNode::ELEMENT_NODE) {
    throw Exception("AkentiCertificateImpl:fromXML");
  }

  issuer.fromXML((DOMElement*) temp);
  temp = temp->getNextSibling();

  if (temp == NULL || temp->getNodeType() != DOMNode::ELEMENT_NODE) {
    throw Exception("AkentiCertificateImpl:fromXML");
  }

  DOMElement* validityPeriod = (DOMElement*) temp;

  string begin = XMLUtil::convert(validityPeriod->getAttribute(XMLUtil::convert("Begin")));
  string end = XMLUtil::convert(validityPeriod->getAttribute(XMLUtil::convert("End")));

  if(begin.size() > 13){
    notBefore = UtcTime(begin,"YYYY-MM-DDTHH:MM:SSZ");
  } else {
    notBefore = UtcTime(begin);
  }

  if(end.size() > 13){
    notAfter = UtcTime(end,"YYYY-MM-DDTHH:MM:SSZ");
  } else {
    notAfter = UtcTime(end);
  }

  child = header->getNextSibling();
  doFromXML((DOMElement*) child);
  child = signable->getNextSibling();

  if (child == NULL) {
    signature = "NO_SIGNATURE";
  } else if ((temp = child->getFirstChild()) == NULL) {
    throw Exception("AkentiCertificateImpl:fromXML:no signature");
  } else {
    rep = XMLUtil::convert(temp->getNodeValue());

    rep = Base64Util::format(rep);

    signature = Base64::decode(rep);
  }
}

string 
ACI::paramString() const {
  ostringstream os;
  
  os << '['
     << "type" << ':' << getType()
     << ','
     << "version" << ":V" << getVersionNumber()
     << ','
     << "uid" << ':' << getUniqueID() 
     << ','
     << "issuer" << ':' << getIssuer().paramString()
     << ',' 
     << "notBefore" << ':' << getNotBefore().paramString()
     << ',' 
     << "notAfter" << ':' << getNotAfter().paramString()
     << ',';

  doParamString(os);

  os << ','
     << "canonAlg" << ':' << getCanonAlg()
     << ','
     << "signatureAlg" << ':' << getSignatureAlg()
     << ','
     <<  "signature " << ':' << Base64::encode(getSignature()) 
     << ']';

  return os.str();
}

string 
ACI::toText(int indentAmt) const {
  ostringstream os;
  os << Util::fText(indentAmt, getType() + " Certificate") << endl
     << Util::fText(indentAmt + 3, "Version: V") << getVersionNumber() << endl 
     << Util::fText(indentAmt + 3, "Cert ID: " + getUniqueID()) << endl
     << Util::fText(indentAmt + 3, "Issuer: ") << endl
     << Util::fText(indentAmt + 6, 
                    "Name: " + getIssuer().getName().toString()) << endl
     << Util::fText(indentAmt + 6, 
                     "CA  : " + getIssuer().getGuarantor().toString()) 
		     << endl
     << Util::fText(indentAmt + 3, "Validity:") << endl
     << Util::fText(indentAmt + 6, 
                     "Not Before: " + getNotBefore().toString("YYYY-MM-DDTHH:MM:SSZ")) << endl
     << Util::fText(indentAmt + 6, 
                     "Not After : " + getNotAfter().toString("YYYY-MM-DDTHH:MM:SSZ")) << endl;
  doToText(indentAmt, os);
  os << Util::fText(indentAmt,
                     "Canon Algorithm: " + getCanonAlg()) << endl
     << Util::fText(indentAmt,
                     "Signature Algorithm: " + getSignatureAlg()) << endl
     << Util::fText(indentAmt, 
                     "Signature: " + Base64::encode(getSignature())) << endl;
  return os.str();
}

string
ACI::toLogString() const {
  string log;

  log += type;
  log += '_'; 
  log += "CERT_INFO";
  log += "=";
  log += "\"";
  log += paramString();
  log += "\"";
  return log;
}

}
