#include "AttributeCertificateImpl.hpp"
#include "AkentiInputStream.hpp"
#include "AkentiOutputStream.hpp"
#include "Exception.hpp"
#include "Error.hpp"
#include "Util.hpp"
#include "AttributeInfo.hpp"
#include "Xerces-Includes.hpp"
#include "XMLUtil.hpp"

using std::endl;

namespace Akenti {

typedef AttributeCertificateImpl ACI;

ACI::AttributeCertificateImpl() 
  : AkentiCertificateImpl("Attribute"), 
    principal(),
    attribute("NO_NAME"), 
    value("NO_VALUE"),
    condition(NULL) {
}

AttributeCertificateImpl* 
ACI::newInstance() const {
  return new AttributeCertificateImpl;
}

ACI::~AttributeCertificateImpl() {
  delete(condition);
}

string
ACI::getHashString() const {
  return 
    Util::getHashDN(principal.getName())
    + Util::getHashDN(principal.getGuarantor())
    + attribute
    + value;
}

void 
ACI::setPrincipal(const AkentiPrincipal& principal) {
  this->principal = principal;
}

const AkentiPrincipal& 
ACI::getPrincipal() const {
  return principal;
}
 
void 
ACI::setAttribute(const string& attribute) {
  if (!attribute.length()) {
    throw Error("AttributeCertificateImpl:setAttribute:attribute = estring");
  }

  this->attribute = attribute;
}

const string&
ACI::getAttribute() const {
  return attribute;
}

void 
ACI::setValue(const string& value) {
  if (!value.length()) {
    throw Error("AttributeCertificateImpl:setValue:value = estring");
  }

  this->value = value;
}

const string&
ACI::getValue() const {
  return value;
}

void
ACI::setCondition(const Condition& condition) {
  this->condition = new Condition(condition);
}

bool 
ACI::containsCondition() const {
  return condition != NULL;
}

const Condition& 
ACI::getCondition() const {
  if (condition == NULL) {
    throw Error("AttributeCertificateImpl:getCondition:no condition set");
  }
  
  return (*condition);
}

const string&
ACI::getConstraint() const {
  if (condition == NULL) {
    throw Error("AttributeCertificateImpl:getConstraint:no condition set");
  }

  return condition->getConstraint();
}

const vector<AttributeInfo>&
ACI::getAttributeInfos() const {
  if (condition == NULL) {
    throw Error("AttributeCertificateImpl:getAttributeInfos:no condition set");
  }

  return condition->getAttributeInfos();
}

void 
ACI::doWrite(AkentiOutputStream& drain) const {
  drain << principal
        << ' ' << attribute 
        << ' ' << value;

  bool haveCondition = (condition != NULL);

  drain << ' ' << haveCondition;
  

  if (haveCondition) {
    drain << ' ' << *condition;
  }
}

void 
ACI::doRead(AkentiInputStream& source) {
  delete(condition);
  condition = NULL;
  source >> principal
         >> attribute 
         >> value;
  
  bool haveCondition; 
  
  source >> haveCondition;

  if (haveCondition) {
    condition = new Condition();
    source >> *condition;
  }
}

void 
ACI::doToXML(DOMElement* attrCertInfo) const {
  DOMDocument *doc = attrCertInfo->getOwnerDocument();
  DOMElement *temp = doc->createElement(XMLUtil::convert("SubjectAndCA"));

  principal.toXML(temp);
  attrCertInfo->appendChild(temp);
  temp = doc->createElement(XMLUtil::convert("AttrName"));
  temp->appendChild(doc->createTextNode(XMLUtil::convert(attribute.c_str())));
  attrCertInfo->appendChild(temp);
  temp = doc->createElement(XMLUtil::convert("AttrValue"));
  temp->appendChild(doc->createTextNode(XMLUtil::convert(value.c_str())));
  attrCertInfo->appendChild(temp);

  if (condition != NULL) {
    temp = doc->createElement(XMLUtil::convert("Condition"));
    condition->toXML(temp);
    attrCertInfo->appendChild(temp);
  }
}

void 
ACI::doFromXML(const DOMElement* attrCertInfo) {
  DOMNode *child = attrCertInfo->getFirstChild();
  
  if (child == NULL || child->getNodeType() != DOMNode::ELEMENT_NODE) {
    throw Exception("AttributeCertificateImpl:doFromXML:");
  }

  principal.fromXML((DOMElement*) child);
  child = child->getNextSibling();

  if (child == NULL || child->getNodeType() != DOMNode::ELEMENT_NODE) {
    throw Exception("AttributeCertificateImpl:doFromXML:");
  }

  DOMNode *temp = child->getFirstChild();

  if (temp == NULL) {
    throw Exception("AttributeCertificateImpl:doFromXML:");
  } 

  attribute = XMLUtil::convert(temp->getNodeValue());

  child = child->getNextSibling();

  if (child == NULL || child->getNodeType() != DOMNode::ELEMENT_NODE) {
    throw Exception("AttributeCertificateImpl:doFromXML:");
  } else if ((temp = child->getFirstChild()) == NULL) {
    throw Exception("AttributeCertificateImpl:doFromXML:");
  } 

  value = XMLUtil::convert(temp->getNodeValue());

  if ((child = child->getNextSibling()) == NULL) {
    condition = NULL;
  } else if (child->getNodeType() != DOMNode::ELEMENT_NODE) {
    throw Exception("AttributeCertificateImpl:doFromXML:");
  } else {
    condition = new Condition();
    condition->fromXML((DOMElement *) child);
  }
}

void
ACI::doParamString(ostream& os) const {
  os << "subject" << ':' << getPrincipal().paramString()
     << ','
     << "attribute" << ':' <<  getAttribute()
     << ','
     << "value" << ':' << getValue()
     << ',';
  
  bool haveCondition = (condition != NULL);

  os << haveCondition;

  if (haveCondition) {
     os << ',';
     os << condition->paramString();
  }
}

void 
ACI::doToText(int indentAmt, ostream& os) const {
  os << Util::fText(indentAmt + 3, "Subject:") << endl 
     << Util::fText(indentAmt + 6, 
                     "Name: " + getPrincipal().getName().toString()) << endl 
     << Util::fText(indentAmt + 6, 
                     "CA  : " + getPrincipal().getGuarantor().toString()) << endl 
     << Util::fText(indentAmt + 3, "AttributeValuePair:") << endl 
     << Util::fText(indentAmt + 6, getAttribute() + " = " + getValue()) << endl;

  if (containsCondition()) {
    os << Util::fText(indentAmt, "Condition: ") << endl 
       << Util::fText(indentAmt + 6, "Constraint: " + getConstraint())
       << endl 
       << Util::fText(indentAmt + 6, "Attribute Infos: ")
       << endl; 

    static string TYPES[] = {"SYSTEM", "X509", "AKENTI", "EXT_AUTH"}; 
    const vector<AttributeInfo>& ais = getAttributeInfos();

    for (int i = 0; i < ais.size(); ++i) {
      AttributeInfo::Type type = ais[i].getType(); 

      os << Util::fText(indentAmt + 9,  
	                 ais[i].getAttribute() + " = " + 
		         ais[i].getValue() + ", type = " + TYPES[type])
         << endl;

      if (type == AttributeInfo::X509) { 
        const vector<DistinguishedName>& cas = ais[i].getCAS();  

        for (int k = 0; k < cas.size(); k++) {
          os << Util::fText(indentAmt + 12, "CA  : " +  cas[k].toString())
             << endl; 
        } 
      } else if (type == AttributeInfo::AKENTI || type == AttributeInfo::EXT_AUTH) {
        const vector<AkentiPrincipal>& 
          aps = type == AttributeInfo::AKENTI? 
            ais[i].getIssuers() :  ais[i].getAuthorities();

        for (int j = 0; j < aps.size(); j++) {
          os << Util::fText(indentAmt + 12, 
	                     "Name: " + aps[j].getName().toString()) << endl 
             << Util::fText(indentAmt + 12, 
	                     "CA  : " + aps[j].getGuarantor().toString()) 
			     << endl; 
        } // end for
        
        const vector<Directory>& dirs = ais[i].getDirectories(); 

        if (dirs.size() > 0) {
          if (type == AttributeInfo::AKENTI) {
            os << Util::fText(indentAmt + 12, "Attribute Directories: ") << endl;
          } else {
            os << Util::fText(indentAmt + 12, "Authority Directories: ") << endl;
          }

          for (int l = 0; l < dirs.size(); l++) {
            os << Util::fText(indentAmt + 15, dirs[l].toString()) << endl;
          }          
        } 

        if (type == AttributeInfo::EXT_AUTH) { 
          const vector<string>& args = ais[i].getArguments();

          if (args.size() > 0) {
            os << Util::fText(indentAmt + 12, "Arguments: ") << endl;

            for (int k = 0; k < args.size(); k++) {
              os << Util::fText(indentAmt + 15, args[k]) << endl;
            }
          }
        }
      } // end if
    } // end for   
  }
}

} // end of namespace
