#include "UseCondCertificateImpl.hpp"
#include "AttributeInfo.hpp"
#include "AkentiInputStream.hpp"
#include "AkentiOutputStream.hpp"
#include "StringTokenizer.hpp"
#include "Exception.hpp"
#include "Error.hpp"
#include "Util.hpp"
#include "Xerces-Includes.hpp"
#include "XMLUtil.hpp"

using namespace std;

namespace Akenti {

typedef UseCondCertificateImpl UCCI;

UCCI::UseCondCertificateImpl() 
  : AkentiCertificateImpl("UseCondition"),
    resource("NO_RESOURCE"),
    localScope(true),
    critical(false),
    condition(), 
    actions() {
}

UseCondCertificateImpl*
UCCI::newInstance() const {
  return new UseCondCertificateImpl;
}

UCCI::~UseCondCertificateImpl() {
}

string
UCCI::getHashString() const {
  return getResource();
}

void 
UCCI::setResource(const string& resource) {
  this->resource = resource;
}

const string&
UCCI::getResource() const {
  return resource;
}

void 
UCCI::setLocalScope(bool localScope) {
  this->localScope = localScope;
}

bool 
UCCI::isScopeLocal() const {
  return localScope; 
}

void 
UCCI::setCritical(bool critical) {
  this->critical = critical;
}

bool
UCCI::isCritical() const {
  return critical;
}

void 
UCCI::setCondition(const Condition& condition) {
 this->condition = condition;
}

const Condition&
UCCI::getCondition() const {
  return condition;
}

const string& 
UCCI::getConstraint() const {
  return condition.getConstraint();
}

const vector<AttributeInfo>& 
UCCI::getAttributeInfos() const {
  return condition.getAttributeInfos();
}

void 
UCCI::setActions(const vector<string>& actions) {
  this->actions = actions;
}

const vector<string>&
UCCI::getActions() const {
  return actions;
}

void
UCCI::doWrite(AkentiOutputStream& drain) const {
  drain << resource
        << ' ' << localScope
        << ' ' << critical
        << ' ' << condition;

  int size = actions.size();

  drain << ' ' << size;

  for (int i = 0; i < size; ++i) {
    drain << ' ' << actions[i];
  }

  if (version == 1) {
    drain << ' ' << (size = subjectCAS.size());

    for (int i = 0; i < size; ++i) {
      drain << ' ' << subjectCAS[i];
    }
  }
}

void
UCCI::doRead(AkentiInputStream& source) {
  actions.clear(); 

  source >> resource 
         >> localScope 
         >> critical
         >> condition;
 
  int size; 
  
  source >> size;
  
  if (size < 0) {
    throw Exception(
      "UseCondCertificateImpl:doRead:reading actions (size < 0)"
    );
  }

  for (int i = 0; i < size; ++i) {
    string action;

    source >> action;
    actions.push_back(action);
  }

  if (version == 1) {
    source >> size;

    if (size < 0) {
      throw Exception(
        "UseCondCertificateImpl:doRead:reading subjectCAS (size < 0)"
      );
    }

    for (int i = 0; i < size; i++) {
      DistinguishedName subjectCA;

      source >> subjectCA;
      subjectCAS.push_back(subjectCA);
    }
  }
}

void 
UCCI::doToXML(DOMElement* ucCertInfo) const {
  ucCertInfo->setAttribute(XMLUtil::convert("scope"), XMLUtil::convert(localScope? "local" : "sub-tree"));
  ucCertInfo->setAttribute(XMLUtil::convert("critical"), XMLUtil::convert(critical? "true" : "false"));

  DOMDocument *doc = ucCertInfo->getOwnerDocument();
  DOMElement *temp = doc->createElement(XMLUtil::convert("ResourceName"));

  temp->appendChild(doc->createTextNode(XMLUtil::convert(resource.c_str())));
  ucCertInfo->appendChild(temp);

  temp = doc->createElement(XMLUtil::convert("Condition"));
  condition.toXML(temp);
  ucCertInfo->appendChild(temp);

  for (int i = 0; i < actions.size(); i++) {
    string sb = actions[i];
    temp = doc->createElement(XMLUtil::convert("Action"));

    temp->appendChild(doc->createTextNode(XMLUtil::convert(sb.c_str())));
    ucCertInfo->appendChild(temp);
  }
}

void
UCCI::doFromXML(const DOMElement* ucCertInfo) {
  actions.clear();

  string temp = XMLUtil::convert(ucCertInfo->getAttribute(XMLUtil::convert("scope")));
   
  if (temp == "local") {
    localScope = true;
  } else if (temp == "sub-tree") {
    localScope = false;
  } else {
    throw Exception("UseConditionImpl:doFromXML:scope:" + temp);
  }

  temp = XMLUtil::convert(ucCertInfo->getAttribute(XMLUtil::convert("critical")));

  if (temp == "true") {
    critical = true;
  } else if (temp == "false") {
    critical = false;
  } else {
    throw Exception("UseConditionImpl:doFromXML:critical:" + temp);
  }

  DOMNode *child = ucCertInfo->getFirstChild();

  if (child == NULL || child->getNodeType() != DOMNode::ELEMENT_NODE) {
    throw Exception("UseConditionImpl:doFromXML:uc info missing");
  }

  DOMNode *node = child->getFirstChild();

  if (node == NULL) {
    throw Exception("UseConditionImpl:doFromXML:resource missing");
  }

  resource = XMLUtil::convert(node->getNodeValue());

  child = child->getNextSibling();

  if (child == NULL || child->getNodeType() != DOMNode::ELEMENT_NODE) {
    throw Exception("UseConditionImpl:doFromXML:condition missing");
  }

  condition.fromXML((DOMElement*) child);

  DOMNodeList *actionList = ucCertInfo->getElementsByTagName(XMLUtil::convert("Action"));

  for (int i = 0; i < actionList->getLength(); i++) {
    node = actionList->item(i);
    string action = XMLUtil::convert(node->getFirstChild()->getNodeValue());

    if (! action.empty()) 
      actions.push_back(action);
  }
}

void
UCCI::doParamString(ostream& os) const {
  os << "resource" << ':' << getResource() 
     << ','
     << "scope" << '=' << (isScopeLocal()? "true" : "false")
     << "critical" << '=' << (isCritical() ? "true" : "false")
     << condition.paramString();

  const vector<string>& actions = getActions();
  int size = actions.size();

  if (size) {
    os << '['
       << actions[0];

    for (int i = 1; i < actions.size(); i++) {
      os << ',' << actions[i];
    }
     
    os << ']';
  }
}

void 
UCCI::doToText(int indentAmt, ostream& os) const {
  os << Util::fText(indentAmt + 3, "Resource: " + getResource())
     << endl; 

  string scope = isScopeLocal()? "local" : "subtree"; 
  string critical = isCritical()? "true" : "false"; 

  os << Util::fText(indentAmt + 3, "Use Condition: scope=" + scope + 
                     " critical=" + critical) << endl
     << Util::fText(indentAmt + 6, "Constraint: " + getConstraint()) << endl;

  const vector<string>& actions = getActions();
  int size = actions.size();

  if (size) {
    os << Util::fText(indentAmt + 6, "Actions: ")
       << Util::fText(0, actions[0]);

    for (int i = 1; i < size; ++i) { 
       os << Util::fText(0, ", " + actions[i]);
    }
  }

  os << endl;
  const vector<AttributeInfo>& ais = getAttributeInfos();

  if (ais.size() > 0) {
    os << Util::fText(indentAmt + 6, "Attribute Infos:") << endl;
  } 

  static string TYPES[] = {"SYSTEM", "X509", "AKENTI", "EXT_AUTH"}; 

  for (int i = 0; i < ais.size(); ++i) {
    AttributeInfo::Type type = ais[i].getType(); 

    os << Util::fText(indentAmt + 9,  
             ais[i].getAttribute() + " = " + 
             ais[i].getValue() + ", type = " + TYPES[type])
       << endl;

    if (type == AttributeInfo::X509) { 
      const vector<DistinguishedName>& cas = ais[i].getCAS();  

      for (int k = 0; k < cas.size(); k++) {
        os << Util::fText(indentAmt + 12, "CA  : " +  cas[k].toString())
           << endl; 
      } 
    } else if (type == AttributeInfo::AKENTI || type == AttributeInfo::EXT_AUTH) {
      const vector<AkentiPrincipal>& 
        aps = type == AttributeInfo::AKENTI? 
          ais[i].getIssuers() :  ais[i].getAuthorities();

      for (int j = 0; j < aps.size(); j++) {
        os << Util::fText(indentAmt + 12, 
                    "Name: " + aps[j].getName().toString()) << endl 
           << Util::fText(indentAmt + 12, 
                    "CA  : " + aps[j].getGuarantor().toString()) 
		     << endl; 
      } // end for
        
      const vector<Directory>& dirs = ais[i].getDirectories(); 

      if (dirs.size() > 0) {
        os << Util::fText(indentAmt + 12, "Directories: ") << endl;

        for (int l = 0; l < dirs.size(); l++) {
          os << Util::fText(indentAmt + 15, dirs[l].toString()) << endl;
        }          
      } 

      if (type == AttributeInfo::EXT_AUTH) { 
        const vector<string>& args = ais[i].getArguments();

        if (args.size() > 0) {
          os << Util::fText(indentAmt + 12, "Arguments: ") << endl;

          for (int k = 0; k < args.size(); k++) {
            os << Util::fText(indentAmt + 15, args[k]) << endl;
          }
        }
      }
    } // end if
  }
}

}
