#include "AttributeInfo.hpp"
#include "StringUtil.hpp"
#include "AkentiOutputStream.hpp"
#include "AkentiInputStream.hpp"
#include "URL.hpp"
#include "Exception.hpp"
#include "Error.hpp"
#include "Util.hpp"
#include "XMLUtil.hpp"
#include "Xerces-Includes.hpp"

#include <iostream>
#include <sstream>
using namespace std;

namespace Akenti {

typedef AttributeInfo AI;

AI::AttributeInfo(const string& attribute, 
                  const string& value) 
  : type(SYSTEM), attribute(attribute), value(value) {
  if (!attribute.size() || !value.size()) {
    throw Error("AttributeInfo:constuctor(1):illegal arg");
  }
}

AI::AttributeInfo(const string& attribute, 
                  const string& value,
                  const vector<DistinguishedName>& cas) 
  : type(X509), attribute(attribute), value(value), cas(cas) {
  if (!attribute.size() || !value.size() || !cas.size()) {
    throw Error("AttributeInfo:constuctor(2):illegal arg");
  }
}

AI::AttributeInfo(bool isAkentiAttribute,
                  const string& attribute, 
                  const string& value,
                  const vector<AkentiPrincipal>& aps,
                  const vector<Directory>& dirs,
                  const vector<string>& args) 
  : type(isAkentiAttribute? AKENTI : EXT_AUTH),
    attribute(attribute),
    value(value),
    aps(aps) ,
    dirs(dirs),
    args(args) {
  if (!attribute.size() || !value.size() || !aps.size()) {
    throw Error("AttributeInfo:constuctor(4):illegal arg");
  } 
}

void 
AI::operator=(const AttributeInfo& ai) {
  type = ai.type;
  attribute = ai.attribute;
  value = ai.value;
  cas = ai.cas;
  aps = ai.aps;
  dirs = ai.dirs;
  args = ai.args;
}

bool 
AI::operator==(const AttributeInfo& ai) const {
  return ai.hasType(type) && ai.hasAttribute(attribute) && ai.hasValue(value);
}

bool 
AI::hasType(Type type) const {
  return this->type == type;
}

bool 
AI::hasAttribute(const string& attribute) const {
    return StringUtil::equalsIgnoreCase(this->attribute, attribute);
}

bool 
AI::hasValue(const string& value) const {
  return this->value == value;
}

bool 
AI::hasCA(const DistinguishedName& ca) const {
  if (type != X509) {
    throw Error("Attribute:hasCA:not x509 attribute");
  }
  
  bool ret = Util::compare(ca, cas[0]);

  for (int i = 1; !ret && i < cas.size(); ++i) {
    ret = Util::compare(ca, cas[i]);
  }

  return ret;
}

bool 
AI::hasIssuer(const AkentiPrincipal& issuer) const {
  if (type != AKENTI) {
    throw Error("Attribute:hasIssuer:not AKENTI attribute");
  }

  bool ret = Util::compare(issuer.getName(), aps[0].getName())
             && 
             Util::compare(issuer.getGuarantor(), aps[0].getGuarantor());

  for (int i = 1; !ret && i < aps.size(); ++i) {
    ret = Util::compare(issuer.getName(), aps[i].getName())
          && 
          Util::compare(issuer.getGuarantor(), aps[i].getGuarantor());
  }

  return ret;
}

bool 
AI::hasAuthority(const AkentiPrincipal& auth) const {
  if (type != EXT_AUTH) {
    throw Error("Attribute:hasIssuer:not AKENTI attribute");
  }

  bool ret = Util::compare(auth.getName(), aps[0].getName())
             && 
             Util::compare(auth.getGuarantor(), aps[0].getGuarantor());

  for (int i = 1; !ret && i < aps.size(); ++i) {
    ret = Util::compare(auth.getName(), aps[i].getName())
          && 
          Util::compare(auth.getGuarantor(), aps[i].getGuarantor());
  }

  return ret;
}

const string& 
AI::getAttribute() const {
  return attribute;
}

const string& 
AI::getValue() const {
  return value;
}

AI::Type 
AI::getType() const {
  return type;
}

const vector<DistinguishedName>& 
AI::getCAS() const {
  if (type != X509) {
    throw Error("AttributeInfo:getCAS:not X509 attribute");
  }

  return cas;
}

const vector<AkentiPrincipal>& 
AI::getIssuers() const {
  if (type != AKENTI) {
    throw Error("AttributeInfo:getIssuers:not AKENTI attribute");
  }

  return aps;
}

const vector<AkentiPrincipal>& 
AI::getAuthorities() const {
  if (type != EXT_AUTH) {
    throw Error("AttributeInfo:getAuthorities:not EXT_AUTH attribute");
  }

  return aps;
}

bool
AI::parseDirs(vector<URL>& urls) const {
  bool ret = true;

  for (int i = 0; i < dirs.size(); ++i) {
    URL url;

    ret = ret ? dirs[i].parseDir(url) : false;
    urls.push_back(url);
  }

  return ret;
}

const vector<Directory>&
AI::getDirectories() const {
  if (type != EXT_AUTH && type != AKENTI) {
    throw Error("AttributeInfo:getDirectories:not EXT_AUTH/AKENTI attribute");
  }

  return dirs;
}

const vector<string>&
AI::getArguments() const {
  if (type != EXT_AUTH) {
    throw Error("AttributeInfo:getArguments:not EXT_AUTH attribute");
  }
  
  return args;
}

void 
AI::writeObject(AkentiOutputStream& drain) const {
  drain << (int) type 
        << ' ' << attribute 
        << ' ' << value;

  if (type == SYSTEM) {
    return;
  }

  // only one of these vectors is used.
  drain << ' ' << (int) (cas.size() + aps.size());

  for (int i = 0; i < cas.size(); ++i) {
    drain << ' ' << cas[i];
  }

  if (type == X509) {
    return;
  }

  for (int i = 0; i < aps.size(); ++i) {
    drain << ' ' << aps[i];
  }

  int size = dirs.size();

  drain << ' ' << size;

  for (int i = 0; i < size; ++i) {
    drain << ' ' << dirs[i];
  }
 
  if (type == AKENTI) {
   return;
  }
  
  size = args.size();

  drain << ' ' << size;

  for (int i = 0; i < args.size(); ++i) {
    drain << ' ' << args[i];
  }
}

void 
AI::readObject(AkentiInputStream& source) {
  cas.clear();
  aps.clear();
  dirs.clear();
  args.clear();
  
  int temp;

  source >> temp 
         >> attribute 
         >> value;

  this->type = (Type) temp;

  if (type == SYSTEM) {
    return;
  }
  
  source >> temp;

  if (temp <= 0) {
    throw Exception("AttributeInfo:readObject:cas/issuers/auths size <= 0");
  }

  if (type == X509) {
    for (int i = 0; i < temp; ++i) {
      DistinguishedName dn;

      source >> dn;
      cas.push_back(dn);
    }
    
    return;

  } else {
    for (int i = 0; i < temp; ++i) {
      AkentiPrincipal ap;

      source >> ap;
      aps.push_back(ap);
    }
  }

  source >> temp;

  if (temp < 0) {
    throw Exception("Attribute:readObject:dirs.size < 0");
  }

  for (int i = 0; i < temp; ++i) {
    Directory dir;

    source >> dir;
    dirs.push_back(dir);
  }
  
  if (type == AKENTI) {
    return;
  }
  
  source >> temp;

  if (temp < 0) {
    throw Exception("AttributeInfo:readObject:args.size < 0");
  }

  for (int i = 0; i < temp; ++i) {
    string arg;
    
    source >> arg;
    args.push_back(arg);
  }
}

void 
AI::toXML(DOMElement* attrInfo) const {
  static string TYPES[] = { "SYSTEM", "X509", "AKENTI", "EXT_AUTH" };

  DOMDocument *doc = attrInfo->getOwnerDocument();
  DOMElement *temp = doc->createElement(XMLUtil::convert("AttrName"));

  attrInfo->setAttribute(XMLUtil::convert("type"), XMLUtil::convert(TYPES[type].c_str()));
  temp->appendChild(doc->createTextNode(XMLUtil::convert(attribute.c_str())));
  attrInfo->appendChild(temp);
  temp =  doc->createElement(XMLUtil::convert("AttrValue"));
  temp->appendChild(doc->createTextNode(XMLUtil::convert(value.c_str())));
  attrInfo->appendChild(temp);

  if (type == SYSTEM) {
    return;
  }

  if (type == X509) {
    for (int i=0; i < cas.size(); i++) {
      temp = doc->createElement(XMLUtil::convert("CADN"));
      temp->appendChild(doc->createTextNode(XMLUtil::convert(cas[i].toString().c_str())));
      attrInfo->appendChild(temp);
    }

    return;
  }

  for (int i = 0; i < aps.size(); i++) {
    temp = doc->createElement(XMLUtil::convert("Principal"));
    aps[i].toXML(temp);
    attrInfo->appendChild(temp);
  }

  if (dirs.size() > 0) {
    DOMElement *attrDirs = doc->createElement(XMLUtil::convert("AttrDirs"));

    for (int i=0; i < dirs.size(); i++) {
      temp = doc->createElement(XMLUtil::convert("URL"));
      dirs[i].toXML(temp);
      attrDirs->appendChild(temp);
    }

    attrInfo->appendChild(attrDirs);
  }

  if (type == AKENTI) {
    return;
  }

  for (int i=0; i < args.size(); i++) {
    DOMElement *arg = doc->createElement(XMLUtil::convert("ExtArg"));

    arg->appendChild(doc->createTextNode(XMLUtil::convert(args[i].c_str())));
    attrInfo->appendChild(arg);
  }
}

void 
AI::fromXML(const DOMElement* attrInfo) {
  cas.clear();
  aps.clear();
  dirs.clear();
  args.clear();

  string typeString = XMLUtil::convert(attrInfo->getAttribute(XMLUtil::convert("type")));
  
  DOMNode *attrNameNode = attrInfo->getFirstChild();
  DOMNode *temp = attrNameNode->getFirstChild();

  attribute = XMLUtil::convert(temp->getNodeValue());

  DOMNode *attrValueNode = attrNameNode->getNextSibling();

  temp = attrValueNode->getFirstChild();
  value = XMLUtil::convert(temp->getNodeValue());
  
  if (typeString == "SYSTEM") {
    type = SYSTEM;
    return;
  } else if (typeString == "X509") {
    type = X509;

    DOMNodeList *caList = attrInfo->getElementsByTagName(XMLUtil::convert("CADN"));
    int len = caList->getLength();
    
    if (!len) {
      throw Exception("AttributeInfo:fromXML:cas size = 0");
    }

    for (int i = 0;i < len; ++i) {
      temp = caList->item(i)->getFirstChild();
      cas.push_back(XMLUtil::convert(temp->getNodeValue()));
    }

    return;
  } 

  DOMNodeList *principalList = attrInfo->getElementsByTagName(XMLUtil::convert("Principal"));
  int len = principalList->getLength();

  if (!len) {
    throw Exception("AttributeInfo:fromXML:issuers/auths size = 0");
  }

  for (int i = 0;i < len; ++i) {
    temp = principalList->item(i);

    DOMElement *element = (DOMElement*) temp;
    AkentiPrincipal ap;

    ap.fromXML(element);
    aps.push_back(ap);
  }
  
  DOMNodeList *attrDirs = attrInfo->getElementsByTagName(XMLUtil::convert("AttrDirs"));
  
  if ((len = attrDirs->getLength()) > 1) {
    throw Exception("AttributeInfo:fromXML:AttrDirs not grouped");
  } else if (len == 1) {
    temp = attrDirs->item(0);
    attrDirs = ((DOMElement*) temp)->getElementsByTagName(XMLUtil::convert("URL"));
    len = attrDirs->getLength();

    for (int i = 0; i < len; ++i) { 
      temp = attrDirs->item(i);

      DOMElement *element = (DOMElement*) temp;
      Directory dir;
   
      dir.fromXML(element);
      dirs.push_back(dir);
    }
  }

  if (typeString == "AKENTI") {
    type = AKENTI;
  } else if (typeString == "EXT_AUTH") {
    type = EXT_AUTH;
    DOMNodeList *argList = attrInfo->getElementsByTagName(XMLUtil::convert("ExtArg"));
    
    len = argList->getLength();
    
    for (int i = 0; i < len; i++) {
      temp = argList->item(i)->getFirstChild();
      
      args.push_back(XMLUtil::convert(temp->getNodeValue()));
    }
  } else {
    throw Exception("AttributeInfo:fromXML:unknown type " + typeString);
  }
}

string 
AI::paramString() const {
  static string TYPES[] = { "SYSTEM", "X509", "AKENTI", "EXT_AUTH" };
  Type type = getType();
  string params;

  params += '[';
  params += TYPES[type];;
  params += ',';
  params += getAttribute();
  params += '=';
  params += getValue();
  params += ',';
  params += '[';

  int temp;

  if (type == X509) {
    temp = cas.size();
    params += cas[0].paramString();

    for (int i = 1; i < temp - 1; ++i) {
      params += ',';
      params += cas[i].paramString();
    }

  } else if (type != SYSTEM) {
    temp = aps.size();
    params += aps[0].paramString();

    for (int i = 1; i < temp; ++i) {
      params += ',';
      params += aps[i].paramString();
    }
  }

  params += ']';
  params += ',';
  params += '[';
  temp = dirs.size();

  if (temp) {
    params += dirs[0].paramString();

    for (int i = 1; i < temp; ++i) {
      params += ',';
      params += dirs[i].paramString();
    }
  }

  params += ']';
  params += ',';
  params += '[';
  temp = args.size();

  if (temp) {
    params += args[0];

    for (int i = 1; i < temp; ++i) {
      params += ',';
      params += args[i];
    }
  }

  params += ']';
  params += ']';
  return params;
}

}
