#ifndef ATTRIBUTE_INFO_HPP
#define ATTRIBUTE_INFO_HPP

#include <vector>
#include "AkentiPrincipal.hpp"
#include "Directory.hpp"

using std::vector;

namespace Akenti {

/**
  * An <i>AttributeInfo</i> consists mainly of an attribute, a value, 
  * and possibly a list of cas/issuers/authorities depending on its type.
  *
  * <BR>
  *
  * Note that these attribute/value pairs can be resource-specific
  * or user-specific.
  *
  * <pre> 
  *
  * We have considered four types of attributes,
  *
  *  1) an X509 attribute.
  *
  *     This attribute/value pair puts a constraint on the 
  *     dn of a user accessing a resource.
  *
  *     Example: OU=LBL
  *      
  *     We ask for a non-empty list of the cas that we 
  *     are willing to accept.
  *
  *     For now the cas have to be the immediate cas of
  *     the user accessing the resource.
  *     
  *  2) an AKENTI attribute.
  *
  *     This is an Akenti Certificate based attribute. 
  *     An attribute certificate is fetched and used 
  *     to evaluate this attribute/value pair.
  *
  *     Example:  group=SOME_GROUP
  *
  *     We ask for a non-empty list of issuers
  *     that can issue such attribute certificates.
  *     
  *  3) an EXT_AUTH attribute.
  *
  *     This attribute/value pair is evaluated or attested
  *     by some authority. Akenti Engine may be configured
  *     to evaluate such attribute/value pairs.  
  *
  *     Example: time=10:00  and SomeTimeAuthority
  *
  *     We ask for a non-empty list of authorities that help 
  *     evaluate such attribute/value pairs.
  *       
  *  4) a SYSTEM attribute.
  *
  *     This is the simplest of all attributes. 
  *     Akenti Engine may be configured to evaluate such 
  *     such attribute/value pairs.
  *
  *     Example: DiscSpace=10 MB
  *
  * AttributeInfo {                                                                             
  *    Type type;                                                                               
  *    string name;  
  *    string value; 
  *    vector&lt;DistinguishedName&gt;     cas; // For X509
  *    vector&lt;AkentiPrincipal&gt;       aps; // For AKENTI and EXT_AUTH attribute certs issuers
  *    vector&lt;Directory&gt;            dirs; // For AKENTI  used to look up attribute certs  
  *                                             //     EXT_AUTH used to contact authoriies        
  *    vector&ltstring&gt                 args; // For EXT_AUTH additional args passed to the authorities
  * } 
  *
  *
  * The writeObject output is: 
  *
  * For a SYSTEM attribute:   
  *     0 attribute value
  *
  * For an X509 attribute: 
  *     1 attribute value caCount ca+
  *
  * For an AKENTI attribute:   
  *     2 attribute value apCount ap+ dirCount dir*
  *
  * For an EXT_AUTH attribute:   
  *     3 attribute value apCount ap+ dirCount dir* argCount arg*
  *
  * Where + means one or more and * means 0 or more
  *
  * </pre>
  *
  * @see AkentiObject
  * @see AkentiPrincipal
  * @see Directory 
  * @see DistinguishedName 
  *
  * @version 1.1 00/05/01
  *
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

class AttributeInfo : public AkentiObject {

public:
  
  /**
    * Type constants.
    *
    * @type Version
    *
    * Values of <i>Type</i> are: <br>
    * <blockquote>
    *   "SYSTEM",  <br>
    *   "X509",    <br>
    *   "AKENTI",  <br>
    *   "EXT_AUTH" <br>
    * </blockquote>
    *
    * @type Type
    */
  enum Type { SYSTEM, X509, AKENTI, EXT_AUTH };

  /**
    * Constructs a SYSTEM attribute/value pair.
    *
    **/
  AttributeInfo(const string& attribute = "NO_ATTRIBUTE", 
                const string& value = "NO_VALUE");

  /**
    * Constructs an X509 attribute/value pair.
    *
    * @param attribute string name of the  attribute
    * @param value string value of the attribute
    * @param cas vector a list of acceptable cas
    *
    * @see DistinguishedName
    */
  AttributeInfo(const string& attribute, 
                const string& value,
                const vector<DistinguishedName>& cas);

  /**
    * Constructs an AKENTI/EXT_AUTH attribute/value pair.
    * 
    * @param isAkentiAttribute    name of this  attribute  
    * @param attribute            name of this  attribute  
    * @param value                value of this attribute
    * @param auths                a list of stakeholders/authorities 
    * @param dirs                 a list of directories
    * @param args                 a list of arguments 
    *
    * @see AkentiPrincipal
    */
  AttributeInfo(bool isAkentiAttribute,
                const string& attribute, 
                const string& value,
                const vector<AkentiPrincipal>& aps,
                const vector<Directory>& dirs = vector<Directory>(),
                const vector<string>& args = vector<string>());

  /**
    * Assignement operator.
    */
  void operator=(const AttributeInfo& ai);

  /**
    * Returns <i>true</i> if this <i>AttributeInfo</i> 
    * is equal <i>ai</i>.
    * 
    * Note here we only compare the types, the attributes
    * and the values. The attribute comparison is not case 
    * sensitive.
    **/
  bool operator==(const AttributeInfo& ai) const;

  /**
    * Returns <i>true</i> if this <i>AttributeInfo</i>
    * is of <i>AttributeInfo::Type type>/i>.
    */
  bool hasType(Type type) const;

  /**
    * Returns <i>true</i> if this <i>AttributeInfo</i>
    * contains <i>attribute</i>.
    */
  bool hasAttribute(const string& attribute) const;

  /**
    * Returns <i>true</i> if this <i>AttributeInfo</i>
    * contains <i>value</i>.
    */
  bool hasValue(const string& value) const;

  /**
    * Returns <i>true</i> if this <i>AttributeInfo</i>'s
    * list of cas contains <i>ca</i>.
    */
  bool hasCA(const DistinguishedName& ca) const;

  /**
    * Returns <i>true</i> if this <i>AttributeInfo</i>'s
    * list of issuers contains <i>issuer</i>.
    */
  bool hasIssuer(const AkentiPrincipal& issuer) const;

  /**
    * Returns <i>true</i> if this <i>AttributeInfo</i>'s
    * list of authorities contains <i>authority</i>.
    */
  bool hasAuthority(const AkentiPrincipal& authority) const;

  /**
    * Returns name of the <i>attribute</i>.
    */
  const string& getAttribute() const;

  /**
    * Returns the <i>value</i> of the attribute.
    */
  const string& getValue() const;

  /**
    * Returns the <i>type</i>.
    */
  Type getType() const;

  /**
    * Returns the <i>cas</cas>.
    *
    * @exception Error if this is not an X509 attribute
    */
  const vector<DistinguishedName>& getCAS() const;

  /**
    * Returns the issuers, <i>aps</i>.
    *
    * @exception Error if this is not an Akenti attribute
    */
  const vector<AkentiPrincipal>& getIssuers() const;

  /**
    * Returns the authorities, <i>aps</i>
    *
    * @exception Error if this is not an EXT_AUTH attribute
    * @return the issuers of this attribute
    */
  const vector<AkentiPrincipal>& getAuthorities() const;

  /**
    * Returns the directories <i>dirs</i>
    *
    * @exception Error if this is not an AKENTI or EXT_AUTH attribute
    */
  const vector<Directory>& getDirectories() const;

  /**
    * Returns true if all dirs parse correctly.
    *
    * @param urls out parameter
    */
  bool parseDirs(vector<URL>& urls) const;

  /**
    * Returns the arguments <i>args</i>.
    *
    * @exception Error if this is not an EXT_AUTH attribute
    */
  const vector<string>& getArguments() const;

  /**
    * Writes this object to the <i>AkentiOutputStream</i>.
    *
    * @param drain AkentiOutputStream used to write this object
    *
    * @see AkentiOutputStream
    */
  void writeObject(AkentiOutputStream& drain) const;

  /**
    * Reads this object from an <i>AkentiInputStream</i>.
    *
    * @param source An AkentiInputStream
    *
    * @see AkentiInputStream
    */
  void readObject(AkentiInputStream& source);

  /**
    * Populates the XML element <i>attrInfo</i> from this <i>AttributeInfo</i>.
    */
  void toXML(DOMElement* attrInfo) const;

  /**
    * Initializes this <i>AttributeInfo</i> from the XML element
    * <i>attrInfo</i>.
    */
  void fromXML(const DOMElement* attrInfo);


  /**
    * Used for debugging.
    *
    * @return a string representation of this attribute
    */
  string paramString() const;

private:

  Type type;                          // SYSTEM, X509, AKENTI, EXT_AUTH  

  string attribute;                   // attribute's name

  string value;                       // attribute's value
  
  vector<DistinguishedName> cas;      // used for  X509 attributes 

  vector<AkentiPrincipal>   aps;      // used for AKENTI/EXT_AUTH attributes

  vector<Directory>        dirs;      // used for AKENTI/EXT_AUTH attributes

  vector<string>           args;      // used for EXT_AUTH attributes

};

} // end of name space

#endif
