#include "ConditionalActions.hpp"
#include "AkentiOutputStream.hpp"
#include "AkentiInputStream.hpp"
#include "Exception.hpp"
#include "Error.hpp"
#include "Xerces-Includes.hpp"
#include "XMLUtil.hpp"

namespace Akenti {

typedef ConditionalActions CA;

CA::ConditionalActions() :
  condition(), actions(), critical(false) {
}
  
CA::ConditionalActions(const Condition& condition, 
                       const vector<string>& actions, 
                       bool critical)
 : condition(condition), actions(actions), critical(critical) {
   int size = actions.size();

   if (size) {
     for (int i = 0; i < size; ++i) {
       if (!actions[i].length()) {
         throw Error("ConditionalActions:constructor:illegal argument(2)");
       }
     }
   }
}

CA::~ConditionalActions() {
}

void 
CA::operator=(const ConditionalActions conditionalActions) {
  condition = conditionalActions.condition;
  actions = conditionalActions.actions;
  critical = conditionalActions.critical;
}

const Condition& 
CA::getCondition() const {
  return condition;
}

bool 
CA::parseConstraint(Expression& expr) const {
  return condition.parseConstraint(expr);
}

const string& 
CA::getConstraint() const {
  return condition.getConstraint();
}

const vector<AttributeInfo>& 
CA::getAttributeInfos() const {
  return condition.getAttributeInfos();
}

const vector<string>& 
CA::getActions() const {
  return actions;
}

bool 
CA::isCritical() const {
  return critical;
}

bool 
CA::isValid() const {
 return condition.isValid();
}

void 
CA::writeObject(AkentiOutputStream& drain) const {
  drain << condition;
  
  int size = actions.size();
  
  drain << ' ' << size;

  for (int i = 0; i < actions.size(); ++i) {
    drain << ' ' << actions[i];
  } 
  
  drain << ' ' << critical;
}

void 
CA::readObject(AkentiInputStream& source) {
  actions.clear();
  source >> condition;
  
  int size;

  source >> size;

  if (size < 0) {
    throw Exception("ConditionalActions:readObject:size <=0");
  }

  for (int i = 0;  i < size; ++i) { 
    string action;
    
    source >> action;
    actions.push_back(action);
  }
    
  source >> critical;
}

void 
CA::toXML(DOMElement* condActInfo) const {  
  int numActions = actions.size();
  
  if (numActions < 1)
    throw Exception("ConditionalActions:toXML:fewer than 1 action specified");

  condActInfo->setAttribute(XMLUtil::convert("critical"), XMLUtil::convert(isCritical()? "true" : "false"));

  DOMDocument *doc = condActInfo->getOwnerDocument();
  DOMElement *temp =  temp = doc->createElement(XMLUtil::convert("Condition"));

  condition.toXML(temp);
  condActInfo->appendChild(temp);

  for (int i = 0; i < numActions; i++) {
    string str = actions[i];
    temp = doc->createElement(XMLUtil::convert("Action"));

    temp->appendChild((DOMNode*) doc->createTextNode(XMLUtil::convert(str.c_str())));
    condActInfo->appendChild(temp);
  }
}

void 
CA::fromXML(const DOMElement* condActInfo) {
  actions.clear();

  string temp = XMLUtil::convert(condActInfo->getAttribute(XMLUtil::convert("critical")));

  if (temp == "true") {
    critical = true;
  } else if (temp == "false") {
    critical = false;
  } else {
    throw Exception("ConditionalActions:fromXML:critical:" + temp);
  }

  DOMNode *child = condActInfo->getFirstChild();
  
  if (child == NULL || child->getNodeType() != DOMNode::ELEMENT_NODE) {
    throw Exception("ConditionalActions:fromXML:condition missing");
  }

  condition.fromXML((DOMElement*) child);
  child = child->getNextSibling();

  if (child == NULL || child->getNodeType() != DOMNode::ELEMENT_NODE) {
    throw Exception("ConditionalActions:fromXML:actions missing");
  }

  DOMNodeList *list = condActInfo->getElementsByTagName(XMLUtil::convert("Action"));
  
  if (list->getLength() < 1)
    throw Exception("ConditionalActions:fromXML:actions missing");

  for (int i = 0; i < list->getLength(); i++) {
    DOMNode *node = list->item(i);

    if (node == NULL) {
      throw Exception("ConditionalActions:fromXML:actions missing in actions tag");
    }

    string action = XMLUtil::convert(node->getFirstChild()->getNodeValue());
  
    actions.push_back(action);
  }
}

string 
CA::paramString() const {
  string params;

  params += '[';
  params += "condition";
  params += '=';
  params += getCondition().paramString();
  params += ',';

  params += ',';
  params += "critical"; 
  params += '=';
  params += isCritical() ? "true" : "false";

  if (actions.size()) {
    params += "actions";
    params += '=';
  
    vector<string> actions = getActions();

    params += '[';
    params += actions[0];

    for (int i = 1;  i < actions.size(); ++i) { 
      params += ',';
      params += actions[i];
    }

    params += ']';
  }

  params += ']';
  return params; 
}

string
CA::toLogString() const {
  string log;

  log = "COND_ACTION_INFO=";
  log += '"';
  log += paramString(); 
  log += '"';
  return log;
}

} // end of namespace Akenti
