#include "LdapCertificateRetrievalImpl.hpp"
#include "IOException.hpp"
#include "Exception.hpp"
#include "Error.hpp"
#include "DistinguishedName.hpp"
#include "Base64.hpp"
#include "URL.hpp"
#include "lber.h"
#include "ldap.h"

using namespace Akenti;

typedef LdapCertificateRetrievalImpl LCRI;

/**
  * form the X500 DN in the proper order
  * i.e. reverse the X400 format provided
  **/

static string 
toX500DN(const string& dn) {
  DistinguishedName dname(dn);
  int size = dname.size();
  pair<string, string> attrval = dname.getPairAt(size - 1);
  string x500dn = attrval.first;

  x500dn +=  '=';
  x500dn += attrval.second;

  for (int i= size - 2; i >= 0; --i) {
    x500dn +=  ',';
    attrval = dname.getPairAt(i);
    x500dn += attrval.first;
    x500dn +=  '=';
    x500dn += attrval.second;
  }

  return x500dn;
}

LCRI::LdapCertificateRetrievalImpl() {
}

LCRI::~LdapCertificateRetrievalImpl() {
}

void
LCRI::getContent(const URL& url, map<string, string>& args, vector<string>& results, string& enc) const {
  char* attributes[2] = { NULL, NULL };

  if (args["type"] == "X509") {
    attributes[0] = "userCertificate;binary";
    enc= "DER";
  } else {
    throw Exception(
      "LdapCertificateRetrievalImpl:getContent:unsupported type:" + args["type"]
    );
  }

  int port = url.getPort();

  if (port < 0) {
    port = 389;
  }

  LDAP* ld = ldap_init((char *) url.getHost().c_str(), port);

  if (!ld) { 
    throw IOException(
      "LdapCertificateRetrievalImpl:getContent:can't connect to" + url.toString()
    );
  }
  
  if (ldap_bind_s(ld, "", "", LDAP_AUTH_SIMPLE) != LDAP_SUCCESS) {
    ldap_unbind(ld);
    ld = NULL;
    
    throw IOException(
      "LdapCertificateRetrievalImpl:getContent:bind failed"
    );
  }

  string base = toX500DN(args["base"]);
  string filter = args["filter"];

  if (!filter.length()) {
    filter = "cn=*";
  }

  int scope;
  
  if (args["scope"] == "onelevel") {
    scope = LDAP_SCOPE_ONELEVEL;
  } else if (args["scope"] == "subtree") {
    scope = LDAP_SCOPE_SUBTREE; 
  } else {
    scope = LDAP_SCOPE_BASE;
  } 

  int ret = ldap_search(
    ld, (char*) base.c_str(), scope, (char*) filter.c_str(), attributes, 0
  );

  if (ret == -1) {
    ldap_unbind(ld); ld = NULL;

    throw IOException(
      "LdapCertificateRetrievalImpl:getContent:ldap search failed"
    );
  }

  LDAPMessage* res = NULL;

  while (ldap_result(ld, LDAP_RES_ANY, 0, NULL, &res) == LDAP_RES_SEARCH_ENTRY) {
    LDAPMessage* entry = ldap_first_entry(ld, res);
    
    for (; entry != NULL; entry = ldap_next_entry(ld, res)) {
      struct berval** values = ldap_get_values_len(ld, entry, attributes[0]);

      if (values != NULL) {
        for (int i = 0; values[i] != NULL; ++i) {
          string cert((char *) values[i]->bv_val, values[i]->bv_len);

          results.push_back(cert);
        }

        ber_bvecfree(values); values = NULL;
      }
    }

    ldap_msgfree(res); res = NULL;
  }
  
  ldap_unbind(ld); ld = NULL;
}

