#include <string>
#include "Expression.hpp"
#include "ExpressionEvaluator.hpp"
#include "ExpressionParser.hpp"
#include "ExpressionVisitor.hpp"
#include "TreeNode.hpp"

using namespace Akenti;

typedef Expression EX;

EX::Expression() : root() {
}

EX::Expression(const string& expression) {
  ExpressionParser parser;

  root = parser.parseExpression(expression);
}

EX::~Expression() {
}

bool 
EX::isNull() {
  return root->getTokenType() == ENULL;
}

string
EX::toString() const {
  return toString(root.get());
}

string 
EX::toString(const TreeNode* n) const {
  switch(n->getTokenType()) {
    case AND: case OR:
      return "( " 
             + toString(n->getLeftChild())
             + ' ' + n->getData() + ' '
             + toString(n->getRightChild())
             + " )";
    case NOT:
      n = n->getLeftChild();

      if (n->getTokenType() == NOT) {
        return "! ( " + toString(n) + " )";
      }

      return "! " + toString(n);
    case CONST:
      return n->getData();
    case ENULL:
      return "()";
    case INVALID:
      throw Error("SHOULD NEVER HAPPEN");
    default:             
      return "( "  
             + toString(n->getLeftChild())  
             + ' ' + n->getData() + ' ' 
             + toString(n->getRightChild()) 
             + " )";
  }
}

string
EX::toLogString() const {
  string log;
  
  log += "EXPRESSION=";
  log += '"';
  log += toString();
  log += '"';
  return log;
}

bool 
EX::evaluate(ExpressionEvaluator& evaluator) const {
  return evaluate(root.get(), evaluator);
}

bool 
EX::evaluate(const TreeNode* n, ExpressionEvaluator& evaluator) const {
  switch(n->getTokenType()) {
    case AND:
      return evaluate(n->getLeftChild(), evaluator) 
             && 
             evaluate(n->getRightChild(), evaluator);
    case OR:
      return evaluate(n->getLeftChild(),evaluator) 
             || 
             evaluate(n->getRightChild(), evaluator);
    case NOT:
      return ! evaluate(n->getLeftChild(), evaluator);
    case CONST:
      throw Error("Expression:evaluate:should never happen(1)");
    case ENULL:
      return true;
    case INVALID:
      throw Error("Expression:evaluate:should never happen(2)");
    default:
      return evaluator.evaluate(
        n->getTokenType(),
        n->getLeftChild()->getData(),
        n->getRightChild()->getData()
      );
  }
}

int 
EX::reduce(ExpressionEvaluator& evaluator, Expression& expr) const {
  TreeNode* n = NULL;
  int ret = reduce(root.get(), evaluator, &n);
    
  switch (ret) {
    case -1:
      expr.root = n;
      break;
  }

  return ret;
}

int 
EX::reduce(const TreeNode* n, ExpressionEvaluator& evaluator, TreeNode** out) const {
  TreeNode* out1 = NULL, *out2 = NULL;
  int ret1, ret2;

  switch(n->getTokenType()) {
    case AND:
      ret1 = reduce(n->getLeftChild(), evaluator, &out1);

      if (!ret1) {
        return 0;
      }  else if (ret1 == 1) {
        return reduce(n->getRightChild(), evaluator, out);
      }

      ret2  = reduce(n->getRightChild(), evaluator, &out2);

      if (!ret2) {
        delete(out1);
        return 0;
      } else if (ret2 == 1) {
        *out = out1;
        return -1;
      } else {
        *out = new TreeNode(n->getTokenType(), n->getData(), out1, out2);
        return -1;
      }
    case OR:
      ret1 = reduce(n->getLeftChild(), evaluator, &out1);

      if (ret1 == 1) {
        return 1;
      } else if (!ret1) {
        return reduce(n->getRightChild(), evaluator, out);
      } 

      ret2 = reduce(n->getRightChild(), evaluator, &out2);

      if (!ret2) {
        *out = out1;
        return -1;
      } else if (ret2 == 1) {
        delete(out1);
        return 1;
      } else {
        *out = new TreeNode(n->getTokenType(), n->getData(), out1, out2);
        return -1;
      }
    case NOT:
      ret1 = reduce(n->getLeftChild(), evaluator, &out1);
        
      if (!ret1 || ret1 == 1) {
        return !ret1;
      }

      *out = new TreeNode(n->getTokenType(), n->getData(), out1, NULL);
      return -1;
    case CONST:
      throw Error("Expression:reduce:should never happen(1)");
    case ENULL:
      return 1;
    case INVALID:
      throw Error("Expression:reduce:should never happen(2)");
    default:
      Expression expr;

      ret1 = evaluator.reduce(
        n->getTokenType(),
        n->getLeftChild()->getData(),
        n->getRightChild()->getData(),
        expr
      );
        
      if (ret1 == -1) {
        if (expr.root) {
          *out = expr.root->clone();
        } else {
          *out = n->clone();
        }
      }
        
      return ret1;
  }
}

void 
EX::visit(ExpressionVisitor& visitor) const {
  visit(root.get(), visitor);
}

void 
EX::visit(const TreeNode* n, ExpressionVisitor& visitor) const {
  switch(n->getTokenType()) {
    case AND: case OR:
      visit(n->getLeftChild(), visitor);
      visit(n->getRightChild(), visitor);
      break;
    case NOT:
      visit(n->getLeftChild(), visitor);
      break;
    case ENULL:
      return;
    case CONST:
      throw Error("Expression:visit:should never happen(1)");
    case INVALID:
      throw Error("Expression:visit:should never happen(2)");
    default:
      visitor.visit(n->getTokenType(), 
                    n->getLeftChild()->getData(),
                    n->getRightChild()->getData());
  }
}
