#include <utility>
#include <memory>
#include <sstream>
#include <map>
#include <set>
#include "AkentiEngine.hpp"
#include "AKRuntime.hpp"
#include "CertificateChain.hpp"
#include "CertificateVerifier.hpp"
#include "CertificateCache.hpp"
#include "AkentiCertificate.hpp"
#include "ServerContext.hpp"
#include "CertificateRetrievalFactory.hpp"
#include "PolicyCollector.hpp"
#include "CapabilityCertGenerator.hpp"
#include "UseCondCollector.hpp"
#include "ConditionalActions.hpp"
#include "Policy.hpp"
#include "PolicyAuthorizer.hpp"
#include "Certificate.hpp"
#include "Properties.hpp"
#include "StringUtil.hpp"
#include "Logger.hpp"
#include "DistinguishedName.hpp"
#include "Exception.hpp"
#include "Error.hpp"
#include "PolicyContext.hpp"
#include "MessageDigest.hpp"
#include "StringTokenizer.hpp"
#include "Util.hpp"

using namespace Akenti;
using namespace std;

typedef AkentiEngine AE;

Properties AE::resourceMap = Properties();

AE::AkentiEngine() : akRuntime(AKRuntime::createAKRuntime()) {
}

void 
AE::setDesiredActions(const set<string>& desiredActions) {
  akRuntime->setDesiredActions(desiredActions);
  akRuntime->setCacheCap(false);
}

void
AE::setRestrictedAttributes(const multimap<string, string>& desiredAttributes) {
  akRuntime->setRestrictedAttributes(desiredAttributes);
  akRuntime->setCacheCap(false);
}

void
AE::setExternalEvaluator(const ExternalEvaluator& externalEvaluator) {
  akRuntime->setExternalEvaluator(externalEvaluator);
}

AE::~AkentiEngine() {
  akRuntime->getExternalEvaluator().setAKRuntime(NULL);
  delete(akRuntime);
}

bool
AE::checkAccess(const string& resource, 
                const AkentiPrincipal& ap, 
                CapabilityCertificate& cert,
                AkentiMessage& mesg) {
  bool ret = checkAccess(resource, ap, akRuntime, mesg);
  
  if (ret) {
    PolicyContext context = akRuntime->getPolicyContext();

    cert = context->getCapabilityCert();
  }

  return ret;
}

bool
AE::checkAccess(const string& resource, 
                const CertificateChain& chain, 
                CapabilityCertificate& cert,
                AkentiMessage& mesg) {
  AkentiPrincipal ap(chain->getPrincipal(), chain->getGuarantor());
  bool ret = checkAccess(resource, ap, akRuntime, mesg);
  
  if (ret) {
    PolicyContext context = akRuntime->getPolicyContext();

    cert = context->getCapabilityCert();
  }

  return ret;
}

bool
AE::checkAccess(const string& resource, 
                const AkentiPrincipal& ap, 
                PolicyContext& context,
                AkentiMessage& mesg) {
  bool ret = checkAccess(resource, ap, akRuntime, mesg);

  context = akRuntime->getPolicyContext();
  return ret;
}

bool
AE::checkAccess(const string& resource, 
                const CertificateChain& chain, 
                PolicyContext& context,
                AkentiMessage& mesg) {
  AkentiPrincipal ap(chain->getPrincipal(), chain->getGuarantor());
  bool ret = checkAccess(resource, ap, akRuntime, mesg);

  context = akRuntime->getPolicyContext();
  return ret;
}

bool
AE::checkAccess(const string& alias, 
                const AkentiPrincipal& ap, 
                AKRuntime* akRuntime,
                AkentiMessage& mesg) const {
  string uid = "NO_UID";

  if (ap.getName().containsAttribute("CN")) {
    uid = ap.getName().getValue("CN");
  }

  std::auto_ptr<Logger> logger(initLogger("AkentiEngine"));

  bLog(LOG_MON,"CHECK_ACCESS","RESOURCE=" + alias + " UID="+uid);
  
  if (alias.length() == 0) {
    fLog(LOG_MON,"CHECK_ACCESS","RESOURCE=empty string");
    mesg = AkentiMessage(401);
    return false;
  }

  string resource = resourceMap.getValue(alias, alias);

  if (resource != alias) {
    iLog(LOG_MON,"RESOURCE_MAPPED_TO","RESOURCE=" + resource);
  }

  akRuntime->setResource(resource);
  akRuntime->setResourceAltName(alias);
  akRuntime->setAkentiPrincipal(ap);
  
  CertificateCache& cache = akRuntime->getCertificateCache();
  StringTokenizer st(resource, "/ ", false);

  cache.setDirectoryName(st.nextToken());

  if (akRuntime->getCacheCap()) {
    MessageDigest md = MessageDigest::getInstance("MD5");

    md->update(alias);
    md->update(Util::getHashDN(ap.getName()));
    md->update(Util::getHashDN(ap.getGuarantor()));

    vector<AKCertHolder> holders = cache.getAKCertHolders(
      md->digest(), "Capability"
    );

    if (holders.size() > 0) { // found a cached capabilty
      bool status = false;
      // NEED TO DO SOME MORE CHECKING
      CapabilityCertificate cert = (CapabilityCertificate&) holders[0]->getCertificate();

      akRuntime->setCapabilityCert(cert);

      if (cert->getConditionalActions().size()) {
        mesg = AkentiMessage(105);
      } else { 
        mesg = AkentiMessage(104);
      }

      status = true;
      sLog(LOG_MON,"CHECK_ACCESS", "found cap in cache");
      return status;
    }
  }

  if (!collectAndVerifyPolicies(akRuntime, mesg)) { 
    cache.updateExternalCache();
    fLog(LOG_MON,"CHECK_ACCESS");
    return false; 
  } else if(!collectAndVerifyUseConds(akRuntime, mesg)) {
    cache.updateExternalCache();
    fLog(LOG_MON,"CHECK_ACCESS");
    return false; 
  } else if (!doAuthorization(akRuntime, mesg)) {
    cache.updateExternalCache();
    fLog(LOG_MON,"CHECK_ACCESS");
    return false; 
  } else {
    long cacheInterval = akRuntime->getCacheInterval();
    PolicyContext ctx = akRuntime->getPolicyContext();
    CapabilityCertificate cert = generateCapabilityCert(ctx, cacheInterval);

    akRuntime->setCapabilityCert(cert);

    if (akRuntime->getCacheCap()) {
      cache.addCertificate(
        cert,   
        akRuntime->getCALevel(akRuntime->getLevel()),
        cert->getNotBefore(),
        cert->getNotAfter()
      );
    }
      
    cache.updateExternalCache();
    sLog(LOG_MON,"CHECK_ACCESS");
    return true;
  }
}

bool
AE::checkPolicy(const string& resource,
                PolicyContext& context,
                AkentiMessage& mesg) {
  string uid = "NO_UID";
  std::auto_ptr<Logger> logger(initLogger("AkentiEngine"));

  bLog(LOG_MON,"CHECK_POLICY","RESOURCE=" + resource);

  string dresource = resourceMap.getValue(resource, resource);

  if (dresource != resource) {
    iLog(LOG_MON,"RESOURCE_MAPPED_TO","RESOURCE=" + dresource);
  }

  akRuntime->setResource(dresource);
  akRuntime->setResourceAltName(resource);

  CertificateCache& cache = akRuntime->getCertificateCache();
  StringTokenizer st(resource, "/ ", false);

  cache.setDirectoryName(st.nextToken());
  context = akRuntime->getPolicyContext();

  if (!collectAndVerifyPolicies(akRuntime, mesg)) {
    cache.updateExternalCache();
    fLog(LOG_MON,"CHECK_POLICY");
    return false;
  } else if (!collectAndVerifyUseConds(akRuntime, mesg)) {
    cache.updateExternalCache();
    fLog(LOG_MON,"CHECK_POLICY");
    return false;
  } 

  cache.updateExternalCache();
  mesg = AkentiMessage(150);
  sLog(LOG_MON,"CHECK_POLICY");
  return true;
}

bool
AE::checkIssuer(const string& resource, const AkentiPrincipal& issuer, AkentiMessage& mesg) {
  string uid = "NO_UID";
  std::auto_ptr<Logger> logger(initLogger("AkentiEngine"));

  bLog(LOG_MON,"CHECK_ISSUER","RESOURCE=" + resource);

  string dresource = resourceMap.getValue(resource, resource);

  if (dresource != resource) {
    iLog(LOG_MON,"RESOURCE_MAPPED_TO","RESOURCE=" + dresource);
  }

  akRuntime->setResource(dresource);
  akRuntime->setResourceAltName(resource);

  CertificateCache& cache = akRuntime->getCertificateCache();
  StringTokenizer st(resource, "/ ", false);

  cache.setDirectoryName(st.nextToken());

  if (!collectAndVerifyPolicies(akRuntime, mesg)) {
    cache.updateExternalCache();
    fLog(LOG_MON,"CHECK_ISSUER");
    return false;
  } else if (!collectAndVerifyUseConds(akRuntime, mesg)) {
    cache.updateExternalCache();
    fLog(LOG_MON,"CHECK_ISSUER");
    return false;
  } 

  bLog(LOG_MON,"VERIFYING_ISSUER", issuer.toLogString());

  PolicyContext context = akRuntime->getPolicyContext();
  const vector<Policy>& policies = context->getPolicies();
  int idx = policies.size() - 1;
  const PolicyCertificate& polCert = policies[idx]->getPolicyCertificate(); 

  const vector<UseCondIssuerGroup>& groups = polCert->getUseCondIssuerGroups();
  bool isIssuer = groups[0].containsUCIssuer(issuer);

  for (int i = 1; !isIssuer && i < groups.size(); ++i) {
    isIssuer = groups[i].containsUCIssuer(issuer);
  }

  if (!isIssuer) {
    fLog(LOG_MON,"VERIFYING_ISSUER");
    mesg = AkentiMessage(350);
    fLog(LOG_MON,"CHECK_ISSUER");
    return false;
  } 

  AkentiX509Certificate x509Cert;

  if (!akRuntime->getCertificateVerifier().verify(issuer, x509Cert)) {
    fLog(LOG_MON,"VERIFYING_ISSUER");
    mesg = AkentiMessage(351);
    fLog(LOG_MON,"CHECK_ISSUER");
    return false;
  }

  sLog(LOG_MON,"VERIFYING_ISSUER");

  cache.updateExternalCache();
  mesg = AkentiMessage(140);
  sLog(LOG_MON,"CHECK_ISSUER");
  return true;
}

bool 
AE::collectAndVerifyPolicies(AKRuntime* akRuntime, 
                             AkentiMessage& mesg) const {
  bLog(LOG_MON,"POLICIES_COLLECTION_PHASE");

  PolicyCollector policyCollector(akRuntime);

  if (!policyCollector.collectPolicies()) {
    fLog(LOG_MON,"POLICIES_COLLECTION_PHASE","COLLECTING_POLICIES=FAIL");
    mesg = policyCollector.getAkentiMessage();
    return false;
  } else if (!policyCollector.verifyPolicies()) {
    fLog(LOG_MON,"POLICIES_COLLECTION_PHASE","CHECKING_POLICIES=FAIL");
    mesg = policyCollector.getAkentiMessage();
    return false;
  }

  sLog(LOG_MON,"POLICIES_COLLECTION_PHASE");
  return true;
}

bool
AE::collectAndVerifyUseConds(AKRuntime* akRuntime, AkentiMessage& mesg) const {
  bLog(LOG_MON,"UC_COLLECTION_PHASE");

  UseCondCollector ucCollector(akRuntime);

  if (!ucCollector.collectUseCondCertificates()) {
    fLog(LOG_MON,"UC_COLLECTION_PHASE");
    mesg = ucCollector.getAkentiMessage();
    return false;
  }

  sLog(LOG_MON,"UC_COLLECTION_PHASE");
  return true;
}

bool
AE::doAuthorization(AKRuntime* akRuntime, 
                    AkentiMessage& mesg) const {
  bLog(LOG_MON,"AUTHORIZATION_PHASE"); 

  AkentiX509Certificate subjectCert;
  const AkentiPrincipal& ap = akRuntime->getAkentiPrincipal();

  bLog(LOG_MON,"VERIFYING_USER", ap.toLogString()); 

  if (!akRuntime->getCertificateVerifier().verify(ap, subjectCert)) {
    fLog(LOG_MON,"VERIFYING_USER");
    fLog(LOG_MON,"AUTHORIZATION_PHASE");
    mesg = AkentiMessage(306);
    return false;
  }

  sLog(LOG_MON,"VERIFYING_USER");
  akRuntime->setSubjectCert(subjectCert);

  PolicyAuthorizer authorizer(akRuntime);
  bool ret = authorizer.authorize();

  mesg = authorizer.getAkentiMessage();

  if (!ret) {
    fLog(LOG_MON,"AUTHORIZATION_PHASE");
    return false;
  }

  const PolicyContext& context = akRuntime->getPolicyContext();
  const vector<string>& actions = context->getUnconditionalActions();
    
  for (int i = 0; i < actions.size(); ++i) {
    iLog(LOG_MON,"UNCONDITIONAL_ACTION", "ACTION=" + actions[i]);
  }

  const vector<ConditionalActions>& condActionsList = context->getConditionalActionsList();

  for (int i = 0; i < condActionsList.size(); ++i) {
    iLog(LOG_MON,"COND_ACTIONS", condActionsList[i].toLogString());
  }

  sLog(LOG_MON,"AUTHORIZATION_PHASE");
  return true;
}

CapabilityCertificate
AE::generateCapabilityCert(const PolicyContext& context, long cacheInterval) const {
  CapabilityCertGenerator gen;
  UtcTime notBefore, notAfter(cacheInterval);

  gen.setVersionNumber(2);
  gen.setNotBefore(notBefore);
  gen.setNotAfter(notAfter);
  gen.setPrincipal(context->getAkentiPrincipal());
  gen.setVersionNumber(2);
  gen.setPublicKey(context->getSubjectCert()->getSubjectPublicKey());
  gen.setResource(context->getResourceAltName());
  gen.setUnConditionalActions(context->getUnconditionalActions());
  gen.setConditionalActions(context->getConditionalActionsList());

  // XXX
  // SETTING UNIQUE ID
  string hash = gen.generate()->hashCode();

  gen.setUniqueID(Util::createUniqueID(Util::createFileName(hash)));

  Properties& props = Properties::getSystemProperties();
  string signCerts = props.getValue("SignCerts", "off");

  if (StringUtil::equalsIgnoreCase(signCerts, "on")) {
    ServerContext& serverContext = ServerContext::getServerContext();
    const Certificate& cert = serverContext.getCertificate();

    gen.setIssuer(
      AkentiPrincipal(
        cert->getPrincipal(),
        cert->getGuarantor()
      )
    );

    const PrivateKey& key = serverContext.getPrivateKey();
    string algorithm = key->getAlgorithm();

    if (algorithm == "RSA") {
      gen.sign(key, "RSA-MD5");
    } else if (algorithm == "DSA") {
      gen.sign(key, "DSA-SHA1");
    } else {
      throw Error(
       "PolicyGenerator:generateCapabilitCertificate:unknown algorithm:" + algorithm
      );
    }
  }

  return gen.generate();
}

Logger*
AE::initLogger(const string& id) const {
  Logger* logger = NULL;

  if (!Logger::hasBeenInitialized()) {
    string message;

    logger = Logger::init(id, message);
  } 
  
  return logger;
}

void 
AE::loadResourceMap(const string& fileName) {
  resourceMap.load(fileName);
}

