#include <sstream>
#include "CertificateVerifier.hpp"
#include "AKRuntime.hpp"
#include "CertificateCache.hpp"
#include "CertificateRetrievalFactory.hpp"
#include "AkentiX509CertificateImpl.hpp"
#include "AKCertHolder.hpp"
#include "Util.hpp"
#include "Directory.hpp"
#include "MessageDigest.hpp"
#include "Certificate.hpp"
#include "Signature.hpp"
#include "UtcTime.hpp"
#include "URL.hpp"
#include "Logger.hpp"
#include "Error.hpp"
#include "StringUtil.hpp" 
#include "Provider.hpp"

using namespace Akenti;

typedef AkentiX509CertificateImpl X509Impl;
typedef CertificateRetrievalFactory CRF;
typedef CertificateVerifier CV;

CV::CertificateVerifier(AKRuntime* akRuntime) : akRuntime(akRuntime) {
}

CV::~CertificateVerifier() {
}

bool
CV::verify(const AkentiCertificate& issuerCert, 
           const AkentiCertificate& issuedCert) {
  const X509Impl* impl = (X509Impl*) issuerCert.get();

  if (!impl) {
    throw Error(
      "CertificateVerifier:verify:akenti x509 cert expected"
    );
  }

  const DistinguishedName& issuer = issuedCert->getIssuer().getName();
  const DistinguishedName& subject = impl->getPrincipal().getName();
  
  bLog(
    LOG_L5, 
    "IS_ISSUER_CERT",
    "ISSUER_CERT=being verified " + issuerCert->toLogString() + 
    " ISSUED_CERT=being verified " + issuedCert->toLogString()
  );
  
  if (!Util::compare(issuer, subject)) {
    fLog(LOG_L5, "IS_ISSUER_CERT", "ISSUER_NAME_CHECK=failed");
    return false;
  }

  sLog(LOG_L5, "ISSUER_NAME_CHECK");

  if (!Signature::isSignatureAlgorithmSupported(issuedCert->getSignatureAlg())) {
    fLog(
      LOG_L5, 
      "IS_ISSUER_CERT", 
      "SIGNATURE_CHECK=failed SIG_ALG=" + issuedCert->getSignatureAlg()
    );

    return false;
  }

  if (!issuedCert->verify(impl->getSubjectPublicKey())) {
    fLog(LOG_L5, "IS_ISSUER_CERT", "SIGNATURE_CHECK=failed");
    return false;
  } 

  sLog(LOG_L5, "IS_ISSUER_CERT");
  return true;
}

AKCertHolder
CV::getIssuerCert(const AkentiCertificate& cert, int& level) {
  const AkentiIssuer& issuer = cert->getIssuer();
  const DistinguishedName& issuerDN = issuer.getName();
  
  bLog(LOG_L4, "TRY_POLICY_CERTS");

  vector<AKCertHolder> holders = akRuntime->getCACerts(level, level);

  for (int i = 0; i < holders.size(); ++i) {
    if (verify(holders[i]->getCertificate(), cert)) {
      if (level != 1) {
        level = akRuntime->getCALevel(level - 1);
      }

      sLog(LOG_L4, "TRY_POLICY_CERTS");
      return holders[i];
    }
  }

  fLog(LOG_L4,"TRY_POLICY_CERTS"); 
  bLog(LOG_L4, "TRY_CACHED_ID_CERTS");

  string dn = Provider::getProvider()->toDerString(issuerDN);
  string hash = MessageDigest::digest("MD5", dn);
  int interval = akRuntime->getCacheInterval();
  UtcTime notBefore(-1 * interval);
  CertificateCache& cache = akRuntime->getCertificateCache();

  holders = cache.getAKCertHolders(
     hash, "X509", notBefore 
  );

  for (int i = 0; i < holders.size(); ++i) {
    if (verify(holders[i]->getCertificate(), cert)) {
      sLog(LOG_L4, "TRY_CACHED_ID_CERTS");
      return holders[i];
    }
  }

  fLog(LOG_L4, "TRY_CACHED_ID_CERTS"); 

  // THESE ARGS WILL SATISFY ALL THE PROTOCOLS
  map<string, string> args;

  args["base"] = Util::getHashDN(issuerDN);
  args["type"] = "X509";
  args["hash"] = Util::createFileName(hash);

  UtcTime now, endCacheTime(interval);
  AKCertHolder holder;

  bLog(LOG_L4, "TRY_ID_DIRS", "FILE_HASH_NAME=" + args["hash"]);
  
  vector<URL> issuerURLS;
  bool ok = issuer.parseIdentityDirs(issuerURLS);

  if (!ok) {
    iLog(LOG_L4, "INFO", "PARSING_ISSUER_URLS=failed was not able to parse all issuer urls");
  }

  const vector<URL>& 
    urls = ok && !issuerURLS.size() ? akRuntime->getIdentityURLS(level) : issuerURLS; 
  int size = urls.size();
  std::ostringstream os;

  os << size;

  iLog(LOG_L4, "INFO", "NB_OF_ID_DIRS=" + os.str());

  // HERE WE CHECK ALL URLS UNTIL WE FIND A WINNER
  for (int i = 0; !holder &&  i < size; ++i) {
    bLog(LOG_L5, "FETCH_ID_CERTS", "DIR=" + urls[i].toString());

    vector<Certificate> idCerts = CRF::getCertificates(
      urls[i], args
    );

    int size = idCerts.size();
    std::ostringstream os;

    os << size;
    
    if (!size) {
      fLog(LOG_L5, "FETCH_ID_CERTS");
    } else {
      sLog(LOG_L5, "FETCH_ID_CERTS", "NUM_OF_CERTS_FOUND= " + os.str());
    }

    // WE ADD ALL CERTS WE FETCHED AS LONG AS THEY HAVE NOT EXPIRED 
    // This may change when dealing with the x509 extension that allows
    // signing past the expiration date

    for (int i = 0; i < size; ++i) {
      AkentiCertificate idCert(new X509Impl(idCerts[i]));
      
      if (idCert->checkValidity(now)) {
        AKCertHolder temp = cache.addCertificate(
          idCert, 0, now, endCacheTime
        );

        if (!holder && verify(idCert, cert)) {
          holder = temp; 
        }
      }
    }
  }

  if (!holder) {
    fLog(LOG_L4, "TRY_ID_DIRS");
  } else {
    sLog(LOG_L4, "TRY_ID_DIRS");
  }

  return holder;
}

bool 
CV::verify(AKCertHolder& certHolder) {
  int caLevel = akRuntime->getCALevel(akRuntime->getLevel());
  
  if (certHolder->getLevel() >= caLevel) {
    return true;
  } 
  
  bool ret = verify(certHolder->getCertificate());

  if (ret) {
    certHolder->setLevel(caLevel);
    return  true;
  } else if (!certHolder->getLevel()) {
    certHolder->setLevel(-1);
  }
  
  return false;
}

bool 
CV::verify(const AkentiCertificate& cert) {
  bLog(LOG_L4,"VERIFYING_CERT");
  iLog(LOG_L4,"CERT_INFO", cert->toLogString());
  
  int level = akRuntime->getCALevel(akRuntime->getLevel());
  AKCertHolder holder = getIssuerCert(cert, level);
  
  if (!holder) {
    fLog(LOG_L4, "VERIFYING_CERT");
    return false;
  } else if (holder->getLevel() < 0) {
    throw Error("CertificateVerifier:verify:should never happen (1)");
  } else if (holder->getLevel() >= level) {
    sLog(LOG_L4, "VERIFYING_CERT");
    return true;
  }
  
  /* Certificate in holder hasn't been verified yet, so verify it now*/
  vector<AKCertHolder> chain;
  int depth = 10;
  
  do {
    chain.push_back(holder);

    iLog(
      LOG_L4, 
      "CERT_IN_CHAIN", 
      holder->getCertificate().toLogString()
    );

    holder = getIssuerCert(holder->getCertificate(), level); 
  } while (holder && (holder->getLevel() < level) && --depth);

  if (!depth) {
    fLog(LOG_L4, "VERIFYING_CERT", "DEPTH_CHECK=failed");
    return false;
  }

  level = !holder ? -1 : akRuntime->getCALevel(akRuntime->getLevel());

  for (int i = 0; i < chain.size(); ++i) {
    chain[i]->setLevel(level);
  }
  
  if (level != -1) {
    sLog(LOG_L4, "VERIFYING_CERT");
    return true;
  } 

  fLog(LOG_L4, "VERIFYING_CERT");
  return false;
}

bool 
CV::verify(const AkentiPrincipal& ap, AkentiX509Certificate& cert) {
  bLog(LOG_L4,"VERIFYING_PRINCIPAL", ap.toLogString());

  const DistinguishedName& issuerDN = ap.getName();
  string dn = Provider::getProvider()->toDerString(issuerDN);
  string hash = MessageDigest::digest("MD5", dn);
  int interval = akRuntime->getCacheInterval();
  UtcTime notBefore(-1 * interval);
  CertificateCache& cache = akRuntime->getCertificateCache();

  vector<AKCertHolder> holders = cache.getAKCertHolders(
     hash, "X509", notBefore
  );

  for (int i = 0; i < holders.size(); ++i) {
    if (verify(holders[i])) {
      cert = (AkentiX509Certificate&) holders[i]->getCertificate(); 
      sLog(LOG_L4, "VERIFYING_PRINCIPAL");
      return true; 
    }
  }

  // THESE ARGS WILL SATISFY ALL THE PROTOCOLS
  map<string, string> args;

  args["base"] = Util::getHashDN(issuerDN);
  args["type"] = "X509";
  args["hash"] = Util::createFileName(hash);

  UtcTime now, endCacheTime(interval);
  int level = akRuntime->getCALevel(akRuntime->getLevel());
  const vector<URL>& urls = akRuntime->getIdentityURLS(level);

  int size = urls.size(); 

  for (int i = 0; i < size; ++i) {
    vector<Certificate> idCerts = CRF::getCertificates(
      urls[i], args
    );

    int size = idCerts.size();

    // WE ADD ALL CERTS WE FETCHED AS LONG AS THEY HAVE NOT EXPIRED
    // This may change when dealing with the x509 extension that allows
    // signing past the expiration date

    for (int i = 0; i < size; ++i) {
      AkentiCertificate idCert(new X509Impl(idCerts[i]));

      if (idCert->checkValidity(now)) {
        AKCertHolder temp = cache.addCertificate(
          idCert, 0, now, endCacheTime
        );

        if (verify(idCert)) {
          cert = (AkentiX509Certificate&) idCert;
          sLog(LOG_L4, "VERIFYING_PRINCIPAL");
          return true;
        }
      }
    }
  }
  
  fLog(LOG_L4, "VERIFYING_PRINCIPAL");
  return false; 
}
