#ifndef EXTERNAL_EVALUATOR_HPP
#define EXTERNAL_EVALUATOR_HPP

#include "TokenType.hpp"
#include "ExternalEvaluatorImpl.hpp"

namespace Akenti {

class AKRuntime;

/**
  * Safe wrapper around the ExternalEvaluatorImpl.
  *
  * <PRE>
  * Usage:
  *
  * class MySpecialEvaluator : public ExternalEvaluatorImpl {
  *   public:
  *
  *   bool canHandleSystemAttribute(const string& attribute) {
  *     return attribute == TIME; 
  *   } 
  *
  *   bool evaluateSystemAttribute(TokenType op, const AttributeInfo& attrInfo) { 
  *     string attribute = attrInfo.getAttribute();
  *     string value = attrInfo.getValue();
  *     bool ret = false;
  *     
  *     switch(op) {
  *       case LT:
  *       ...
  *       case GT:
  *       ....
  *     }
  *
  *     return ret;
  *   }
  * };
  *
  * // INSTALLING AN EVALUATOR.
  *
  * Evaluator mySpecialEvaluator(new MySpecialEvalutor);
  * AkentiEngine eng;
  *
  * eng.setExternalEvaluator(mySpecialEvaluator);
  * eng.checkAccess(...);
  * </PRE>
  *
  * 
  * @see ExternalEvaluatorImpl
  * @version 1.1 03/04/2003
  * @author Abdelilah Essiari
  */

class ExternalEvaluator {

public:
  
  /**
    * Constructor.
    */
  ExternalEvaluator() : impl(new ExternalEvaluatorImpl) {
  }

  ExternalEvaluator(ExternalEvaluatorImpl* impl) : impl(impl) {
    if (!impl) {
      throw Error("ExternalEvaluator:ExternalEvaluator:illegal argument");
    }
  }

  /**
    * Destructor.
    */

  ~ExternalEvaluator() {
  }

  /**
    * Returns <i>true</i> if this evaluator knows how to 
    * handle the <i>SYSTEM attribute</i> specified in the
    * argument.
    *
    * @see AttributeInfo 
    */
  bool canHandleSystemAttribute(const string& attribute) {
    return impl->canHandleSystemAttribute(attribute);
  }

  /**
    * Returns <i>true</i> if (attrInfo.getAttribute() op  attrInfo.getValue())
    *
    * Note: the engine will call this method only if 
    *      canHandleSystemAttribute<attrInfo.getAttribute() returned true.
    */
  bool evaluateSystemAttribute(TokenType op, const AttributeInfo& attrInfo) { 
    return impl->evaluateSystemAttribute(op, attrInfo);
  }

  /**
    * Returns <i>true</i> if this evaluator knows how to 
    * handle the <i>EXT_AUTH attribute</i> specified in the
    * argument.
    *
    * @see AttributeInfo 
    */
  bool canHandleExternalAttribute(const string& attribute) {
    return impl->canHandleExternalAttribute(attribute);
  }

  /**
    * Returns <i>true</i> if (attrInfo.getAttribute() op  attrInfo.getValue())
    *
    * Note: the engine will call this method only if 
    *        canHandleHandleExternalAttribute<attrInfo.getAttribute()> returned true.
    */
  bool evaluateExternalAttribute(TokenType op, const AttributeInfo& attrInfo) {
    return impl->evaluateExternalAttribute(op, attrInfo);
  }

private:

  void setAKRuntime(AKRuntime* akRuntime) {
    impl->setAKRuntime(akRuntime);
  }

  friend class AkentiEngine;

  ref_ptr<ExternalEvaluatorImpl> impl;
};

}

#endif

