#include <sstream>
#include "PolicyCollector.hpp"
#include "AKRuntime.hpp"
#include "AkentiCertificate.hpp"
#include "CertificateVerifier.hpp"
#include "Certificate.hpp"
#include "UseCondIssuerGroup.hpp"
#include "UtcTime.hpp"
#include "Properties.hpp"
#include "StringTokenizer.hpp"
#include "File.hpp"
#include "Exception.hpp"
#include "Logger.hpp"
#include "AkentiConstants.hpp"

using namespace Akenti;

namespace Akenti {
  static const string ROOT_RESOURCE_NAME = "RootResourceName";
};

typedef PolicyCollector PC;

PC::PolicyCollector(AKRuntime* akRuntime) : akRuntime(akRuntime) {
}

PC::~PolicyCollector() {
}
  
bool
PC::collectPolicies() {
  Properties& props = Properties::getSystemProperties();
  vector<string> resources;
  
  bLog(LOG_MON,"COLLECT_ROOT_POLICY");
  
  if (!parseResource(akRuntime->getResource(), resources)) {
    fLog(LOG_MON,"COLLECT_ROOT_POLICY","PARSE_RESOURCE_NAME=FAIL");
    message = AkentiMessage(401);
    return false;
  }
  
  string base = props.getValue(std::make_pair(ROOT_RESOURCE_NAME, resources[0]));

  if (!base.length()) {
    fLog(LOG_MON,"COLLECT_ROOT_POLICY","BASE_NAME=not specified");
    message = AkentiMessage(402); 
    return false;
  }

  string encType = props.getValue("AK_CERT_ENC_TYPE", AK_CERT_ENC_TYPE);
  string policyFileName = props.getValue("PolicyFileName", PolicyFileName);
  string temp = base + '/' + resources[0] + '/' + policyFileName;
  File file(temp);

  if (!file.isAbsolute()) {
    fLog(
      LOG_MON, "COLLECT_ROOT_POLICY", 
      "BASE_NAME_CHECK=" + 
      file.toLogString() +
      "is not absolute"
    );

    message = AkentiMessage(403); 
    return false;
  } else if (!file.exists()) {
    fLog(LOG_MON, "COLLECT_ROOT_POLICY", "POL_FILE_DOES_NOT_EXIST=" + file.toLogString());
    message = AkentiMessage(404); 
    return false;
  } 
  
  string mesg;
  PolicyCertificate policyCert = readPolicyFile(encType, temp, mesg);

  if (!policyCert) {
    fLog(LOG_MON, "COLLECT_ROOT_POLICY", mesg);
    message = AkentiMessage(405);
    return false;
  } else if (policyCert->getResource() != resources[0]) {
    fLog(LOG_MON, "COLLECT_ROOT_POLICY", "WRONG_RESOURCE=" + policyCert->getResource());
    message = AkentiMessage(406,  " " + resources[0] + " != " + policyCert->getResource()); 
    return false;
  } else if (!policyCert->containsCertificateAuthorityInfos()) {
    fLog(LOG_MON, "COLLECT_ROOT_POLICY", "ROOT_CAS_CHECK=FAIL");
    message = AkentiMessage(407); 
    return false;
  } 
  
  int size = resources.size();
  
  if (size == 1) {
    akRuntime->addPolicy(new PolicyImpl(true, policyCert));
  } else if (size == 2) {
    file = File(base + '/' + resources[1]);
    akRuntime->addPolicy(new PolicyImpl(!file.isDirectory(), policyCert));
  } else {
    akRuntime->addPolicy(new PolicyImpl(false, policyCert));
  }

  sLog(LOG_MON,"COLLECT_ROOT_POLICY");

  for (int i = 1; i < size; ++i) {
    temp = base + '/' + resources[i];
    file = File(temp);

    if (file.isDirectory()) {
      temp += '/' + policyFileName;
      file = File(temp);

      if (file.exists()) {
        bLog(LOG_MON,"COLLECT_POLICY");
        policyCert = readPolicyFile(encType, temp, mesg);

        if (!policyCert) {
          fLog(LOG_MON, "COLLECT_POLICY", mesg);
          message = AkentiMessage(430, " for " + resources[i]); 
          return false;
        } else if (policyCert->getResource() != resources[i]) {
          fLog(LOG_MON, "COLLECT_POLICY", "WRONG_RESOURCE=" + policyCert->getResource());
          message = AkentiMessage(431, resources[i] + " != " + policyCert->getResource()); 
          return false;
        }

        if (i == size - 1) {
          akRuntime->addPolicy(new PolicyImpl(true, policyCert));
        } else if (i == size - 2) {
          file = File(base + '/' + resources[size - 1]);
          akRuntime->addPolicy(new PolicyImpl(!file.isDirectory(), policyCert));
        } else {
          akRuntime->addPolicy(new PolicyImpl(false, policyCert));
        }

        sLog(LOG_MON,"COLLECT_POLICY");
      }
    } else if (i != size - 1) {
      bLog(LOG_MON,"COLLECT_POLICY");
      fLog(LOG_MON, "COLLECT_POLICY", "RESOURCE_TREE_CHECK=FAIL");
      message = AkentiMessage(432); 
      return false;
    }
  }   // end loop collecting a policy file for each resource in the tree

  message = AkentiMessage(101);
  return true;
}

bool 
PC::verifyPolicies() {
  int size = akRuntime->numOfPolicies();

  if (!size) {
    throw Error("PolicyCollector:verifyPolicies:illegal operation");
  }
  
  bLog(LOG_MON,"CHECK_POLICIES");
  
  bool ret = true;
  int i = -1;
  
  while (ret && ++i < size) {
    akRuntime->setLevel(i + 1);
    ret = verifyPolicy(akRuntime->getPolicy(i));
  }
  
  if (ret) {
    sLog(LOG_MON,"CHECK_POLICIES");
    message = AkentiMessage(102);
  } else {
    fLog(LOG_MON,"CHECK_POLICIES");
    message = AkentiMessage(450); 
  }

  return ret;
}

const AkentiMessage&
PC::getAkentiMessage() const {
  return message;
}

bool
PC::verifyPolicy(const Policy& policy) {
  bLog(LOG_L3, "CHECK_POLICY", "RESOURCE=" + policy->getResource());

  PolicyCertificate cert = policy->getPolicyCertificate();

  if (!cert->checkValidity()) {
    fLog(LOG_L3, "CHECK_POLICY", "TIME_CHECK=failed");
    return false;
  }

  const AkentiIssuer& issuer = policy->getIssuer();
  bool nameVerified = false;

  for (int i = 0; !nameVerified && i < policy->numOfUseCondIssuerGroups(); ++i) {
    const UseCondIssuerGroup& group = policy->getUseCondIssuerGroup(i);

    nameVerified = group.containsUCIssuer(issuer.getIssuer());
  }

  if (!nameVerified) {
    fLog(LOG_L3, "CHECK_POLICY", "GROUP_CHECK=issuer must be in one of the UCGroups");
    return false;
  }

  if (akRuntime->getCertificateVerifier().verify(cert)) {
    sLog(LOG_L3, "CHECK_POLICY");  
    return true;
  }

  fLog(LOG_L3, "CHECK_POLICY", "TRUST_CHECK=FAIL");  
  return false;
}

bool 
PC::parseResource(const string& resource, vector<string>& resources) const {
  StringTokenizer st(resource, "/ ", true);
  bool ret = false;
  string name;

  while (st.hasMoreTokens()) {
    string temp = st.nextToken();

    if (temp == "/" || temp == " ") {
      break;
    }

    resources.push_back(name += temp);
    ret = true;

    if (!st.hasMoreTokens()) {
      break;
    }

    if (st.nextToken() == " ") {
      ret = false;
      break;
    }

    name += "/";
    ret = false;
  }

  return ret;
}

PolicyCertificate
PC::readPolicyFile(const string& encType, const string& fileName, string& mesg) const {
  PolicyCertificate policyCert;

  try {
    AkentiCertificate cert = AkentiCertificate::initFromFile(encType, fileName);
    
    if (!cert.instanceOf<PolicyCertificate>()) {
      mesg = "WONG_TYPE=" + cert->getType();
    } else {
      policyCert = (PolicyCertificate&) cert;
      mesg = "OK";
    }
  } catch(Exception& ex) { 
    mesg = "READING_POLICY=failed";
  }

  return policyCert;;
}
