#include <sstream>
#include "UseCondCollector.hpp"
#include "Policy.hpp"
#include "AKRuntime.hpp"
#include "CertificateRetrievalFactory.hpp"
#include "CertificateCache.hpp"
#include "UseCondIssuerGroup.hpp"
#include "CertificateVerifier.hpp"
#include "UtcTime.hpp"
#include "Policy.hpp"
#include "URL.hpp"
#include "Util.hpp"
#include "Logger.hpp"

using std::ostringstream;

using namespace Akenti;

typedef CertificateRetrievalFactory CRF;

typedef UseCondCollector UC;

UC::UseCondCollector(AKRuntime* akRuntime) 
  : akRuntime(akRuntime) {
}

UC::~UseCondCollector() {
}
  
const AkentiMessage&
UC::getAkentiMessage() const {
  return message;
}

vector<AKCertHolder>
UC::getUCCertHolders(Policy& policy) {
  string hash = policy->hashCode();
  int cacheInterval = akRuntime->getCacheInterval();
  UtcTime notBefore(-1 * cacheInterval);

  bLog(LOG_MON,"TRY_CACHED_UC_CERTS");

  CertificateCache& certCache = akRuntime->getCertificateCache();
  vector<AKCertHolder> ucCertHolders = certCache.getAKCertHolders(
    hash, "UseCondition", notBefore
  );

  if (ucCertHolders.size()) {
    sLog(LOG_MON,"TRY_CACHED_UC_CERTS");
    return ucCertHolders;
  }

  fLog(LOG_MON, "TRY_CACHED_UC_CERTS");

  map<string, string> args;
  int numOfGroups = policy->numOfUseCondIssuerGroups();
  ostringstream os;

  args["type"] = "UseCondition";
  args["hash"] = Util::createFileName(hash);
  os << numOfGroups;

  bLog(
    LOG_MON, 
    "TRY_UC_DIRS", 
    "FILE_HASH_NAME=" + args["hash"] + " NB_OF_UC_GROUPS=" + os.str()
  );

  vector<AkentiCertificate> ucCerts, trustedUCCerts;
  int groupIndex = 0;
  bool foundValidUC;
  UtcTime now;
  
  CertificateVerifier& certVerifier = akRuntime->getCertificateVerifier();

  do {
    const UseCondIssuerGroup& group = policy->getUseCondIssuerGroup(groupIndex);
    const vector<Directory>& dirs = group.getDirectories();

    foundValidUC = false;

    for (int j = 0; !foundValidUC && j < dirs.size(); ++j) {
      bLog(LOG_MON, "FETCH_UC_CERTS", "DIR=" + dirs[j].toString());

      vector<AkentiCertificate> certs = CRF::getAkentiCertificates(
        dirs[j], args
      );

      int size = certs.size();
      ostringstream os;

      os << size;
          
      if (!size) {
        fLog(LOG_MON, "FETCH_UC_CERTS");
      } else {
        sLog(LOG_MON, "FETCH_UC_CERTS", "NB_OF_UC_CERTS_FOUND=" + os.str());
      }

      int k;

      // We must find trusted uc cert for each group
      for (k = 0; !foundValidUC && k < size; ++k) {
        UseCondCertificate cert = (UseCondCertificate&) certs[k];
        const AkentiIssuer& issuer = cert->getIssuer();

        bLog( LOG_L3, "CHECK_UC_CERT", cert->toLogString());

        if (cert->hashCode() != hash) {
          fLog(LOG_L3, "CHECK_UC_CERT", "HASH_CHECK=failed");
        } else if (!cert->checkValidity(now)) {
          fLog(LOG_L3, "CHECK_UC_CERT", "TIME_CHECK=failed");
        } else if (!group.containsUCIssuer(issuer.getIssuer())) {
          fLog(LOG_L3, "CHECK_UC_CERT", "GROUP_CHECK=failed");
        } else if (!cert->getCondition().isValid()) {
          fLog(LOG_L3,"CHECK_UC_CERT", "COND_CHECK_VALIDITY=FAIL");
        } else if (!certVerifier.verify(cert)) {
          fLog(LOG_L3, "CHECK_UC_CERT", "TRUST_CHECK=FAIL");
        } else {
          foundValidUC = true;
          trustedUCCerts.push_back(cert);
          sLog(LOG_L3, "CHECK_UC_CERT");
        }
      }

      // We will check the trust later 
      for (int l = k; l < certs.size(); ++l) {
        UseCondCertificate cert = (UseCondCertificate&) certs[l];
        const AkentiIssuer& issuer = cert->getIssuer();

        bLog (LOG_L3, "CHECK_UC_CERT", cert->toLogString());

        if (cert->hashCode() != hash) {
          fLog(LOG_L3, "CHECK_UC_CERT", "HASH_CHECK=failed");
        } else if (!cert->checkValidity(now)) {
          fLog(LOG_L3, "CHECK_UC_CERT", "TIME_CHECK=failed");
        } else if (!group.containsUCIssuer(issuer.getIssuer())) {
          fLog(LOG_L3, "CHECK_UC_CERT", "GROUP_CHECK=failed");
        } else if (!cert->getCondition().isValid()) {
          fLog(LOG_L3,"CHECK_UC_CERT", "COND_CHECK_VALIDITY=FAIL");
        } else {
          ucCerts.push_back(cert);
          sLog(LOG_L3, "CHECK_UC_CERT");
        }
      }
    }
  } while (foundValidUC && ++groupIndex < numOfGroups);

  if (foundValidUC) {
    UtcTime endCacheTime(cacheInterval);
    int level = akRuntime->getLevel();

    for (int i = 0; i < trustedUCCerts.size(); ++i) {
      ucCertHolders.push_back(
        certCache.addCertificate(trustedUCCerts[i], level, now, endCacheTime) 
      ); 
    }

    for (int i = 0; i < ucCerts.size(); ++i) {
      ucCertHolders.push_back(
        certCache.addCertificate(ucCerts[i], 0, now, endCacheTime) 
      ); 
    }
  }

  if (ucCertHolders.size()) {
    sLog(LOG_MON,"TRY_UC_DIRS");
  } else {
    ostringstream os;

    os << groupIndex;
    fLog(LOG_MON, "TRY_UC_DIRS", "NO_VALID_UC_FOR_GROUP=" + os.str());
  }

  return ucCertHolders;
}

bool
UC::collectUseCondCertificates(Policy& policy) {
  bLog(LOG_MON,"COLLECT_UC_CERTS", "RESOURCE=" + policy->getResource());

  vector<AKCertHolder> ucCertHolders = getUCCertHolders(policy);
  int size = ucCertHolders.size();

  if (!size) {
    fLog(
      LOG_MON,"COLLECT_UC_CERTS", 
      "NOT_FOUND_UC=did not find any UCCerts"
    );

    message = AkentiMessage(
      501, 
      " for " + policy->getResource()
    );

    return false;
  }

  ostringstream os1;
  
  os1 << size;

  iLog(LOG_MON, "INFO", "NB_OF_UC_CERTS_FOUND=" + os1.str());

  CertificateVerifier& certVerifier = akRuntime->getCertificateVerifier();
  int count = 0;

  for (int i = 0; i < ucCertHolders.size(); ++i) {
    UseCondCertificate ucCert = (UseCondCertificate&) ucCertHolders[i]->getCertificate();

    if (!ucCert->isScopeLocal() || policy->isLeaf()) {
      if (certVerifier.verify(ucCertHolders[i])) {
        policy->addUCCertificate(ucCert);
        ++count;
      }
    }
  }

  ostringstream os2;
  
  os2 << count;

  iLog(LOG_MON,"INFO", "NB_OF_ACCEPTABLE_UC_CERTS=" + os2.str());
  sLog(LOG_MON,"COLLECT_UC_CERTS");

  message = AkentiMessage(
    103, 
    " for " + policy->getResource()
  );

  return true;
}

bool
UC::collectUseCondCertificates() {
  bool ret = true;
  int size = akRuntime->numOfPolicies();

  for (int i = 0; ret && i < size; ++i) {
    akRuntime->setLevel(i + 1);
    ret = collectUseCondCertificates(akRuntime->getPolicy(i));
  }
  
  return ret;
}
