#include <sstream>
#include "CheckPolCertCollector.hpp"
#include "AKRuntime.hpp"
#include "AkentiCertificate.hpp"
#include "DummyCertificate.hpp"
#include "CertificateVerifier.hpp"
#include "Certificate.hpp"
#include "UseCondIssuerGroup.hpp"
#include "UtcTime.hpp"
#include "Properties.hpp"
#include "StringTokenizer.hpp"
#include "File.hpp"
#include "Exception.hpp"
#include "Logger.hpp"
#include "CertInfoCollector.hpp"
#include "AkentiConstants.hpp"

// for debugging statements only
#include <iostream> 
using namespace std; 

using namespace Akenti;

namespace Akenti {
  static const string ROOT_RESOURCE_NAME = "RootResourceName";
};

typedef CheckPolCertCollector PCC;


PCC::CheckPolCertCollector(AKRuntime* akRuntime) : akRuntime(akRuntime) {
}

PCC::~CheckPolCertCollector() {
}

/* In collectPolicies: if its a root policy - return after a single 
 * error - in all other cases, continue as far as you can 
 */
  
bool
PCC::collectPolicies() {
  Properties& props = Properties::getSystemProperties();
  vector<string> resources;
  DummyCertificate dummyCert; // dummy - should handle this differently

  CertInfoCollector &cinfoCol = akRuntime->getCertInfoCollector ();
  
  bLog(LOG_MON,"COLLECT_ROOT_POLICY");
  
  if (!parseResource(akRuntime->getResource(), resources)) {
    fLog(LOG_MON,"COLLECT_ROOT_POLICY","PARSE_RESOURCE_NAME=FAIL");
    CertInfo cinfo (false, dummyCert, (AkentiMessage(401))); 
    cinfoCol.addCertInfo(cinfo);
    cinfoCol.setStatus (false);
    return false;
  }
  
  string base = props.getValue(std::make_pair(ROOT_RESOURCE_NAME, resources[0]));

  if (!base.length()) {

    fLog(LOG_MON,"COLLECT_ROOT_POLICY","BASE_NAME=not specified");
    CertInfo cinfo (false, dummyCert, (AkentiMessage (402))); 
    cinfoCol.addCertInfo(cinfo);
    cinfoCol.setStatus (false);
    return false;
  }

  string encType = props.getValue("AK_CERT_ENC_TYPE", AK_CERT_ENC_TYPE);
  string policyFileName = props.getValue("PolicyFileName", PolicyFileName);
  string temp = base + '/' + resources[0] + '/' + policyFileName;
  File file(temp);

  if (!file.isAbsolute()) {

    fLog(
      LOG_MON, "COLLECT_ROOT_POLICY", 
      "BASE_NAME_CHECK=" + 
      file.toLogString() +
      "is not absolute"
    );

    CertInfo cinfo (false, dummyCert, (AkentiMessage (403))); 
    cinfoCol.addCertInfo(cinfo);
    cinfoCol.setStatus (false);
    return false;
  } else if (!file.exists()) {
    fLog(LOG_MON, "COLLECT_ROOT_POLICY", "POL_FILE_DOES_NOT_EXIST=" + file.toLogString());
    CertInfo cinfo (false, dummyCert, (AkentiMessage (404)));
    cinfoCol.addCertInfo(cinfo);
    cinfoCol.setStatus (false);
    return false;
  } 
  

  string mesg; 
  PolicyCertificate policyCert = readPolicyFile(encType, temp, mesg);

  if (!policyCert || (mesg == "READING_POLICY=failed")) {
    fLog(LOG_MON, "COLLECT_ROOT_POLICY", mesg);
    CertInfo cinfo (false, dummyCert, (AkentiMessage (405, " or error in policy file " ))); 
    cinfoCol.addCertInfo(cinfo);
    cinfoCol.setStatus (false);
    return false;
  } else if (policyCert->getResource() != resources[0]) {
    fLog(LOG_MON, "COLLECT_ROOT_POLICY", "WRONG_RESOURCE=" + policyCert->getResource());
    int index = cinfoCol.lookup (policyCert->hashCode(), "Policy");
    if (index < 0) { // doesnt exist, add it
        CertInfo cinfo (false, policyCert, (AkentiMessage (406)),
               policyCert->hashCode(), "Policy", policyCert->getResource()); 
        cinfoCol.addCertInfo (cinfo);
    } 
    else {
        cinfoCol.addMessage (index, AkentiMessage (406)); 
    }
    cinfoCol.setStatus (false);
    return false;
  } else if (!policyCert->containsCertificateAuthorityInfos()) {
    fLog(LOG_MON, "COLLECT_ROOT_POLICY", "ROOT_CAS_CHECK=FAIL");
    int index = cinfoCol.lookup (policyCert->hashCode(), "Policy");
    if (index < 0) { // doesnt exist, add it
        CertInfo cinfo (false, policyCert, (AkentiMessage (407)),
               policyCert->hashCode(), "Policy", policyCert->getResource()); 
        cinfoCol.addCertInfo (cinfo);
    } 
    else {
        cinfoCol.addMessage (index, AkentiMessage (407)); 
    }
    cinfoCol.setStatus (false);
    return false;
  } 

  // If root policies failed - at this point you should exit
  if (!cinfoCol.getStatus()) return false;
  
  int size = resources.size();

  if (cinfoCol.getStatus()) {
          bool leaf = false;

          if (size == 1) {
            leaf = true;
          } else if (size == 2) {
            file = File(base + '/' + resources[1]);
            leaf = !file.isDirectory();
          }

          int index = cinfoCol.lookup (policyCert->hashCode(), "Policy");

          if (index < 0) { // doesnt exist, add it
                CertInfo cinfo (true, policyCert, (AkentiMessage (101)),
                  policyCert->hashCode(), "Policy", policyCert->getResource()); 
                cinfo.setLeaf (leaf);
                cinfoCol.addCertInfo (cinfo);
          } 
          else {
              cinfoCol.addMessage (index, AkentiMessage (101)); 
          }

          akRuntime->addPolicy(new PolicyImpl(leaf, policyCert));

     sLog(LOG_MON,"COLLECT_ROOT_POLICY");
  }

  for (int i = 1; i < size; ++i) {
    temp = base + '/' + resources[i];
    file = File(temp);
    if (file.isDirectory()) {
      temp += '/' + policyFileName;
      file = File(temp);

      if (file.exists()) {
        bLog(LOG_MON,"COLLECT_POLICY");
        policyCert = readPolicyFile(encType, temp, mesg);

        if (!policyCert || (mesg == "READING_POLICY=failed")) {
          fLog(LOG_MON, "COLLECT_POLICY", mesg);
          int index = cinfoCol.lookup (policyCert->hashCode(), "Policy");
          if (index < 0) { // doesnt exist, add it
                CertInfo cinfo (false, policyCert, 
                 (AkentiMessage (430, " for " + resources[i])),
                 policyCert->hashCode(), "Policy", policyCert->getResource(),
                 true); 
                cinfoCol.addCertInfo (cinfo);
          } 
          else {
              cinfoCol.addMessage (index, AkentiMessage (430, " for " + resources[i]));
              cinfoCol.setUsed (index, false);
              cinfoCol.setError (index, true);
          }
          cinfoCol.setStatus (false);
        } else if (policyCert->getResource() != resources[i]) {
          fLog(LOG_MON, "COLLECT_POLICY", "WRONG_RESOURCE=" + policyCert->getResource());
          int index = cinfoCol.lookup (policyCert->hashCode(), "Policy");
          if (index < 0) { // doesnt exist, add it
              CertInfo cinfo (false, policyCert, 
                (AkentiMessage (431, " " + policyCert->getResource())),
                policyCert->hashCode(), "Policy", policyCert->getResource(),
                true); 
                cinfoCol.addCertInfo (cinfo);
          } 
          else {
              cinfoCol.addMessage (index, AkentiMessage (431, " " + policyCert->getResource()));
              cinfoCol.setUsed (index, false);
              cinfoCol.setError (index, true);
          }
          cinfoCol.setStatus (false);
        }
        else {
          bool leaf = false;

          if (i == size - 1) {
            leaf = true;
          } else if (i == size - 2) {
            file = File(base + '/' + resources[size - 1]);
            leaf = !file.isDirectory();
          }

          int index = cinfoCol.lookup (policyCert->hashCode(), "Policy");
          if (index < 0) { // doesnt exist, add it
                CertInfo cinfo (true, policyCert, (AkentiMessage (101)),
                  policyCert->hashCode(), "Policy", policyCert->getResource()); 
                cinfoCol.addCertInfo (cinfo);
                cinfo.setLeaf (leaf);
          } 
          else {
              cinfoCol.addMessage (index, AkentiMessage (101)); 
          }
        akRuntime->addPolicy(new PolicyImpl(leaf, policyCert));
        sLog(LOG_MON,"COLLECT_POLICY");
        }
      }

    } else if (i != size - 1) {
      bLog(LOG_MON,"COLLECT_POLICY");
      fLog(LOG_MON, "COLLECT_POLICY", "RESOURCE_TREE_CHECK=FAIL");
      CertInfo cinfo (false, dummyCert, (AkentiMessage (432))); 

      cinfo.setResource(resources[i]);
      cinfoCol.addCertInfo (cinfo);
      cinfoCol.setStatus(false);
      break;

/* XXX SUSHI I HAVE CHANGED THIS 
      int index = cinfoCol.lookup (policyCert->hashCode(), "Policy");
      if (index < 0) { // doesnt exist, add it
        CertInfo cinfo (false, policyCert, (AkentiMessage (432)),
            policyCert->hashCode(), "Policy", policyCert->getResource(),
            true); 
        cinfoCol.addCertInfo (cinfo);
      } 
      else {
          cinfoCol.addMessage (index, AkentiMessage (432)); 
          cinfoCol.setUsed (index, false);
          cinfoCol.setError (index, true);
      }
      cinfoCol.setStatus(false);
*/
    }
  }   // end loop collecting a policy file for each resource in the tree

  message = AkentiMessage(101);
  return true; // Necessary for Test case -doesnt really matter otherwise 
}

bool 
PCC::verifyPolicies() {

  CertInfoCollector &cinfoCol = akRuntime->getCertInfoCollector();

  bool ret = true;
  int i = -1;

  int size = cinfoCol.getCertInfoVec().size();

  int j = 1; // used to set akRuntime levels

  for (int i = 0; i < size; ++i) {
      CertInfo& cinfo = cinfoCol.getCertInfo(i);
      if (cinfo.getType() == "Policy") {
         AkentiCertificate akcert = cinfo.getCertificate();
         PolicyCertificate pcert = (PolicyCertificate &)akcert;
         Policy policy  = new PolicyImpl(cinfo.isLeaf(), pcert);
         if (cinfo.getUsed()) {
           akRuntime->setLevel (j++);
            verifyPolicy(policy);
         }
      }
  }

  return true; // Necessary for Test case -doesnt really matter otherwise 
}

const AkentiMessage&
PCC::getAkentiMessage() const {
  return message;
}

bool
PCC::verifyPolicy(const Policy& policy) {
  CertInfoCollector &cinfoCol = akRuntime->getCertInfoCollector();

  bLog(LOG_L3, "CHECK_POLICY", "RESOURCE=" + policy->getResource());
  PolicyCertificate pcert = policy->getPolicyCertificate();

  if (!pcert->checkValidity()) {
    fLog(LOG_L3, "CHECK_POLICY", "TIME_CHECK=failed");
     int index = cinfoCol.lookup (pcert->hashCode(), "Policy");
     if (index < 0) { // doesnt exist, add it
       CertInfo cinfo (false, pcert, 
          (AkentiMessage (480, " Validation failed")),
          pcert->hashCode(), "Policy", pcert->getResource(), true); 
       cinfoCol.addCertInfo (cinfo);
     } 
     else {
         cinfoCol.addMessage (index, AkentiMessage (480, " Validation failed")); 
         cinfoCol.setUsed (index, false);
         cinfoCol.setError (index, true);
     }
      cinfoCol.setStatus (false);
  }

  const AkentiIssuer& issuer = policy->getIssuer();
  bool nameVerified = false;

  for (int i = 0; !nameVerified && i < policy->numOfUseCondIssuerGroups(); ++i) {
    const UseCondIssuerGroup& group = policy->getUseCondIssuerGroup(i);

    nameVerified = group.containsUCIssuer(issuer.getIssuer());
  }

  if (!nameVerified) {
    fLog(LOG_L3, "CHECK_POLICY", "GROUP_CHECK=issuer must be in one of the UCGroups");
    int index = cinfoCol.lookup (pcert->hashCode(), "Policy");
    if (index < 0) { // doesnt exist, add it
      CertInfo cinfo (false, pcert, 
          (AkentiMessage (480, " Group check failed")),
          pcert->hashCode(), "Policy", pcert->getResource (), true); 
      cinfoCol.addCertInfo (cinfo);
    } 
    else {
        cinfoCol.addMessage (index, AkentiMessage (480, " Group check failed")); 
        cinfoCol.setUsed (index, false);
        cinfoCol.setError (index, true);
    }
    cinfoCol.setStatus (false);
  }

  if (akRuntime->getCertificateVerifier().verify(pcert)) {
    sLog(LOG_L3, "CHECK_POLICY");  
    int index = cinfoCol.lookup (pcert->hashCode(), "Policy");
    if (index < 0) { // doesnt exist, add it
       CertInfo cinfo (true, pcert, (AkentiMessage (102)),
                 pcert->hashCode(), "Policy", pcert->getResource()); 
       cinfoCol.addCertInfo (cinfo);
    } 
    else {
        cinfoCol.addMessage (index, AkentiMessage (102)); 
    }
    return true;
  }

  fLog(LOG_L3, "CHECK_POLICY", "TRUST_CHECK=FAIL");  
}

bool 
PCC::parseResource(const string& resource, vector<string>& resources) const {
  StringTokenizer st(resource, "/ ", true);
  bool ret = false;
  string name;

  while (st.hasMoreTokens()) {
    string temp = st.nextToken();

    if (temp == "/" || temp == " ") {
      break;
    }

    resources.push_back(name += temp);
    ret = true;

    if (!st.hasMoreTokens()) {
      break;
    }

    if (st.nextToken() == " ") {
      ret = false;
      break;
    }

    name += "/";
    ret = false;
  }

  return ret;
}

PolicyCertificate
PCC::readPolicyFile(const string& encType, const string& fileName, string& mesg) const {

  PolicyCertificate policyCert;
  bool failed = false;

  try {
    AkentiCertificate cert = AkentiCertificate::initFromFile(encType, fileName);
    
    if (!cert.instanceOf<PolicyCertificate>()) {
      mesg = "WRONG_TYPE=" + cert->getType();
    } else {
      policyCert = (PolicyCertificate&) cert;
      mesg = "OK";
    }
  } catch(Exception& ex) { 
    failed = true;
  }

  if (failed)
    mesg = "READING_POLICY=failed";

  return policyCert;
}
