#include <iostream>
#include <fstream>
#include <sstream>
#include <map>
#include "Xerces-Includes.hpp"
#include "ProtocolHandler.hpp"
#include "AkentiRequests.hpp"
#include "AkentiResponses.hpp"
#include "AkentiRequest.hpp"
#include "AkentiResponse.hpp"
#include "AkentiEngine.hpp"
#include "SoapHttpRequest.hpp"
#include "SoapHttpResponse.hpp"
#include "AkentiRequests.hpp"
#include "AkentiResponse.hpp"
#include "AkentiResponses.hpp"
#include "CheckAccessRequest.hpp"
#include "CheckAccessResponse.hpp"
#include "CheckPolicyRequest.hpp"
#include "CheckPolicyResponse.hpp"
#include "PolicyChecker.hpp"
#include "ResourceBrowseRequest.hpp"
#include "ResourceBrowseResponse.hpp"
#include "CreateResourceRequest.hpp"
#include "CreateResourceResponse.hpp"
#include "UploadPolicyCertificateRequest.hpp"
#include "UploadPolicyCertificateResponse.hpp"
#include "DeletePolicyCertificateRequest.hpp"
#include "DeletePolicyCertificateResponse.hpp"
#include "GetResourceInfoRequest.hpp"
#include "GetResourceInfoResponse.hpp"
#include "AkentiMessage.hpp"
#include "CapabilityCertificate.hpp"
#include "PolicyContext.hpp"
#include "AkentiEngine.hpp"
#include "Properties.hpp"
#include "File.hpp"
#include "StringTokenizer.hpp"
#include "StringUtil.hpp"
#include "SSLSocket.hpp"
#include "Socket.hpp"
#include "XMLUtil.hpp"
#include "AkentiConstants.hpp"
#include "Exception.hpp"
#include "Error.hpp"
#include "ParseResourceUtil.hpp"

using namespace std;

#define DEBUG 0

namespace Akenti {
  
  typedef ProtocolHandler PH;

  // BEGIN REMOTE
  static Socket
  createSocket() {
    Properties& props = Properties::getSystemProperties();
    string sslMode = props.getValue("SslMode");
    string host = props.getValue("Host", "localhost");
    bool sslOn = StringUtil::equalsIgnoreCase(sslMode, "on");
    string portString = props.getValue("Port", sslOn? "8443" : "8444");
    istringstream is(portString);
    int port;

    if (!(is >> port)) {
      throw Exception("Invalid port " + portString);
    }

    return sslOn? SSLSocket(host, port) : Socket(host, port);
  }


  static void 
  doRemote(const AkentiRequests& requests, AkentiResponses& responses) {
    if (DEBUG) {
      cerr << "Sending Requests..." << endl << endl << endl;
      cerr << requests.toString() << endl;
    }

    SoapHttpRequest soapRequest;

    soapRequest.setParams(requests.toXML());

    Socket sock = createSocket();

    sock.connect();

    SoapHttpResponse soapResponse;

    soapRequest.toStream(sock.getOutputStream());
    soapResponse.fromStream(sock.getInputStream());

    if (soapResponse.getFaultCode() != NONE) {
      if (DEBUG) {
	cerr << "SOAP FAULT:" << soapResponse.getFaultMessage() << endl;
      }

      throw Exception(soapResponse.getFaultMessage());
    }

    responses.fromXML(soapResponse.getReturnValue());

    if (DEBUG) {
      cerr << "Receiving Responses..." << endl << endl << endl;
      cerr << responses.toString() << endl;
    }
  }

  static void 
  doRemote(const AkentiRequest& req, AkentiResponse& resp) {
    AkentiRequests requests;
    AkentiResponses responses;

    requests.addRequest(req);
    doRemote(requests, responses);
    resp = responses.getResponseAt(0);
  }

  // END REMOTE

  // BEGIN LOCAL

  static void
  handle(const CheckAccessRequest& req, CheckAccessResponse& resp) {
    CapabilityCertificate cert;
    AkentiMessage mesg;
    AkentiEngine eng;

    eng.checkAccess(req.getResource(), req.getAkentiPrincipal(), cert, mesg);

    if (cert) {
      resp.setCapabilityCertificate(cert);
    }

    resp.setID(req.getID());
    resp.setAkentiMessage(mesg);
  }

  static void
  handle(const ResourceBrowseRequest& req, ResourceBrowseResponse& resp){
    AkentiMessage mesg;
    Properties& props = Properties::getSystemProperties();
    vector< pair<string,string> > keyPairs = props.getKeyPairs();
    string resource = req.getResource();
    vector<string> dirs;

    if (resource == "") {
      for (int i=0; i < keyPairs.size(); i++) {
	if (StringUtil::equalsIgnoreCase(keyPairs[i].first, "RootResourceName")) {
	  dirs.push_back(keyPairs.at(i).second);
	}
      }

      resp.setResources(dirs);
      resp.setAkentiMessage(AkentiMessage(162));
      return;
    }

    vector<string> resources;

    if (!ParseResourceUtil::parseResource(resource,resources)) {
      resp.setAkentiMessage(AkentiMessage(701));
      return;
    }

    string base = props.getValue(std::make_pair("RootResourceName", resources[0]));

    if (!base.length()) {
      resp.setAkentiMessage(AkentiMessage(702));
      return;
    }

    string temp = base + '/' + resource;
    File dir(temp);

    if (!dir.isAbsolute()) {
      resp.setAkentiMessage(AkentiMessage(703));
      return;
    } else if (!dir.isDirectory()) {
      resp.setAkentiMessage(AkentiMessage(704));
      return;
    }

    for (vector<string> contents = dir.list(); !contents.empty(); contents.pop_back()) {
      File contentObject(dir.getAbsolutePath() + "/" + contents.back());

      if (contentObject.isDirectory()) {
	dirs.push_back(contentObject.getName());
      }
    }

    resp.setResources(dirs);
    resp.setAkentiMessage(AkentiMessage(161));
  }


  static void
  handle(const CreateResourceRequest& req, 
         CreateResourceResponse& resp,
         const Certificate& x509Cert){
    string resource = req.getResource();

    if (x509Cert) {
      AkentiEngine eng;
      AkentiMessage mesg;
      AkentiPrincipal issuer(x509Cert->getPrincipal(), x509Cert->getGuarantor()); 
      
      if (!eng.checkIssuer(resource, issuer, mesg)) {
        resp.setAkentiMessage(AkentiMessage(mesg));
        return;
      }
    }

    Properties& props = Properties::getSystemProperties();
    vector< pair<string,string> > keyPairs = props.getKeyPairs();
    vector<string> resources;

    if (!ParseResourceUtil::parseResource(resource, resources)) {
      resp.setAkentiMessage(AkentiMessage(711));
      return;
    }

    string base = props.getValue(std::make_pair("RootResourceName", resources[0]));

    if (!base.length()) {
      resp.setAkentiMessage(AkentiMessage(712));
      return;
    }

    string temp = base + '/' + resource;
    File dir(temp);

    if (!dir.isAbsolute()) {
      resp.setAkentiMessage(AkentiMessage(713));
      return;
    } else if (dir.exists()) {
      resp.setAkentiMessage(dir.isDirectory()? AkentiMessage(714) 
			    : AkentiMessage(715));
      return;
    }

    if (resources.size() == 1) {
      resp.setAkentiMessage(AkentiMessage(716));
      return;
    }

    try {
      dir.mkdir();
    } catch (Exception ex) {
      resp.setAkentiMessage(AkentiMessage(717));
      return;
    }

    resp.setAkentiMessage(AkentiMessage(171));
  }

  static void
  handle(const UploadPolicyCertificateRequest& req, 
         UploadPolicyCertificateResponse& resp,
         const Certificate& x509Cert) {
    const PolicyCertificate& polCert = req.getCertificate();
    string resource = polCert->getResource();

    if (x509Cert) {
      AkentiEngine eng;
      AkentiMessage mesg;
      AkentiPrincipal issuer(x509Cert->getPrincipal(), x509Cert->getGuarantor()); 
      
      if (!eng.checkIssuer(resource, issuer, mesg)) {
        resp.setAkentiMessage(AkentiMessage(mesg));
        return;
      }
    }

    Properties& props = Properties::getSystemProperties();
    vector< pair<string,string> > keyPairs = props.getKeyPairs();
    vector<string> resources;

    if (!ParseResourceUtil::parseResource(resource,resources)) {
      resp.setAkentiMessage(AkentiMessage(721));
      return;
    }

    string base = props.getValue(std::make_pair("RootResourceName", resources[0]));

    if (!base.length()) {
      resp.setAkentiMessage(AkentiMessage(722));
      return;
    }

    string temp = base + '/' + resource;
    File dir(temp);

    if (!dir.isAbsolute()) {
      resp.setAkentiMessage(AkentiMessage(723));
      return;
    } else if (!dir.isDirectory()) {
      resp.setAkentiMessage(AkentiMessage(724));
      return;
    }

    string filename = temp + "/" + props.getValue("PolicyFileName");
    string encType = props.getValue("AK_CERT_ENC_TYPE", AK_CERT_ENC_TYPE);

    AkentiCertificate::writeToFile(encType, filename, polCert);
    resp.setAkentiMessage(AkentiMessage(181));
  }

  static void
  handle(const DeletePolicyCertificateRequest& req, 
         DeletePolicyCertificateResponse& resp,
         const Certificate& x509Cert) {
    string resource = req.getResource();

    if (x509Cert) {
      AkentiEngine eng;
      AkentiMessage mesg;
      AkentiPrincipal issuer(x509Cert->getPrincipal(), x509Cert->getGuarantor()); 
      
      if (!eng.checkIssuer(resource, issuer, mesg)) {
        resp.setAkentiMessage(AkentiMessage(mesg));
        return;
      }
    }

    Properties& props = Properties::getSystemProperties();
    vector< pair<string,string> > keyPairs = props.getKeyPairs();
    vector<string> resources;

    if (!ParseResourceUtil::parseResource(resource,resources)) {
      resp.setAkentiMessage(AkentiMessage(731));
      return;
    }

    string base = props.getValue(std::make_pair("RootResourceName", resources[0]));

    if (!base.length()) {
      resp.setAkentiMessage(AkentiMessage(732));
      return;
    }

    string temp = base + '/' + resource + '/' + props.getValue("PolicyFileName");
    File certFile(temp);

    if (!certFile.isAbsolute()) {
      resp.setAkentiMessage(AkentiMessage(733));
      return;
    } else if (!certFile.isFile()) {
      resp.setAkentiMessage(AkentiMessage(734));
      return;
    }

    certFile.remove();
    resp.setAkentiMessage(AkentiMessage(191));
  }

  static void
  handle(const GetResourceInfoRequest& req, GetResourceInfoResponse& resp) {
    Properties& props = Properties::getSystemProperties();
    vector< pair<string,string> > keyPairs = props.getKeyPairs();
    string resource = req.getResource();
    vector<string> resources;

    if (!ParseResourceUtil::parseResource(resource, resources)) {
      resp.setAkentiMessage(AkentiMessage(741));
      return;
    }

    string base = props.getValue(std::make_pair("RootResourceName", resources[0]));

    if (!base.length()) {
      resp.setAkentiMessage(AkentiMessage(742));
      return;
    }

    File dir(base + '/' + resource);

    if (!dir.isAbsolute()) {
      resp.setAkentiMessage(AkentiMessage(743));
      return;
    } else if (!dir.isDirectory()) {
      resp.setAkentiMessage(AkentiMessage(744));
      return;
    }

    string hintFileName;
    string hintName = props.getValue("ResAttrFileName", ResAttrFileName);

    for (int i=resources.size()-1; i >= 0; i--) {
      File hintFile(base + '/' + resources[i] + '/' + hintName);
      if (hintFile.exists() && hintFile.isFile()) {
	hintFileName = hintFile.getAbsolutePath();
	break;
      }
    }

    if (hintFileName == "") {
      resp.setAkentiMessage(AkentiMessage(746));
    } else {
      ifstream is(hintFileName.c_str());
      is.seekg(0, std::ios::end);

      int len = is.tellg();
      string content(len, '\0');

      is.seekg(0, std::ios::beg);

      if (!is.read((char *) content.data(), len)) {
	resp.setAkentiMessage(AkentiMessage(745));
      } else {
	is.close();
	resp.setAkentiMessage(AkentiMessage(196));
	resp.setResourceInfo(content);
      }
    }
    return;

  }

  static void
  handle(const CheckPolicyRequest& req, CheckPolicyResponse& resp){
    PolicyChecker policyChecker;
    CertInfoCollector ciCol;
    AkentiMessage msg;
    bool ret = policyChecker.checkPolicy(req.getResource(), ciCol, msg);
    bool worked = ciCol.getStatus(); 

    resp.setAkentiMessage(worked? AkentiMessage (150) : AkentiMessage(450));
 
    vector<CertInfo>& cinfoVec = (vector<CertInfo>&) ciCol.getCertInfoVec();
 
    if (cinfoVec.size() == 0) {
      return;
    }
 
    vector<PolicyInfo>& pinfoVec = (vector<PolicyInfo>&) resp.getPolicyInfos();

    for (int i = 0; i <cinfoVec.size(); ++i) {  
      CertInfo &certInfo = cinfoVec[i];
      string type = certInfo.getType();
      CertificateInfo info;
   
      if (type == "dummy") { // root level error 
	PolicyInfo pinfo;
	pinfo.setResource (certInfo.getResource());
	pinfo.setPolicyCorrect (!certInfo.getError());
	vector<CertificateInfo>&
	  v1 = (vector<CertificateInfo>&)pinfo.getCertificateInfos();
	info.setUsed(certInfo.getUsed());
	info.setCertificateMessages(certInfo.getMessages());
	v1.push_back(info);
	pinfoVec.push_back (pinfo);
	if (certInfo.getMessages().size() == 1) { // should always be 1 here
	  vector<AkentiMessage>& mesgs = (vector<AkentiMessage>&)certInfo.getMessages(); 
	  resp.setAkentiMessage (mesgs[0]);
	}
      }
      else if (type == "Policy") { // found a decent policy cert
	PolicyInfo pinfo;
	pinfo.setResource (certInfo.getResource());
	pinfo.setPolicyCorrect (!certInfo.getError());
	vector<CertificateInfo>&
	  v1 = (vector<CertificateInfo>&)pinfo.getCertificateInfos();
	info.setUsed(certInfo.getUsed());
	info.setCertificateMessages(certInfo.getMessages());
	info.setCertificate (certInfo.getCertificate());
	v1.push_back(info);
	pinfoVec.push_back (pinfo);
      }
    }

    for (int i = 0; i <cinfoVec.size(); ++i) {  
      CertInfo &cinfo = cinfoVec[i];
      string type = cinfo.getType();
      CertificateInfo info;
   
    if (type == "Use Condition") { // Found Use Condition cert
       info.setUsed(cinfo.getUsed());
       info.setCertificateMessages(cinfo.getMessages());
       info.setCertificate (cinfo.getCertificate());
       info.setLocation (cinfo.getLocation());
       for (int j = 0; j < pinfoVec.size(); j++) {
          if (pinfoVec[j].getResource() == cinfo.getResource()) {
	    PolicyInfo& pinfoRef = pinfoVec[j];
	    vector<CertificateInfo>&
	      v1 = (vector<CertificateInfo>&) pinfoRef.getCertificateInfos();
 
	    v1.push_back(info);
	    break;
	  }
	}
      }
    }
  }

  static void 
  doLocal(const AkentiRequest& req, AkentiResponse& resp, const Certificate& cert, PH::Type type) {
    resp.setID(req.getID());

    if (req.instanceOf<CheckAccessRequest>()) {
      handle((CheckAccessRequest&) req, (CheckAccessResponse&) resp);
    } else if (req.instanceOf<CheckPolicyRequest>()) { // SRS 
      handle((CheckPolicyRequest&) req, (CheckPolicyResponse&) resp);
    } else if (req.instanceOf<ResourceBrowseRequest>()) {
      handle((ResourceBrowseRequest&) req, (ResourceBrowseResponse&) resp);
    } else if (req.instanceOf<GetResourceInfoRequest>()) {
      handle((GetResourceInfoRequest&) req, (GetResourceInfoResponse&) resp);
    } else if (type == PH::SERVER && !cert) {
      resp.setAkentiMessage(380);
    } else if (req.instanceOf<CreateResourceRequest>()) {
      handle((CreateResourceRequest&) req, 
             (CreateResourceResponse&) resp,
             cert);
    } else if (req.instanceOf<UploadPolicyCertificateRequest>()) {
      handle((UploadPolicyCertificateRequest&) req, 
             (UploadPolicyCertificateResponse&) resp,
             cert);
    } else if (req.instanceOf<DeletePolicyCertificateRequest>()) {
      handle((DeletePolicyCertificateRequest&) req, 
             (DeletePolicyCertificateResponse&) resp,
             cert);
    } else {
      throw Error("ProtocolHandler:doLocal:bad request type");
    }
  }

  static void 
  doLocal(const AkentiRequests& requests, 
	  AkentiResponses& responses,
	  const Certificate& cert,
          PH::Type type) {
    for (int i = 0; i < requests.numOfRequests(); ++i) {
      AkentiRequest req = requests.getRequestAt(i);
      AkentiResponse resp;
 
      if (req.instanceOf<CheckAccessRequest>()) {
	doLocal(req, resp = CheckAccessResponse(), cert, type);
      } else if (req.instanceOf<CheckPolicyRequest>()) {
	doLocal(req, resp = CheckPolicyResponse(), cert, type);
      } else if (req.instanceOf<ResourceBrowseRequest>()) {
	doLocal(req, resp = ResourceBrowseResponse(), cert, type);
      } else if (req.instanceOf<CreateResourceRequest>()) {
	doLocal(req, resp = CreateResourceResponse(), cert, type);
      } else if (req.instanceOf<UploadPolicyCertificateRequest>()) {
	doLocal(req, resp = UploadPolicyCertificateResponse(), cert, type);
      } else if (req.instanceOf<DeletePolicyCertificateRequest>()) {
	doLocal(req, resp = DeletePolicyCertificateResponse(), cert, type);
      } else if (req.instanceOf<GetResourceInfoRequest>()) {
	doLocal(req, resp = GetResourceInfoResponse(), cert, type);
      } else {
	throw Error("ProtocolHandler:doLocal:bad request type");
      }

      responses.addResponse(resp);
    }
  }

  // END LOCAL

  void 
  PH::handle(const AkentiRequest& req, 
	     AkentiResponse& resp,
	     const Certificate& cert) {
    type == CLIENT? doRemote(req, resp) : doLocal(req, resp, cert, type);
  }

  void 
  PH::handle(const AkentiRequests& requests, 
	     AkentiResponses& responses,
	     const Certificate& cert) {
    type == CLIENT? doRemote(requests, responses) 
      : doLocal(requests, responses, cert, type);
  }

  ProtocolHandler& 
  PH::getInstance(Type type) {
    static ProtocolHandler handler(type);

    return handler;
  }

}
