#ifndef AK_URL_CONNECTION_IMPL_HPP
#define AK_URL_CONNECTION_IMPL_HPP

#include <sstream>
#include "Error.hpp"
#include "InputStream.hpp"
#include "OutputStream.hpp"
#include "URL.hpp"

using std::ostringstream;
	
/**
  * A <i>URLConnectionImpl<i> is an abstract class that  holds a URL and its
  * connection state, and input and/or output streams. The open, close,
  * getInputSream, getOutputStream and getContent calls are implmented
  * by the protocol-specific ConnectionImpls which are implemented in the same
  * files as the protocol-specific URLHandlerImpls.
  *
  * @see URL
  * @see InputStream
  * @see OutputStream
  *
  * @version 1.1 00/11/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

namespace Akenti {

class URLConnectionImpl {

public:

  /**
    * Destructor
    */
  virtual ~URLConnectionImpl() {
  }

  /**
    * Returns the url 
    *
    * @return the url
    */
  const URL& getURL() const {
    return url;
  }

  /**
    * Sets the connection to doIn. Must be called before connect.
    *
    * @param doInp if true, connection is going to be used for input, if
    *   false, it will not.
    * @exception Error if setDoInput: illegal operation if it is 
    *   already connected
    */
  void setDoInput(bool doInp) {
    if (isConnected()) {
      throw Error("URLConnectionImpl:setDoInput:illegal operation");
    }

    this->doInput = doInp;
  }

  /**
    * Returns an indication if this connection can be used for input.
    *
    * @return true if this connection can be used for input, otherwise false
    */
  bool getDoInput() const {
    return doInput;
  } 

  /**
    * Sets the connection to doOutp. Must be called before connect.
    *
    * @param doOutput if true, connection is going to be used for output, if
    *   false, it will not
    * @exception Error if setDoOutput: illegal operation if it is already 
    *   connected.
    */
  void setDoOutput(bool doOutp) {
    if (isConnected()) {
      throw Error("URLConnectionImpl:setDoOutput:illegal operation");
    }

    this->doOutput = doOutp;
  }

  /**
    * Returns an indication if this connection can be used for output.
    *
    * @return true if this connection can be used for output, otherwise false
    */
  bool getDoOutput() const {
    return doOutput;
  } 

  /**
    * Determines if there is a connection or not
    *
    * @return true if there is connection, false otherwise
    */
  bool isConnected() const {
    return connected;
  }
  
  /**
    * Establishs a connection by calling the subclass doOpen method. Sets
    * isConnected to true.
    *
    * @exception Error if a connection already existed
    * @exception Error if an illegal operation if neither DoInput or DoOutput
    *     has been set
    * @see doOpen() in the subclass for each protocol - part of the subclass
    *       URLHandlerImpl.
    * @see setConnected(bool);
    */
  void connect() {
    if (isConnected()) {
      throw Error("URLConnectionImpl:connect:connection already established");
    } else if (!getDoInput() && !getDoOutput()) {
      throw Error("URLConnectionImpl:connect:illegal operation");
    } 

    doOpen();
    setConnected(true);
  } 

  /**
    * Closes the connection by calling doClose in the subclass.
    * 
    * @exception Error if the connection is not open 
    * @see doClose()
    * @see setConnected(bool)
    */
  void close() {
    if (!isConnected()) {
      throw Error("URLConnectionImpl:connect:connection is not open");
    }

    doClose();
    setConnected(false);
    is = NULL;
    os = NULL;
  }

  /**
    * Returns the input stream. The first time this is called, it calls
    * doGetInputStream in the subclass to open an input stream specific
    * the the protocol type. It then saves the streamObject util a close
    * is called.
    *
    * @exception Error if the connection is not open
    * @exception Error if the getInputStream is not set for input
    * @return the input stream
    */
  InputStream& getInputStream() {
    if (!isConnected()) {
      throw Error("URLConnectionImpl:getInputStream:connection is not open");
    } else if (!getDoInput()) {
      throw Error("URLConnectionImpl:getInputStream:not set for input");
    } else if(!is) {
      is = doGetInputStream();
    }

    return is;
  }

  /**
    * Returns the output stream. The first time this is called, it calls
    * doGetOutputStream in the subclass to open an output stream specific
    * the the protocol type. It then saves the streamObject util a close
    * is called.
    *
    * @exception Error if the connection is not open
    * @exception Error if the getOutputStream is not set for output
    * @return the output stream
    */
  OutputStream& getOutputStream() {
    if (!isConnected()) {
      throw Error("URLConnectionImpl:getOutputStream:connection is not open");
    } else if (!getDoOutput()) {
      throw Error("URLConnectionImpl:getOutputStream:not set for output");
    } else if(!os) {
      os = doGetOutputStream();
    }

    return os;
  }

  /**
    * Returns the content of the input stream by calling the subclass method
    * doGetContent that know how to read for this type of protocol.
    * 
    * @return the content of the input stream
    */
  string getContent() {
    return doGetContent(getInputStream());
  }

  /**
    * Returns a string representation for URLConnectionImpl which is used  
    * for debugging purposes.
    *
    * @return debugging info
    */
  virtual string paramString() const {
    ostringstream oss;

    oss << '[' 
       << "url"
       << '='
       << '"'
       << getURL().paramString()
       << '"'
       << ','
       << "doInput"
       << '='
       << getDoInput() 
       << ','
       << "doOutput"
       << '='
       << getDoOutput()
       << ','
       << "connected" 
       << '='
       << isConnected() 
       << ']';
    return oss.str(); 
  } 
 
  
protected:
  
  /** 
    * Create this object. Defaults to allowing both input and output.
    *
    * @param url the url to make connection
    */
  URLConnectionImpl(const URL& url)
    : url(url), doInput(true), doOutput(true), connected(false) {
  } 

  /**
    * To open the connection. Must be implemented by the subclasses.
    */
  virtual void doOpen() = 0;

  /**
    * To get the content from the input stream. Must be implemented by 
    * the subclasses. 
    *
    * @param is the input stream
    * @return the content of the input stream
    */
  virtual string doGetContent(InputStream &is) = 0;
  
  /**
    * Closes the connection. Must be implemented by the subclasses.
    */
  virtual void doClose() = 0;
  
  /**
    * Returns the input stream Must be implemented by the subclasses.
    *
    * @return the input stream
    */
  virtual InputStream doGetInputStream() = 0; 

  /**
    * Returns the output stream. Must be implemented by the subclasses. 
    *
    * @return the output stream
    */
  virtual OutputStream doGetOutputStream() = 0;
 
private:
  
  /**
    * Sets the value to conn, to indicate whether a connection is 
    * is made or not  
    *
    * @param conn the value to indicate to indicate the connection
    */
  void setConnected(bool conn) {
    this->connected = conn;
  }

private:

  URL url;

  InputStream is;

  OutputStream os;

  bool doInput;
  
  bool doOutput;

  bool connected;
  
};

} // end of namespace Akenti

#endif
