#include "OpenSSLContextImpl.hpp" 
#include "OpenSSLX509CertificateImpl.hpp"
#include "StringTokenizer.hpp"
#include "File.hpp"
#include "SecurityException.hpp"
#include <vector>

typedef Akenti::OpenSSLContextImpl OSCI;

OSCI::OpenSSLContextImpl(Version ver, bool isClnt) 
  : SSLContextImpl(ver, isClnt) {
   
   switch(ver) {
     case SSL2:
       throw SecurityException("OpenSSLContextImpl:constructor: ssl2 not supported");
     case SSL3:
       if (isClnt) {
         ctx = SSL_CTX_new(SSLv3_client_method());
       } else {
         ctx = SSL_CTX_new(SSLv3_server_method());
       }

       break;
     case TLS1:
       if (isClnt) {
         ctx = SSL_CTX_new(TLSv1_client_method());
       }  else {
         ctx = SSL_CTX_new(TLSv1_server_method());
       }
       break;
     default:
       throw Error("OpenSSLContextImpl:constructor:bad ssl version type");
   }

   if (!ctx) {
     // XXX We need to do better here.
     throw SecurityException("OpenSSLContextImpl:constructor:failed init context");
   }

   // SSL_CTX_set_options(ctx, SSL_OP_TLS_ROLLBACK_BUG);
}

OSCI::~OpenSSLContextImpl() {
  SSL_CTX_free(ctx);
}

void 
OSCI::doSetVerify(int verify) {
  SSL_CTX_set_verify(ctx, verify, NULL);
}

void 
OSCI::setCertificateFile(const string& certFile, FileType type) {
  int dtype = (type == PEM) ?  SSL_FILETYPE_PEM : SSL_FILETYPE_ASN1;

  /* 
  if (SSL_CTX_use_certificate_file(ctx, certFile.c_str(), dtype) <= 0) {
    throw SecurityException("OpenSSLContextImpl:setCertificateFile:failed");
  }
  */

  if (SSL_CTX_use_certificate_chain_file(ctx, certFile.c_str()) <= 0) {
    throw SecurityException("OpenSSLContextImpl:setCertificateFile:failed");
  }
}

void 
OSCI::setPrivateKeyFile(const string& keyFile, FileType type) {
  int dtype = (type == PEM) ?  SSL_FILETYPE_PEM : SSL_FILETYPE_ASN1;

  if (SSL_CTX_use_PrivateKey_file(ctx, keyFile.c_str(), dtype) <= 0) {
    throw SecurityException("OpenSSLContextImpl:setPrivateKeyFile:failed");
  }
}

void
OSCI::addTrustedCA(const Certificate& cert) {
  OpenSSLX509CertificateImpl* impl = dynamic_cast<OpenSSLX509CertificateImpl *> (cert.get());
   
  if (impl == NULL) {
    throw Error("OpenSSLContextImpl:doAddTrustedCA:bad argument");
  }

  // TODO: This may fail
  X509_STORE_add_cert(ctx->cert_store, (X509 *) impl->getX509());
}

void
OSCI::setTrustedCAS(const string& filename) {
  File file(filename);

  if (!file.isDirectory()) {
    if (!(SSL_CTX_load_verify_locations(ctx, filename.c_str(), NULL))) {
      throw SecurityException("OpenSSLContextImpl:setCAFile:setting CA file failed");
    }
    
    return;
  }

  vector<string> files = file.list();
  int num = 0;

  for (int i=0; i < files.size(); i++) {
    string temp = filename + "/" + files[i];
    
    if (!File(temp).isDirectory()) {
      if ((SSL_CTX_load_verify_locations(ctx, temp.c_str(), NULL))) {
        num++;
      }
    }
  }

  if (num == 0) {
    throw SecurityException("SecureSocketImpl:setCAS:setting CA list failed");
  }
}
