#include <openssl/evp.h>
#include <openssl/rand.h>
#include <openssl/bio.h>
#include <openssl/pem.h>
#include "OpenSSLKeyFactoryImpl.hpp"
#include "OpenSSLPublicKeyImpl.hpp"
#include "OpenSSLPrivateKeyImpl.hpp"
#include "PublicKey.hpp"
#include "PrivateKey.hpp"
#include "KeyPair.hpp"
#include "OpenSSLKeyImpl.hpp"
#include "SecurityException.hpp"
#include "IOException.hpp"
#include "Error.hpp"

typedef Akenti::OpenSSLKeyFactoryImpl OSKFI;

OSKFI::OpenSSLKeyFactoryImpl() {
}

OSKFI::~OpenSSLKeyFactoryImpl() {
}

/**
  * It would be nice if openssl provided the same 
  * hack it provided for private keys.
  *
  * To keep the interface consistent, we provide our own hack.
  */
Akenti::PublicKey 
OSKFI::initPublicKey(const string& format, const string& encoding) const {
  if (format == "DER") {
    unsigned char* ptr = (unsigned char *) encoding.data();
    EVP_PKEY *evp = d2i_PUBKEY(NULL, &ptr, encoding.length()); 
 
    if (evp == NULL) {
      throw SecurityException("OpenSSLKeyFactoryImpl:initPublicKey:unknown type");
    }

    return PublicKey(new OpenSSLPublicKeyImpl(evp));

/*
    unsigned char* ptr = (unsigned char *) encoding.data();
    EVP_PKEY *evp = d2i_PublicKey(EVP_PKEY_RSA, NULL, &ptr, encoding.length());

    if (evp == NULL) {
      evp = d2i_PublicKey(EVP_PKEY_DSA, NULL, &ptr, encoding.length());
    }
    
    if (evp == NULL) {
      throw SecurityException("OpenSSLKeyFactoryImpl:initPublicKey:unknown type");
    }
*/

    return PublicKey(new OpenSSLPublicKeyImpl(evp));
  } else {
    throw SecurityException("OpenSSLKeyFactoryImpl:initPublicKey:unknown format " + format);
  }
}

string
OSKFI::getKeyEncoded(const string& format, const PublicKey& publicKey) const {
  if (format == "DER") {
    OpenSSLPublicKeyImpl *impl = dynamic_cast<OpenSSLPublicKeyImpl *> (publicKey.get());

    if (!impl) {
      throw Error("OpenSSLKeyFactoryImpl:getKeyEncoded:bad key");
    }

    EVP_PKEY* evp = impl->getEVP_PKEY();

    int len = i2d_PUBKEY(evp, NULL);

    if (len == -1) {
      throw SecurityException("OpenSSLKeyFactoryImpl:getEncoded:encoding failed");
    }

    string encoding(len, '\0');
    unsigned char* temp = (unsigned char *) encoding.data();

    if (i2d_PUBKEY(evp, &temp) != len) {
      throw SecurityException("OpenSSLKeyFactoryImpl:getEncoded:encoding failed");
    }

    return encoding;
  } else {
    throw SecurityException("OpenSSLKeyFactoryImpl:getKeyEncoded:unknown format" + format);
  }
}

Akenti::PrivateKey 
OSKFI::initPrivateKey(const string& format, const string& encoding) const {
  if (format == "DER") {
    unsigned char* ptr = (unsigned char *) encoding.data();
    EVP_PKEY *evp = d2i_AutoPrivateKey(0, &ptr, encoding.length());
    
    if (!evp) {
      throw SecurityException(
        "OpenSSLKeyFactoryImpl:initPrivateKey:failed to decode key"
      );
    }
    
    return PrivateKey(new OpenSSLPrivateKeyImpl(evp));
  } else if (format == "PEM") {
    BIO* bio = BIO_new_mem_buf((char*) encoding.data(), encoding.length());

    if (!bio) {
      throw Error("OpenSSLKeyFactoryImpl:initPrivateKey:bad alloc");
    }

    EVP_PKEY *evp = PEM_read_bio_PrivateKey(bio, NULL, NULL, NULL);
    
    BIO_free(bio);

    if (!evp) {
      throw SecurityException("OpenSSLKeyFactoryImpl:initPrivateKey:bad pem encoding");
    }

    return PrivateKey(new OpenSSLPrivateKeyImpl(evp));
  } else {
    throw SecurityException("OpenSSLKeyFactoryImpl:initPrivateKey:unknown format " + format);
  }
}

Akenti::PrivateKey 
OSKFI::initPrivateKeyFromFile(const string& format, 
                              const string& fileName) const {
  if (format == "PEM") {
    BIO* bio = BIO_new_file((char *) fileName.c_str(), "r");

    if (!bio) {
      throw IOException(
        "OpenSSLKeyFactoryFactoryImpl:initPrivateKeyFromFile:can't open file " + fileName
      );
    }

    EVP_PKEY *evp = PEM_read_bio_PrivateKey(bio, NULL, NULL, NULL);

    BIO_free(bio);

    if (!evp) {
      throw SecurityException(
        "OpenSSLKeyFactoryImpl:initPrivateKeyFromFile:bad pem encoding"
      );
    }

    return PrivateKey(new OpenSSLPrivateKeyImpl(evp));
  } else {
    throw SecurityException(
      "OpenSSLKeyFactoryImpl:initPrivateKeyFromFile:unknown format " + format)
    ;
  }
}

string 
OSKFI::getKeyEncoded(const string& format, const PrivateKey& privateKey) const {
  OpenSSLPrivateKeyImpl* impl = dynamic_cast<OpenSSLPrivateKeyImpl *> (privateKey.get());

  if (!impl) {
    throw Error("OpenSSLSignatureImpl:initSign:bad key");
  }

  EVP_PKEY* evp = impl->getEVP_PKEY();

  if (format == "DER") {
    int len = i2d_PrivateKey(evp, NULL);

    if (len == -1) {
      throw SecurityException("OpenSSLRSAPrivateKeyImpl:getEncoded:encoding failed");
    }

    string encoding(len, '\0');
    unsigned char* temp = (unsigned char *) encoding.data();

    if (i2d_PrivateKey(evp, &temp) != len) {
      throw SecurityException("OpenSSLRSAPrivateKeyImpl:getEncoded:encoding failed");
    }

    return encoding;
  } else if (format == "PEM") {
    BIO *bio = BIO_new(BIO_s_mem());

    PEM_write_bio_PrivateKey(bio, evp, NULL, NULL, 0, NULL, NULL);

    int len = bio->num_write;
    string encoding(len, '\0');

    BIO_read(bio, (char *) encoding.c_str(), len);
    BIO_free(bio);
    return encoding;
  } else {
    throw SecurityException("OpenSSLKeyFactoryImpl:getKeyEncoded:unknown format " + format);
  }
}


Akenti::KeyPair 
OSKFI::generateRSAKeyPair(int size, unsigned long exponent) const {
  RSA *rsa = RSA_generate_key(size * 8, exponent, NULL, NULL);
    
  if (!rsa) {
    throw SecurityException(
      "OpenSSLKeyFactoryImpl:generateRSAKeyPair:failed to generate RSA key"
    );
  } 
    
  EVP_PKEY *evp = EVP_PKEY_new();

  if (!evp) {
    RSA_free(rsa);

    throw Error(
      "OpenSSLKeyFactoryImpl:generateRSAKeyPair:EVP_PKEY_new failed"
    ); 
  }

  EVP_PKEY_assign(evp, EVP_PKEY_RSA, (char *) rsa);

  /**
    * This is OK even when threads are involved. :)
    */
  evp->references = 2;                       

  return KeyPair(
    PublicKey(new OpenSSLPublicKeyImpl(evp)),
    PrivateKey(new OpenSSLPrivateKeyImpl(evp))
  );
}

Akenti::KeyPair 
OSKFI::generateDSAKeyPair(int size) const {
  DSA *dsa = DSA_generate_parameters(
    size * 8,
    NULL, 
    0, 
    NULL, 
    NULL, 
    NULL, 
    NULL
  );
   
  if (!dsa) {
    throw SecurityException(
      "OpenSSLKeyFactoryImpl:generateKeyPair:failed to generate DSA parameters"
    );
  } else if(!DSA_generate_key(dsa)) {
    DSA_free(dsa);
    throw SecurityException(
      "OpenSSLKeyFactoryImpl:generateKeyPair:failed to generate DSA key"
    );
  }

  EVP_PKEY* evp = EVP_PKEY_new();

  if (!evp) {
    DSA_free(dsa);
    throw Error(
      "OpenSSLKeyFactoryImpl:generateDSAKeyPair:EVP_PKEY_new failed"
    ); 
  }

  EVP_PKEY_assign(evp, EVP_PKEY_DSA, (char *) dsa);

  /**
    * This is OK even when threads are involved. :)
    */
  evp->references = 2;

  return KeyPair(
    PublicKey(new OpenSSLPublicKeyImpl(evp)),
    PrivateKey(new OpenSSLPrivateKeyImpl(evp))
  );
}

