#ifndef OPEN_SSL_MESSAGE_DIGEST_IMPL_HPP 
#define OPEN_SSL_MESSAGE_DIGEST_IMPL_HPP 

#include <openssl/evp.h>
#include "MessageDigestImpl.hpp"

namespace Akenti {

/**
  * This class is instantiated from superclass <i>MessageDigestImpl</i> and 
  * uses hash algorithms such as MD5 or SHA1 to produce a fixed length
  * output from an arbitrary-length input message.It wraps a ssl EVP_MD_CTX.
  *
  * @see MessageDigestImpl
  * 
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */  

class OpenSSLMessageDigestImpl : public MessageDigestImpl {

public:

  /**
    * Constructs this object with the specified algorithm. e.g. MD5 or SHA1.
    *  
    * @param alg const string & the name of the algorithm 
    * @exception SecurityException if no such algorithm
    * @see EVP_get_digestbyname
    */
  OpenSSLMessageDigestImpl(const string& alg);

  /**
    * Deletes this object.
    */
  ~OpenSSLMessageDigestImpl();

  /**
    * Resets the engine for further use.
    */
  void reset();

protected:

  /**
    * Updates the engine using the specified number of bytes, starting
    * at the specified offset. The user calls the update method which is 
    * implemented by the superclass which in turn calls this protected method
    * to do the openSSL specific actions.
    *
    * @param input string of bytes.
    * @param offset the offset to start from.
    * @param len the number of bytes to use starting at <i>offset</i>.
    * @exception Error if input buffer too short
    * @see update(const string&)
    */
  void doUpdate(const string& data, int offset, int len);

  /**
    * Completes the hash computation.
    * You should reset the digest after this call is made.
    *
    * @return a string of bytes for the resulting hash value.
    */
  string doDigest();

private:

  EVP_MD_CTX context;
  
};

} // end namespace Akenti

#endif
