#ifndef AK_CERTIFICATE_FACTORY_IMPL_HPP
#define AK_CERTIFICATE_FACTORY_IMPL_HPP

#include <string>
#include <vector>

using std::vector;
using std::string;

namespace Akenti {

class Certificate;
class CertificateChain;
class CertificateGenerator;
class CertificateRequest;
class CertificateRequestGenerator;
class CertificateRevocationList;

/**
  * This abstract class provides applications with an interface for
  * X509 certificate  generation, encoding and decoding.The actual 
  * implemenation of this class is done by a provider class, for example
  * OpenSSLCertificateFactoryImpl. Akenti certificates are handled by
  * the AkentiCertificateFactory class, rather than this class.
  *
  * Typically, an instance of this class is instantiated 
  * and wrapped in a <i>CertificateFactory</i> object which is
  * merely a <b>reference-count </b> pointer.
  *
  * Applications should never have to instantiate a subclass of
  * this class. Instead, the <i>CertificateFactory#getCertificateFactory()</i>
  * should be used. 
  *
  * @see CertificateFactory
  * @see Provider
  * @see ProviderImpl
  * @see Certificate 
  * @see CertificateGenerator 
  * @see CertificateRequestGenerator 
  * @see CertificateRequest
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

class CertificateFactoryImpl {

public:

  /**
    * Deletes this object.
    */
  virtual ~CertificateFactoryImpl() {
  }

  /**
    * Returns the <i>Certificate</i> from a file containing a PEM encoded
    *  X.509 certificate.
    *
    * @param type the encoding type of Certificate, e.g. "PEM" 
    * @param fileName name of file from which to  initialize the Certificate 
    * @return collection of Certificates
    */
  virtual Certificate
  initCertificateFromFile(const string& type, const string& fileName) const = 0;

  /**
    * Returns a collection <i>Certificates</i> from a file containing a 
    * one or more PEM encoded X.509 certificates.
    *
    * @param type the encoding type of Certificate, e.g. "PEM
    * @param fileName name of file from which to initialize the Certificate 
    * @return Certificate
    */
  virtual vector<Certificate>
  initCertificatesFromFile(const string& type, const string& fileName) const = 0;

  /**
    * Returns a <i>Certificate</i> class given an encoded string containing the 
    * certificate. 
    * The <i>type</i> specifies the way the Certificate string is encoded, 
    * e.g. PEM, DER
    *
    * @param type the Certificate's type of encoding
    * @param encoding a string of bytes
    * @return  a Certificate
    * @exception SecurityException if either type is unsupported
    *   or if decoding fails
    * @see CertificateFactory 
    * @see CertificateFactoryImpl  
    */
  virtual Certificate
  initCertificate(const string& type, const string& encoding) const = 0;

  /**
    * Returns collection of <i>Certificates</i> from a a string containing
    * one or more PEM encoded certificates.
    *
    * @param type string specifies the encoding type. Only PEM is 
    *     currently supported
    * @param encoding a string of bytes 
    * @exception Error if out of memory
    * @exception SecurityException if type is unsupported
    * @see Error
    * @see SecurityException
    */
  virtual vector<Certificate>
  initCertificates(const string& type, const string& encoding) const = 0;

  /**
    * Returns a CertificateChain from a file containing a PEM encoded
    * certificate chain.
    *
    * @param type the encoding type of Certificate, e.g. "PEM
    * @param fileName name of file from which to initialize the Certificate
    *   chain
    * @return CertificateChain
    */
  virtual CertificateChain
  initCertificateChainFromFile(const string& type, const string& fname) const = 0;

  /**
    * Returns a CertificateChain from a string containing a PEM encoded
    * certificate chain.
    *
    * @param type the encoding type of Certificate, e.g. "PEM
    * @param encoding a string of bytes 
    * @return CertificateChain
    */
  virtual CertificateChain
  initCertificateChain(const string& type, const string& encoding) const = 0;

  /**
    * Encodes the <i>Certificate</i> using the encoding type specified by 
    * <i>type</i>.
    *
    * @param type string specifies the encoding type to be used
    * @param certificate Certificate to be encoded
    * @return encoding a string of bytes
    * @exception Error if certificate is unitialized or unrecognized
    * @exception SecurityException if type is unsupported or if the
    *   encoding fails
    * @see Certificate 
    * @see CertificateImpl
    */
  virtual string
  getCertificateEncoded(const string& type, 
                        const Certificate& certificate) const = 0;

  /**
    * Return an instance of <i>CertificateGenerator</i> using 
    * a type specified by <i>type</i>
    *
    * @param type string specifies the type to be used 
    * @see CertificateGenerator
    * @see CertificateGeneratorImpl
    */ 
  virtual CertificateGenerator
  getCertificateGenerator(const string& type) const = 0;

  /**
    * Returns a <i>CertificateRequest</i>.
    * The <i>type</i> specifies the kind of CertificateRequest
    * we are decoding.
    *
    * @param type the CertificateRequest's type
    * @param encoding a string of bytes representing a CertificateRequest.
    * @return  a CertificateRequest
    * @exception SecurityException if type  is unsupported
    *  or if decoding fails
    * @see CertificateRequest
    * @see CertificateRequestImpl
    */
  virtual CertificateRequest
  initCertificateRequest(const string& type, const string& encoding) const = 0;

  /**
    * Encodes the <i>CertificateRequst</i> using the encoding type 
    * specified by <i>type</i>.
    *
    * @param type string specifies the type to be used
    * @param request CertificateRequest to be encoded
    * @return encoding a string of bytes representating the certificateRequest
    *   that could be sent to a CA for signing.
    * @exception Error if request is unitialized or unrecognized
    * @exception SecurityException if type is unsupported or if the
    *   encoding fails
    * @see CertificateRequest
    * @see CertificateRequestImpl
    */
  virtual string
  getCertificateRequestEncoded(const string& type, 
                               const CertificateRequest& request) const = 0;

  /**
    * Return an instance of <i>CertificateRequestGenerator</i> using
    * a type specified by <i>type</i>
    *
    * @param type string specifies the type to be used
    * @see CertificateRequestGenerator
    * @see CertificateRequestGeneratorImpl
    */
  virtual CertificateRequestGenerator
  getCertificateRequestGenerator(const string& type) const = 0;

  /**
    * Return an instance of <i>CertificateRevocationList</i> using
    * the type specified.
    *
    *   @param crl string that refers to the crl encoding
    * @param type string that refers to the type of crl encoding
    * @see CertificateRevocationList
    */
  virtual CertificateRevocationList
  initCRL(const string& crl, const string& type) const = 0;

  /**
    * Return an instance of <i>CertificateRevocationList</i> from
    * the given filename and the encoding type
    *
    * @param crlfile string that refers to the crl filename
    * @param type string that refers to the type of crl encoding
    * @see CertificateRevocationList
    */
  virtual CertificateRevocationList
  initCRLFromFile(const string& crlfile, const string& type) const = 0;
};

} // end of namespace Akenti

#endif
