#include "DistinguishedName.hpp"
#include "StringTokenizer.hpp"
#include "ParsingException.hpp"
#include "StringUtil.hpp"
#include "Error.hpp"

/** 
  * commonName                   "CN"
  * user ID                      "UID"
  * countryName                  "C"
  * localityName                 "L"
  * Email Address                "E"
  * Postal Address               "PA" 
  * stateOrProvinceName          "ST"
  * organizationName             "O"
  * organizationalUnitName       "OU"
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  **/






static const char* DEFAULTS[] = {
  "CN", "UID", "C", "L", "E", "PA", "ST", "O", "OU", NULL 
};

typedef Akenti::DistinguishedName DN;

DN::DistinguishedName(const string& dn) {
  int size = dn.size();

  if (!size || dn[0] != '/' || dn[size - 1] == '/') {
    throw ParsingException("DistinguishedName:DistinguishedName:bad dn:" + dn);
  }

  StringTokenizer st(dn, "/", false);

  if (!st.hasMoreTokens()) {
     throw ParsingException(dn);
  }

  string temp = st.nextToken();

  int idx = temp.find("=");
    
  if (idx <= 0 || idx == temp.size() - 1) {
    throw ParsingException("DistinguishedName:DistinguishedName:bad dn:" + dn);
  }

  pairs.push_back(Pair(temp.substr(0, idx), temp.substr(idx + 1)));

  while (st.hasMoreTokens()) {
    temp = st.nextToken();

    idx = temp.find("=");

    if (idx == 0 || idx == temp.size() - 1) {
       throw ParsingException("DistinguishedName:DistinguishedName:bad dn:" + dn);
    } else if (idx != -1) {
       pairs.push_back(Pair(temp.substr(0, idx), temp.substr(idx + 1)));
    } else {
       pairs[pairs.size() - 1].second += '/';
       pairs[pairs.size() - 1].second += temp;
    }
  }
}

DN::~DistinguishedName() {
}

string
DN::toString() const {
  string dn;

  for(int i = 0; i < pairs.size(); ++i) {
    dn += "/";
    dn += pairs[i].first;
    dn += "=";
    dn += pairs[i].second;
  }
  
  return dn;
}

bool
DN::operator==(const DistinguishedName& dn) const {
  bool ret = pairs.size() == dn.pairs.size();

  for (int i = 0; ret && i < pairs.size(); ++i) {
    ret = pairs[i].second == dn.pairs[i].second;
    ret = ret && StringUtil::equalsIgnoreCase(
      pairs[i].first, dn.pairs[i].first
    );
  }
  
  return ret;
}

bool 
DN::containsAttributeValue(const string& attribute, const string& value) const {

  bool found = pairs[0].second == value; 

  found = found && StringUtil::equalsIgnoreCase(
    pairs[0].first, attribute
  );

  for (int i = 1; !found  && i < pairs.size(); ++i) {
    found = pairs[i].second == value; 
    found = found && StringUtil::equalsIgnoreCase(
      pairs[i].first, attribute
    );
  }
  
  return found;
}

bool
DN::containsAttribute(const string& attribute) const {
  return numOfOcurrences(attribute) != 0;
}

bool
DN::containsValue(const string& value) const {
  bool found = pairs[0].second == value;
 
  for(int i = 1; !found && i < pairs.size(); ++i) {
    found = pairs[i].second == value;
  }
  
  return found;
}

int 
DN::numOfOcurrences(const string& attribute) const {
  int count = StringUtil::equalsIgnoreCase(
      pairs[0].first, attribute
  );
  
  for(int i = 1; i < pairs.size(); ++i) {
    count += StringUtil::equalsIgnoreCase(
      pairs[i].first, attribute
    );
  }
  
  return count;
}

const string&
DN::getValue(const string& attribute) const {
  for(int i = 0; i < pairs.size(); ++i) {
    if (StringUtil::equalsIgnoreCase(pairs[i].first, attribute)) {
      return pairs[i].second;
    }
  }

  throw Exception("DistinguishedName:getValue: no value found for" + attribute);
}

vector<string>
DN::getValues(const string& attribute)  {
  vector<string> values;

  for(int i = 0; i < pairs.size(); ++i) {
    if (StringUtil::equalsIgnoreCase(pairs[i].first, attribute)) {
      values.push_back(pairs[i].second);
    }
  }
 
  return values;
}

int 
DN::size() const {
  return pairs.size();
}

bool 
DN::hasMorePairs(int index) const {
  return index < size();
}

const pair<string, string>& 
DN::getPairAt(int index) const {
  return pairs[index]; 
}

string 
DN::paramString() const {
  string params;

  params += '[';
  params += '"'; 
  params += toString();
  params += '"'; 
  params += ']';
  return params;
}

string
DN::toLogString() const {
  string log;
  
  log += "DN=";
  log += '"';
  log += toString();
  log += '"';
  return log;
}
