#ifndef AK_SIGNATURE_HPP
#define AK_SIGNATURE_HPP

#include "ref_ptr.hpp"
#include "SignatureImpl.hpp"
#include "Provider.hpp"

namespace Akenti {

/**
  * A <i>Signature</i> object is a <b>reference-count</b> pointer that wraps 
  * a <i>SignatureImpl</i> object. It provides applications, albeit indirectly,
  * with the functionality of a digital signature algorithm such as RSA-MD5 
  * or DSA-SHA1. Digital signatures are used for authentication and integrity 
  * assurance of digital data.
  * 
  * <PRE>
  * <b>Usage:</b>
  *      try {
  *        Signature sig = Signature::getInstance("RSA-MD5");
  *
  *        sig->update("Date");
  *        sig->update("to");
  *        sig->update("be signed");
  *
  *        string signature = sig->sign();
  *      } catch(SecurityException e) { ... }
  * </PRE>
  *
  * @see ref_ptr
  * @see SignatureImpl
  * @see Provider#getProvider()
  * @see ProviderImpl#getSignature(const string& alg)
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

class Signature : public ref_ptr<SignatureImpl> {

public:

   /**
     * Contructs a <i>Signature</i> object given a <i>SignatureImpl</i>
     * object. Typically the <i>getInstance(const string&)</i> method is
     * used to create an instance of this class.
     *
     * @param impl SignatureImpl wrapped object
     * @see ref_ptr
     * @see SignatureImpl
     * @see getInstance(const string&) 
     */
   Signature(SignatureImpl *impl = NULL) : ref_ptr<SignatureImpl> (impl) {
   }

  /**
    * Decrements the reference_count and when it goes to zero destroys
    * the object.
    *
    * @see ref_ptr
    */
  ~Signature() {
   }

  /**
    * Returns a <i>Signature</i> object of the type  set in the security
    * provider for the type of algorithm specified by <i>alg</i>. 
    *
    * @param alg string name of the algorithm
    * @exception Error if a provider has not been installed
    * @exception SecurityException if algorithm is not supported
    *
    * @see ProviderImpl
    * @see isSignatureAlgorithmSupported(const string&)
    */
  static Signature getInstance(const string& alg) {
    Provider& po = Provider::getProvider();

    return po->getSignature(alg);
  }

  /**
    * Returns <i>true</i> if the signature algorithm specified  
    * by <i>alg</i> is supported.
    *
    * @param alg string name of the algorithm
    * @exception Error if a provider has not been installed
    */
  static bool isSignatureAlgorithmSupported(const string& alg) {
    Provider& po = Provider::getProvider();
    
    return po->isSignatureAlgorithmSupported(alg);
  }

};

} // end of namespace Akenti

#endif
