#include <sstream> 
#include <map> 
#include "Exception.hpp"
#include "AkentiMessage.hpp"
#include "Xerces-Includes.hpp"
#include "XMLUtil.hpp"

using std::istringstream;
using std::ostringstream;
using std::map;
using namespace Akenti;

typedef AkentiMessage AM;

namespace Akenti {

static map<int, string> codes;

class CodesLoader {

public:

  CodesLoader() {
    /* 100-199 Ok messages */
    codes[101] = "collecting policies ok"; // PolicyCollector
    codes[102] = "verifying policies ok";  // PolicyCollector

    codes[103] = "use condition certs verified"; // UseCondCollector
    
    codes[104] = "access granted"; // PolicyAuthorizer
    codes[105] = "further checks must be made to resolve ConditionalActions"; // PolicyAuthorizer
    codes[106] = "User cannot do all desired actions";// PolicyAuthorizer
    codes[107] = "User will never be able to do all desired actions"; //PolicyAuthorizer

    codes[140] = "stakeholder is accpetable"; // AkentiEngine 

    codes[150] = "Policy is OK"; // PolicyChecker
    codes[151] = "Policy is OK but some CAs are bad"; // PolicyChecker

    codes[161] = "Resource Browse is OK"; // ResourceBrowse
    codes[162] = "Resource Browse is OK, listing root resources"; // ResourceBrowse

    codes[171] = "Create Resource is OK"; // CreateResource

    codes[181] = "Upload Certificate is OK"; // UploadCertificate

    codes[191] = "Delete Certificate is OK"; // DeleteCertificate

    codes[196] = "Get Resource Info is OK"; // GetResourceInfo

  /* 200-249 Problems that the client finds : cannot connect to server */
    codes[200] = "Illegal argument";
    codes[201] = "Connection Error"; 
    codes[202] = "Soap Fault"; 

    codes[205] = "Exception thrown:";
    codes[206] = "Error thrown:";
    codes[207] = "Unknown error thrown";

  /* 250-299 Problems that server finds: wrong request, bad credential
   *         - user needs to fix something */

  /* 300-399 Permission denied - errors from the Authorizer   
   *         - maps to user as permission denied,  */
    codes[301] = "did not satisfy critical UC"; // PolicyAuthorizer
    codes[302] = "User is allowed no actions"; // PolicyAuthorizer
    codes[306] = "cannot verify user's x509 cert"; // AkentiEngine

    codes[350] = "not an acceptale stakeholder"; // AkentiEngine
    codes[351] = "cannot verify stakeholder's x509 cert"; // AkentiEngine

    codes[380] = "request denied"; // ProtocolHandler 

  /* 400-499 Problems in policy collection and verififcation  */

    codes[401] = "parsing resource name failed"; // PolicyCollector
    codes[402] = "cannot find rootPolicyPath in config file"; // PolicyCollector
    codes[403] = "RootPolicyPath found in config is not absolute"; //PolicyCollector
    codes[404] = "RootPolicyCertificate does not exist"; // PolicyCollector
    codes[405] = "unable to read root policy"; // PolicyCollector
    codes[406] = "wrong root resource name";  // PolicyCollector
    codes[407] = "no root cas in root policy cert"; // PolicyCollector 
    codes[430] = "unable to read policy";  // PolicyCollector
    codes[431] = "wrong resource name"; // PolicyCollector
    codes[432] = "missing resource"; // PolicyCollector
    codes[450] = "failed verifying policies"; // PolicyCollector

    codes[460] = "All the CAs are unacceptable"; //PolicyChecker
    codes[470] = "All UC issuers are unacceptable in group"; // PolicyChecker
    
  /* 500-599 Problems in UseCondition collection and verification
   *         - both map to user as policy is inconsistant */
    codes[501] = "failed to find uc certs";     // UseCondCollector

   /* 600-699 ?? */
    codes[601] = "failed to satisfy use conds for a policy"; // PolicyAuthorizer
    codes[602] = "user did not satisfy any use condition for all policies"; // PolicyAuthorizer

    /*700-799 Problems with ResourceBrowse, Create Resource, Upload Certificate, 
     Delete Certificate, and Get Resource Info*/
    codes[701] = "parsing resource name failed"; // ResourceBrowse
    codes[702] = "cannot find RootPath in config file"; // ResourceBrowse
    codes[703] = "RootPath found in config is not absolute"; // ResourceBrowse
    codes[704] = "resource does not exist"; // ResourceBrowse

    codes[711] = "parsing resource name failed"; // CreateResource
    codes[712] = "cannot find RootPath in config file"; // CreateResource
    codes[713] = "RootPath found in config is not absolute"; // CreateResource
    codes[714] = "resource already exists"; // CreateResource
    codes[715] = "resource name already in use"; // CreateResource
    codes[716] = "root resource cannot be created through a request"; // CreateResource
    codes[717] = "resource could not be created"; // CreateResource

    codes[721] = "parsing resource name failed"; // UploadPolicyCertificate
    codes[722] = "cannot find RootPath in config file"; // UploadPolicyCertificate
    codes[723] = "RootPath found in config is not absolute"; // UploadPolicyCertificate
    codes[724] = "resource does not exist"; // UploadPolicyCertificate

    codes[731] = "parsing resource name failed"; // DeletePolicyCertificate
    codes[732] = "cannot find RootPath in config file"; // DeletePolicyCertificate
    codes[733] = "RootPath found in config is not absolute"; // DeletePolicyCertificate
    codes[734] = "resource policy certificate does not exist"; // DeletePolicyCertificate

    codes[741] = "parsing resource name failed"; // GetResourceInfo
    codes[742] = "root resource does not exist"; // GetResourceInfo
    codes[743] = "root resource is not absolute in config file"; // GetResourceInfo
    codes[744] = "resource does not exist"; // GetResourceInfo
    codes[745] = "there were errors reading the hints file"; // GetResourceInfo
    codes[746] = "no hints file found"; // GetResourceInfo


  /* 900-999 Programming errors exceptions and errors caught
   *         - maps to user as code is wrong */
  }

} codesLoader;

}

AM::AkentiMessage()
  : code(-1), message("NO_MESSAGE") {
}

AM::AkentiMessage(int code)
  : code(code), message(codes[code]) {
}

AM::AkentiMessage(int code, const string& message)
  : code(code), message(codes[code] + message) {
}

AM::~AkentiMessage() {
}

int
AM::getCode() const {
  return code;
}

const string& 
AM::getMessage() const {
  return message;
}

string 
AM::toString() const {
  ostringstream os;

  os << code
     << ' ' 
     << message;

  return os.str();
}

bool 
AM::operator==(const AkentiMessage& am) const {
  return 
    getCode() == am.getCode()
	  && 
    getMessage() == am.getMessage();
}

void
AM::toXML(DOMElement* mesg) const {
  DOMDocument* doc = mesg->getOwnerDocument();
  DOMElement* temp = doc->createElement(XMLUtil::convert("Code"));

  ostringstream oss;
  oss << code;

  temp->appendChild(doc->createTextNode(XMLUtil::convert(oss.str().c_str())));
  mesg->appendChild(temp);

  temp = doc->createElement(XMLUtil::convert("Status"));

  temp->appendChild(doc->createTextNode(XMLUtil::convert(message.c_str())));
  mesg->appendChild(temp);
}

void 
AM::fromXML(const DOMElement* mesg) {
  DOMNode* child = mesg->getFirstChild();

  if (child == NULL || child->getNodeType() != DOMNode::ELEMENT_NODE) {
    throw Exception("AkentiMessage:fromXML:code missing");
  }

  DOMNode* node = child->getFirstChild();

  if (node == NULL) {
    throw Exception("AkentiMessage:doFromXML:code missing");
  }

  string str(XMLUtil::convert(node->getNodeValue()));
  istringstream iss(str);

  iss >> code;

  child = child->getNextSibling();

  if (child == NULL || child->getNodeType() != DOMNode::ELEMENT_NODE) {
    throw Exception("AkentiMessage:fromXML:message missing");
  }

  node = child->getFirstChild();

  if (node == NULL) {
    throw Exception("AkentiMessage:doFromXML:message missing");
  }

  message = XMLUtil::convert(node->getNodeValue());
}
