#include <iostream>
#include <sstream>
#include <fstream>
#include <memory>
#include <sys/time.h>
#include <NetLogger.h>
#include <logging.h>        // for NL_APPEND
#include "Logger.hpp"
#include "Properties.hpp"
#include "Error.hpp"
#include "File.hpp"

extern "C" {
  int getpid();
}

using std::auto_ptr;

namespace Akenti {

static Logger* DEFAULT_LOGGER = new Logger();
static auto_ptr<Logger> DEFAULT_LOGGER_PTR(DEFAULT_LOGGER);

// map to hold loggers.
static map<int, Logger*> loggers;

static int 
getThreadID() {
  return 0;
}

static Logger*
getLogger() {
  Logger* logger = loggers[getThreadID()];

  return logger == NULL ? DEFAULT_LOGGER : logger;
}

static void 
add(Logger* logger) {
  loggers[getThreadID()] = logger;
}
 
static void 
remove(Logger* logger) {
  if (logger != DEFAULT_LOGGER) {
    loggers.erase(getThreadID());
  }
}

Logger::~Logger() {
  remove(this);
}

int 
Logger::doWrite(int, const string&, const string&, const string&) {
  return 0;
}

int 
Logger::doSetLevel(int level) {
  return 0;
}

void  
Logger::doFlush() {
}

class DLogger : public Logger {

public:

  DLogger(NetLogger* delegate) : delegate(delegate) {
  }

  ~DLogger() {
    string tempFile = delegate->getFileName();

    try { 
      delegate->close(); 
      delete(delegate);
    } catch (exception& ex) { 
      delete(delegate);
      return;
    }
    
    if (!tempFile.length()) {
      return;
    }

    Properties& config = Properties::getSystemProperties();
    string permFile = config.getValue("MonitorLogFile", "akentiLogs");
    std::ifstream is((char*) tempFile.c_str());
    std::ofstream os((char*) permFile.c_str(), std::ios::app);

    if (!is) {
      return;
    }
      /* XXXX the lock and unlock methods no longer exist in g++ 3.0
	 so let the bytes fall where they may

	 else if (!lock(os)) { return;}
      */

    string line;

    os << "************************************" << '\n';

    while (std::getline(is, line)) {
      os << line << '\n';
    }

    os << "************************************" << '\n';

    os.close();
    is.close();

    File(tempFile).remove();

    /* XXXX
      if (!unlock(os)) { return; }
    */ 

    return;
  }

  int doWrite(int level, 
              const string& key, 
              const string& levelMessage, 
              const string& message) {
    try { 
      delegate->write(level, key, levelMessage + ' ' + message + '\n');
      return 0;
    } catch(...) {
      return -1;
    }
  }

  int getLevel() {
    return delegate->getLevel();
  }

  int doSetLevel(int level) {
    int old = delegate->getLevel();

    delegate->setLevel(level);
    return old;
  }

  void 
  doFlush() {
    std::ofstream os(".globus/xxxx", std::ios::app);

    os << "FLUSHING" << delegate->flush() << std::endl;
    os.close();
    delegate->flush();
  }

private: 
  
  NetLogger* delegate;

};

bool 
Logger::hasBeenInitialized() {
  return loggers[getThreadID()] != NULL;
}


Logger*
Logger::init(const string& id, string& message) {
  return init(id, Properties::getSystemProperties(), message);
}

Logger* 
Logger::init(const string& id, const Properties& config, string& message) {
  if (Logger::hasBeenInitialized()) {
    throw Error("Logger:init:logger already initialized");
  }

  string type = config.getValue("Logging", "off");

  if (type == "off" || type == "OFF") {
    Logger* logger = new Logger();

    if (!logger) {
      throw Error("Logger:init:bad alloc(1)");
    }

    add(logger);
    return logger;
  } 
  
  NetLogger* delegate = new NetLogger();

  if (!delegate) {
    throw Error("Logger:init:bad alloc(2)");
  }

  string quotedID  = "\"" + id + "\"";

  try {
    if (type == "server" || type == "SERVER") {
      string server = config.getValue("MonitorServer", "localhost");
      string port = config.getValue("MonitorPort", "10000");
    
      message = "opening connection to server " + server + " at port" + port;
      delegate->open(quotedID, "x-netlog://" + server + ":" + port, NL_APPEND);
    } else if (type == "file" || type == "FILE") {
      File dir(config.getValue("TempLogDir", "/tmp"));
      std::ostringstream os;
      struct timeval tv;

      gettimeofday(&tv, 0);

      os << dir.getAbsolutePath() 
         << '/' << "aklogs" 
         << '.' << getpid() 
         << '.' << tv.tv_usec;

      string file = os.str();

      message = "opening file " + file;
      delegate->open(quotedID, "file://" + file, NL_APPEND);  
    } else if (type == "stdout" || type == "STDOUT") {
      message = "opening stdout";
      delegate->open(quotedID, 1, NL_APPEND);
    } else if (type == "stderr" || type == "STDERR") {
      message = "opening stderr";
      delegate->open(quotedID, 2, NL_APPEND);
    } else {
      message = "invalid logging type " + type;
      message += " opening stderr";
      delegate->open(quotedID, 2, NL_APPEND);
    }

    message += " succeded";
  } catch(...) {
    delete(delegate);
    message += " failed,";
    return NULL;
  }

  // LOG_AUD = 1 
  std::istringstream is(config.getValue("LogLevel", "1"));
  int level = LOG_AUD;
  
  if (!(is >> level) || level < LOG_AUD) {
    level = LOG_AUD;
  }

  delegate->setLevel(level);
  
  DLogger* logger = new DLogger(delegate);

  if (!logger) {
    delegate->close();
    delete(delegate);
    throw Error("Logger:init:bad alloc(4)");
  }

  add(logger);
  return logger;
}

int
Logger::write(int level, const string& key, const string& message) { 
  string temp;

  switch(level) {
    case LOG_AUD:
      temp = "LVL=AUDIT";
      break;
    case LOG_MON: 
      temp = "LVL=MONITOR";
      break;
    case LOG_L3: 
      temp = "LVL=DEBUG1";
      break;
    case LOG_L4: 
      temp  = "LVL=DEBUG2";
      break;
    case LOG_L5: 
      temp = "LVL=DEBUG3";
      break;
    default: 
      temp = "LVL=UKNOWN";
      break;
  }

  return getLogger()->doWrite(level, key, temp, message);
}

int
Logger::setLevel(int level) {
  return getLogger()->doSetLevel(level);
}

void Logger::flush() {
  getLogger()->doFlush();
}

} // end of namespace Akenti

