#include <iostream>
#include <sstream>
#include <string>
#include <stdlib.h>
#include "Xerces-Includes.hpp"
#include <stdlib.h>
#include "Exception.hpp"
#include "Error.hpp"
#include "InputStream.hpp"
#include "XMLUtil.hpp"
#include "File.hpp"
#include "URL.hpp"
#include "AkentiConstants.hpp"

#define DEBUG 0 

using namespace std;

namespace Akenti {

typedef XMLUtil XU;

XStr::XStr(const char* const toTranscode) {
  if (!toTranscode) {
    throw Error("XStr:XStr:illegal argument");
  } 

  unicodeForm = XMLString::transcode(toTranscode, XMLPlatformUtils::fgMemoryManager);

  if (!unicodeForm) {
     throw Error("XStr:XStr:transcode failed");
  }
}

XStr::~XStr() {
  XMLPlatformUtils::fgMemoryManager->deallocate(unicodeForm);
  unicodeForm = NULL;
}

XStr::operator const XMLCh*() const {
  return unicodeForm;
}

class MyDOMPrintFormatTarget : public XMLFormatTarget {

public:

  MyDOMPrintFormatTarget(ostream& os) : os(os) {
  }

  void writeChars(const XMLByte* const  toWrite,
                  const unsigned int  count,
                  XMLFormatter * const formatter) {
    os.write((char *) toWrite, (int) count);
  }

private:

  ostream& os;

};

class STDBinInputStream : public BinInputStream {

public:

  STDBinInputStream(istream& is) : is(is) {
  }

  ~STDBinInputStream() {
  }

  unsigned int curPos() const {
    throw Error("STDBinInputStream:curPos:Unsupported");
  }

  unsigned int readBytes(XMLByte* const toFill , const unsigned int maxToRead) {
    is.read((char *) toFill, maxToRead);
    return is.gcount();
  }

private:

istream& is;

};

class STDInputSource : public InputSource {

public:

  STDInputSource(istream& is) : is(is) {
  }

  ~STDInputSource() {
  }

  BinInputStream* makeStream() const {
    return new STDBinInputStream(is);
  }

private:

istream& is;

};

class DOMTreeErrorReporter : public ErrorHandler {
 
public:
 
  DOMTreeErrorReporter() {
  }
 
  void warning(const SAXParseException& toCatch) {
    os << "Parsing Warning at line "
       << toCatch.getLineNumber()
       << ", column " << toCatch.getColumnNumber()
       << "(" << XMLUtil::convert(toCatch.getMessage())
       << ")\n";
  }
 
  void DOMTreeErrorReporter::error(const SAXParseException& toCatch) {
    os << "Parsing Error at line "
       << toCatch.getLineNumber()
       << ", column " << toCatch.getColumnNumber()
       << "(" << XMLUtil::convert(toCatch.getMessage())
       << ")\n";
  }
 
  void DOMTreeErrorReporter::fatalError(const SAXParseException& toCatch) {
    os << "Parsing Fatal Error at line "
       << toCatch.getLineNumber()
       << ", column " << toCatch.getColumnNumber()
       << "(" << XMLUtil::convert(toCatch.getMessage())
       << ")\n";
  }

  void resetErrors() {
    throw Error("XMLUtil:DOMTreeErrorReporter:reset:notimplemented");
  }

  string getReport() const {
    return os.str();
  }

private:
 
 ostringstream os;
 
};

static string 
getSchemaLocation() {
  string temp;
  char* var = NULL;
  int idx = 0;

  while (NAME_SPACES_MAP[idx][0]) {
    string ns = NAME_SPACES_MAP[idx][0]; 
    string loc = NAME_SPACES_MAP[idx++][1]; 

#if DEBUG
    cerr << ns << endl;
    cerr << loc << endl;
#endif 
    
    if (loc.find("http:") == string::npos) {
      if (loc.find("file:") == 0) {
        loc = URL(loc).getFile();
      }

      File file(loc);

      if (!file.isAbsolute()) {
        if (var == NULL) { 
          if ((var = getenv("AKENTI_HOME")) == NULL) {
            throw Exception("AKENTI_HOME is not set and $AKENTI_HOME/xml/schema/" +
                            (" must contain schema " + loc));
          } 
        }

        file = File(loc = var + ("/xml/schema/" + loc));
      }

      if (!file.exists() || file.isDirectory()) {
        throw Exception("XMLUtil:Can't open Schema file at " +  file.getAbsolutePath());
      }
    }

    if (temp.length() != 0) {
      temp += ' ';
    }

    temp += ns + ' ' + loc; 
  }

  return temp;
}

static bool INIT = false;
static string SCHEMA_LOCATION;

void 
XU::init() { 
  try {
    if (!INIT) {
      SCHEMA_LOCATION = getSchemaLocation();

#if DEBUG
     cerr << SCHEMA_LOCATION << endl;
#endif

      XMLPlatformUtils::Initialize();
      INIT = true;
    }
  } catch(XMLException& ex) {
    throw Exception("XMLUtil:init:failed");
  }
}

auto_ptr<AbstractDOMParser> 
XU::parse(istream& is) {
  if (!INIT) {
    throw Error("XMLUtil:convert:must call XMLUtil::init");
  }

  try {
    auto_ptr<XercesDOMParser> parser(new XercesDOMParser);
    STDInputSource mis(is);
    DOMTreeErrorReporter errReporter;

    parser->setErrorHandler(&errReporter);
    parser->setValidationScheme(XercesDOMParser::Val_Always);
    parser->setDoNamespaces(true);
    parser->setExternalSchemaLocation(SCHEMA_LOCATION.c_str());
    parser->setDoSchema(true);
    parser->setValidationSchemaFullChecking(true);
    parser->parse(mis);

    if (parser->getErrorCount() > 0) {
      throw Exception("XMLUtil:parse:Errors occured while parsing\n" + errReporter.getReport());
    }   

    return (auto_ptr<AbstractDOMParser>&) parser;
  } catch (const DOMException& ex) {
    throw Exception("XMLUtil:parse:Unknown DOMException");
  } catch (const XMLException& e) {
    throw Exception("XMLUtil:parse:" + convert(e.getMessage()));
  }
}  

auto_ptr<AbstractDOMParser> 
XU::parse(InputStream& is) {
  return parse(*(is.get()));
}

XStr
XU::convert(const char* ch) {
  if (!INIT) {
    throw Error("XMLUtil:convert:must call XMLUtil::init");
  }

  return XStr(ch);
}

string 
XU::convert(const XMLCh* s) {
  if (!INIT) {
    throw Error("XMLUtil:convert:must call XMLUtil::init");
  }

  static auto_ptr<XMLLCPTranscoder> coder(XMLPlatformUtils::fgTransService->makeNewLCPTranscoder());

  if (!coder.get()) {
    throw Exception("XMLUtil:convert:no XMLLCPTranscoder");
  }

  int len = coder->calcRequiredSize(s);
  string rep(len, '\0');

  if (coder->transcode(s, (char *) rep.data(), len) == 0) {
    throw Exception("XMLUtil:convert:failed");
  }

  return rep;
}

void
XU::filter(DOMNode* n) {
  if (!INIT) {
    throw Error("XMLUtil:convert:must call XMLUtil::init");
  }

  DOMNode *c = n->getFirstChild();
  
  while (c != NULL) {
    DOMNode *temp = c;

    c = temp->getNextSibling();
    
    if (temp->getNodeType() == DOMNode::TEXT_NODE) { 
      if (XMLString::isAllWhiteSpace(temp->getNodeValue())) {
        n->removeChild(temp);
      } 
    } else {
      filter(temp);
    }
  }
}  

static const XMLCh  gEndElement[] = { chOpenAngle, chForwardSlash, chNull };
static const XMLCh  gEndPI[] = { chQuestion, chCloseAngle, chNull};
static const XMLCh  gStartPI[] = { chOpenAngle, chQuestion, chNull };
static const XMLCh  gXMLDecl1[] =
{
        chOpenAngle, chQuestion, chLatin_x, chLatin_m, chLatin_l
    ,   chSpace, chLatin_v, chLatin_e, chLatin_r, chLatin_s, chLatin_i
    ,   chLatin_o, chLatin_n, chEqual, chDoubleQuote, chNull
};
static const XMLCh  gXMLDecl2[] =
{
        chDoubleQuote, chSpace, chLatin_e, chLatin_n, chLatin_c
    ,   chLatin_o, chLatin_d, chLatin_i, chLatin_n, chLatin_g, chEqual
    ,   chDoubleQuote, chNull
};
static const XMLCh  gXMLDecl3[] =
{
        chDoubleQuote, chSpace, chLatin_s, chLatin_t, chLatin_a
    ,   chLatin_n, chLatin_d, chLatin_a, chLatin_l, chLatin_o
    ,   chLatin_n, chLatin_e, chEqual, chDoubleQuote, chNull
};

static const XMLCh  gXMLDecl4[] =
{
        chDoubleQuote, chQuestion, chCloseAngle
    ,   chLF, chNull
};

static const XMLCh  gStartCDATA[] =
{
        chOpenAngle, chBang, chOpenSquare, chLatin_C, chLatin_D,
        chLatin_A, chLatin_T, chLatin_A, chOpenSquare, chNull
};

static const XMLCh  gEndCDATA[] =
{
    chCloseSquare, chCloseSquare, chCloseAngle, chNull
};
static const XMLCh  gStartComment[] =
{
    chOpenAngle, chBang, chDash, chDash, chNull
};

static const XMLCh  gEndComment[] =
{
    chDash, chDash, chCloseAngle, chNull
};

static const XMLCh  gStartDoctype[] =
{
    chOpenAngle, chBang, chLatin_D, chLatin_O, chLatin_C, chLatin_T,
    chLatin_Y, chLatin_P, chLatin_E, chSpace, chNull
};
static const XMLCh  gPublic[] =
{
    chLatin_P, chLatin_U, chLatin_B, chLatin_L, chLatin_I,
    chLatin_C, chSpace, chDoubleQuote, chNull
};
static const XMLCh  gSystem[] =
{
    chLatin_S, chLatin_Y, chLatin_S, chLatin_T, chLatin_E,
    chLatin_M, chSpace, chDoubleQuote, chNull
};
static const XMLCh  gStartEntity[] =
{
    chOpenAngle, chBang, chLatin_E, chLatin_N, chLatin_T, chLatin_I,
    chLatin_T, chLatin_Y, chSpace, chNull
};
static const XMLCh  gNotation[] =
{
    chLatin_N, chLatin_D, chLatin_A, chLatin_T, chLatin_A,
    chSpace, chDoubleQuote, chNull
};

XMLFormatter& operator<<(XMLFormatter& strm, const XMLCh* s) {
  unsigned int len = XMLString::stringLen(s);

  if (len <= 0) {
    return strm;
  }

  XMLCh*  buf = new XMLCh[len + 1];
  XMLString::copyNString(buf, s, len);
  buf[len] = 0;
  strm << *buf;
  delete [] buf;
  return strm;
}

void writeSpaces(XMLFormatter& formatter, int ident) {
  for (int i = 0; i < ident; ++i) {
    formatter << chSpace;
  }
}

void writeToStream(ostream& target, XMLFormatter& formatter, DOMNode* toWrite, int ident);

void
writeToStream(ostream& target, XMLFormatter& formatter, DOMDocument* toWrite, int ident) {
  DOMNode *child = toWrite->getFirstChild();

  while(child != 0) {
    writeToStream(target, formatter, child, ident);
    formatter << chLF;
    target << flush;
    child = child->getNextSibling();
  }
}

void
writeToStream(ostream& target, XMLFormatter& formatter, DOMText* toWrite) {
  const XMLCh   *nodeValue = toWrite->getNodeValue();
  unsigned long len = XMLString::stringLen(nodeValue);

  formatter.formatBuf(nodeValue, len, XMLFormatter::CharEscapes);
}

void
writeToStream(ostream& target, XMLFormatter& formatter, DOMElement* toWrite, int ident) {
  const XMLCh *nodeName = toWrite->getNodeName();

  writeSpaces(formatter, ident);

  formatter << XMLFormatter::NoEscapes << chOpenAngle;

  int len = XMLString::stringLen(nodeName);
  
  for (int i=0; i < len; i++)
    formatter << nodeName[i];

  DOMNamedNodeMap *attributes = toWrite->getAttributes();
  int attrCount = attributes->getLength();

  for (int i = 0; i < attrCount; i++) {
     DOMNode  *attribute = attributes->item(i);

     formatter  << XMLFormatter::NoEscapes
                << chSpace;
     
     len = XMLString::stringLen(attribute->getNodeName());

     for (int i=0; i < len; i++)
       formatter << attribute->getNodeName()[i];

     formatter << chEqual << chDoubleQuote
	       << XMLFormatter::AttrEscapes;

     len = XMLString::stringLen(attribute->getNodeValue());

     for (int i=0; i < len; i++)
       formatter << attribute->getNodeValue()[i];
     
     formatter << XMLFormatter::NoEscapes
	       << chDoubleQuote;
  }

  DOMNode *child = toWrite->getFirstChild();

  if (child != 0) {
    formatter << XMLFormatter::NoEscapes << chCloseAngle;

    bool textNode = child->getNodeType() == DOMNode::TEXT_NODE;

    if (!textNode) {
      formatter << chLF;
    }

    while (child != 0) {
      writeToStream(target, formatter, child, ident + 3);
      child = child->getNextSibling();
    }

    if (!textNode) {
      writeSpaces(formatter, ident);
    }
    formatter <<  XMLFormatter::NoEscapes;

    len = XMLString::stringLen(gEndElement);

    for (int i=0; i < len; i++)
      formatter << gEndElement[i];

    len = XMLString::stringLen(nodeName);

    for (int i=0; i < len; i++)
      formatter << nodeName[i];

    formatter << chCloseAngle;
  } else {
    formatter <<  XMLFormatter::NoEscapes << chForwardSlash << chCloseAngle;
  }

  formatter << chLF;   // ADDED
}

void 
writeToStream(ostream& target, XMLFormatter& formatter, DOMNode* toWrite, int ident) {
  switch (toWrite->getNodeType()) {
    case DOMNode::TEXT_NODE: 
      writeToStream(target, formatter, (DOMText*) toWrite);
      break;
    case DOMNode::DOCUMENT_NODE :
      writeToStream(target, formatter, (DOMDocument*) toWrite, ident);
      break;
    case DOMNode::ELEMENT_NODE : 
      writeToStream(target, formatter, (DOMElement*) toWrite, ident);
      break;
    default:
      cerr << "UNKNOWN TYPE " << endl;
      exit(1);
  }
}

void 
XU::write(ostream& target, DOMNode* toWrite) {
  if (!INIT) {
    throw Error("XMLUtil:convert:must call XMLUtil::init");
  }

  MyDOMPrintFormatTarget formatTarget(target);
  XMLFormatter formatter(
    XMLUtil::convert("UTF-8"), 
    XMLUtil::convert("1.0"),
    &formatTarget, 
    XMLFormatter::NoEscapes, 
    XMLFormatter::UnRep_CharRef 
  );

  writeToStream(target, formatter, toWrite, 0);
}

}
