#include <fstream>
#include <iostream>
#include <sstream>
#include "cppunit/TestCaller.h"
#include "cppunit/Exception.h"
#include "cppunit/Message.h"
#include "cppunit/SourceLine.h"

#include "CapabilityCertGeneratorTestCase.hpp"
#include "CapabilityCertGenerator.hpp"
#include "DistinguishedName.hpp"
#include "AttributeInfo.hpp"
#include "KeyFactory.hpp"
#include "KeyPair.hpp"
#include "PrivateKey.hpp"
#include "PublicKey.hpp"
#include "ConditionalActions.hpp"
#include "UtcTime.hpp"
#include "URL.hpp"
#include "File.hpp"
#include "SecurityConstants.hpp" 
#include "CertificateFactory.hpp"
#include "Certificate.hpp"
#include "../shared/TestCaseConstants.hpp"
#include "XMLUtil.hpp"

using namespace Akenti;
using namespace std;

typedef CapabilityCertGeneratorTestCase CCGTC;

void 
CCGTC::generateCapabilityCerts() {
  string method;

  try {
    method = "CapabilityCertGeneratorTestCase:generateCapabilityCert1:";
    generateCapabilityCert1();
    method = "CapabilityCertGeneratorTestCase:generateCapabilityCert2:";
    generateCapabilityCert2();
    method = "CapabilityCertGeneratorTestCase:generateCapabilityCert3:";
    generateCapabilityCert3();
    method = "CapabilityCertGeneratorTestCase:generateCapabilityCert4:";
    generateCapabilityCert4();
  } catch (Exception& ex) {
    CppUnit::Message m(method + ex.getMessage());

    throw CppUnit::Exception(m, CPPUNIT_SOURCELINE());
  } catch (Error& err) {
    CppUnit::Message m(method + err.getMessage());

    throw CppUnit::Exception(m, CPPUNIT_SOURCELINE());
  } 
}

void
CCGTC::generateCapabilityCert1() {
#if DEBUG
  cout << "\nGenerating Capability Certificate 1 ..." << endl;
#endif 

  UtcTime now;
  UtcTime after(365 * 24 * 60 * 60);

  DistinguishedName ca(CAIA_DN);
  DistinguishedName issuer(SHI_DN);
  AkentiPrincipal guarantor(issuer, ca);

  DistinguishedName user(USERI_DN);
  AkentiPrincipal principal(user, ca);

  CapabilityCertGenerator gen;

  gen.setVersionNumber(2);
  gen.setUniqueID("1");
  gen.setIssuer(guarantor);
  gen.setNotBefore(now);
  gen.setNotAfter(after);
  gen.setResource(ROOT1);
  gen.setPrincipal(principal);

  KeyFactory kfac = KeyFactory::getInstance();
  KeyPair pair = kfac->generateRSAKeyPair();

  gen.setPublicKey(pair.getPublicKey());
  gen.sign(pair.getPrivateKey(), SecurityConstants::RSA_MD5);

  AkentiCertificate cert = gen.generate();
   
  ostringstream oos;
  AkentiCertificate::writeToStream("PEM", oos, cert);
  istringstream iis(oos.str());
  AkentiCertificate certFromStream = AkentiCertificate::initFromStream("PEM", iis);
 
  CPPUNIT_ASSERT(cert == certFromStream);

  ostringstream oos2;
  AkentiCertificate::writeToStream("XML", oos2, cert);
  istringstream iis2(oos2.str());
  certFromStream = AkentiCertificate::initFromStream("XML", iis2);
 
  CPPUNIT_ASSERT(cert == certFromStream);

#if DEBUG
   cout << iis2.str() << endl;
#endif
}

void
CCGTC::generateCapabilityCert2() {
#if DEBUG
  cout << "\nGenerating Capability Certificate 2 ..." << endl;
#endif 

  UtcTime now;
  UtcTime after(365 * 24 * 60 * 60);

  DistinguishedName ca(CAIA_DN);
  DistinguishedName issuer(SHI_DN);
  AkentiPrincipal guarantor(issuer, ca);

  DistinguishedName user(USERI_DN);
  AkentiPrincipal principal(user, ca);
  vector<string> actions;
  string act1("write");
  string act2("execute");
  string act3("read");

  actions.push_back(act1);
  actions.push_back(act2);
  actions.push_back(act3);
  
  CapabilityCertGenerator gen;

  gen.setVersionNumber(2);
  gen.setUniqueID("1");
  gen.setIssuer(guarantor);
  gen.setNotBefore(now);
  gen.setNotAfter(after);
  gen.setResource(ROOT1);
  gen.setPrincipal(principal);
  gen.setUnConditionalActions(actions);

  KeyFactory kfac = KeyFactory::getInstance();
  KeyPair pair = kfac->generateRSAKeyPair();

  gen.setPublicKey(pair.getPublicKey());
  gen.sign(pair.getPrivateKey(), SecurityConstants::RSA_MD5);

  AkentiCertificate cert = gen.generate();
   
  /* Write the cert in PEM and XML formats to a stream then read from the stream
   * to make sure the values are the same
   */
   ostringstream oos;
   AkentiCertificate::writeToStream("PEM", oos, cert);
   istringstream iis(oos.str());
   AkentiCertificate certFromStream = AkentiCertificate::initFromStream("PEM", iis);
 
   CPPUNIT_ASSERT(cert == certFromStream);

   ostringstream oos2;
   AkentiCertificate::writeToStream("XML", oos2, cert);
   istringstream iis2(oos2.str());
   certFromStream = AkentiCertificate::initFromStream("XML", iis2);
 
   CPPUNIT_ASSERT(cert == certFromStream);

#if DEBUG
   cout << iis2.str() << endl;
#endif
}

void
CCGTC::generateCapabilityCert3() {
#if DEBUG
  cout << "\nGenerating Capability Certificate 3 ..." << endl;
#endif 

  UtcTime now;
  UtcTime after(365 * 24 * 60 * 60);

  DistinguishedName ca(CAIA_DN);
  DistinguishedName issuer(SHI_DN);
  AkentiPrincipal guarantor(issuer, ca);

  DistinguishedName user(USERI_DN);
  AkentiPrincipal principal(user, ca);
  string act1("read");
  string act2("write");
  string act3("execute");
  vector<string> condAct1Acts;
  vector<string> condAct2Acts;

  condAct1Acts.push_back(act3);
  condAct2Acts.push_back(act1);
  condAct2Acts.push_back(act2);

  vector<DistinguishedName> cas(1, ca);
  vector<AkentiPrincipal> attrIssuers(1, guarantor);

  string val = user.getValue("OU");
  string val2 = user.getValue("CN");

  AttributeInfo at1("OU", val, cas);
  AttributeInfo at2(true, "JOB", "SUPERVISOR", attrIssuers);
  AttributeInfo at3("CN", val2, cas);

  vector<AttributeInfo> attrInfos;
  vector<AttributeInfo> attrInfos2;

  attrInfos.push_back(at1); 
  attrInfos.push_back(at2);
  attrInfos2.push_back(at3);

  Condition cond1("OU =" + val + " && JOB = SUPERVISOR", attrInfos);
  Condition cond2("CN = " + val2, attrInfos2);

  vector<ConditionalActions> condActs;

  ConditionalActions condAct1(cond1, condAct1Acts);
  ConditionalActions condAct2(cond2, condAct2Acts, true);

  condActs.push_back(condAct1);
  condActs.push_back(condAct2);

  CapabilityCertGenerator gen;

  gen.setVersionNumber(2);
  gen.setUniqueID("1");
  gen.setIssuer(guarantor);
  gen.setNotBefore(now);
  gen.setNotAfter(after);
  gen.setResource(ROOT1);
  gen.setPrincipal(principal);
  gen.setConditionalActions(condActs);

  KeyFactory kfac = KeyFactory::getInstance();
  KeyPair pair = kfac->generateRSAKeyPair();

  gen.setPublicKey(pair.getPublicKey());
  gen.sign(pair.getPrivateKey(), SecurityConstants::RSA_MD5);

  AkentiCertificate cert = gen.generate();
   
  /* Write the cert in PEM and XML formats to a stream then read from the stream
   * to make sure the values are the same
   */
   ostringstream oos;
   AkentiCertificate::writeToStream("PEM", oos, cert);
   istringstream iis(oos.str());
   AkentiCertificate certFromStream = AkentiCertificate::initFromStream("PEM", iis);
 
   CPPUNIT_ASSERT(cert == certFromStream);

   ostringstream oos2;
   AkentiCertificate::writeToStream("XML", oos2, cert);
   istringstream iis2(oos2.str());
   certFromStream = AkentiCertificate::initFromStream("XML", iis2);
 
   CPPUNIT_ASSERT(cert == certFromStream);

#if DEBUG
   cout << iis2.str() << endl;
#endif
}

void
CCGTC::generateCapabilityCert4() {
#if DEBUG
  cout << "\nGenerating Capability Certificate 4 ..." << endl;
#endif 

  UtcTime now;
  UtcTime after(365 * 24 * 60 * 60);

  DistinguishedName ca(CAIA_DN);
  DistinguishedName issuer(SHI_DN);
  AkentiPrincipal guarantor(issuer, ca);

  DistinguishedName user(USERI_DN);
  AkentiPrincipal principal(user, ca);
  string act1("read");
  string act2("write");
  string act3("execute");

  vector<string> unCondActs;

  unCondActs.push_back(act1);
  unCondActs.push_back(act2);

  vector<string> condAct1Acts;
  vector<string> condAct2Acts;

  condAct1Acts.push_back(act3);
  condAct2Acts.push_back(act1);
  condAct2Acts.push_back(act2);

  vector<DistinguishedName> cas(1, ca);
  vector<AkentiPrincipal> attrIssuers(1, guarantor);

  string val = user.getValue("OU");
  string val2 = user.getValue("CN");

  AttributeInfo at1("OU", val, cas);
  AttributeInfo at2(true, "JOB", "SUPERVISOR", attrIssuers);
  AttributeInfo at3("CN", val2, cas);

  vector<AttributeInfo> attrInfos;
  vector<AttributeInfo> attrInfos2;

  attrInfos.push_back(at1); 
  attrInfos.push_back(at2);
  attrInfos2.push_back(at3);

  Condition cond1("OU =" + val + " && JOB = SUPERVISOR", attrInfos);
  Condition cond2("CN = " + val2, attrInfos2);

  vector<ConditionalActions> condActs;

  ConditionalActions condAct1(cond1, condAct1Acts);
  ConditionalActions condAct2(cond2, condAct2Acts, true);

  condActs.push_back(condAct1);
  condActs.push_back(condAct2);

  CapabilityCertGenerator gen;

  gen.setVersionNumber(2);
  gen.setUniqueID("1");
  gen.setIssuer(guarantor);
  gen.setNotBefore(now);
  gen.setNotAfter(after);
  gen.setResource(ROOT1);
  gen.setPrincipal(principal);
  gen.setUnConditionalActions(unCondActs);
  gen.setConditionalActions(condActs);

  KeyFactory kfac = KeyFactory::getInstance();
  KeyPair pair = kfac->generateRSAKeyPair();

  gen.setPublicKey(pair.getPublicKey());
  gen.sign(pair.getPrivateKey(), SecurityConstants::RSA_MD5);

  AkentiCertificate cert = gen.generate();
   
  /* Write the cert in PEM and XML formats to a stream then read from the stream
   * to make sure the values are the same
   */
   ostringstream oos;
   AkentiCertificate::writeToStream("PEM", oos, cert);
   istringstream iis(oos.str());
   AkentiCertificate certFromStream = AkentiCertificate::initFromStream("PEM", iis);
 
   CPPUNIT_ASSERT(cert == certFromStream);

   ostringstream oos2;
   AkentiCertificate::writeToStream("XML", oos2, cert);
   istringstream iis2(oos2.str());
   certFromStream = AkentiCertificate::initFromStream("XML", iis2);
 
   CPPUNIT_ASSERT(cert == certFromStream);

#if DEBUG
   cout << iis2.str() << endl;
#endif
}

CppUnit::TestSuite*
CCGTC::suite() {
  CppUnit::TestSuite* testSuite = new CppUnit::TestSuite("CapabilityCertGeneratorTestCase");

  testSuite->addTest(
    new CppUnit::TestCaller<CapabilityCertGeneratorTestCase> (
      "generateCapabilityCerts",
      &CapabilityCertGeneratorTestCase::generateCapabilityCerts
    )
  );

  return testSuite;
}
