#include <vector>
#include <iostream>
#include <fstream>
#include <sstream>
#include "cppunit/TestCaller.h"
#include "cppunit/Exception.h"
#include "cppunit/Message.h"
#include "cppunit/SourceLine.h"

#include "../shared/TestCaseConstants.hpp"
#include "UseCondCertGeneratorTestCase.hpp"
#include "DistinguishedName.hpp"
#include "AttributeInfo.hpp"
#include "PrivateKey.hpp"
#include "KeyFactory.hpp"
#include "KeyPair.hpp"
#include "UseCondCertGenerator.hpp"
#include "Condition.hpp"
#include "Util.hpp"
#include "SecurityConstants.hpp"
#include "File.hpp"
#include "Exception.hpp"
#include "Error.hpp"

using namespace std;
using namespace Akenti;

typedef UseCondCertGeneratorTestCase UCGTC;

void 
UCGTC::generateUseCondCerts() {
  string method; 

  try {
    method = "UseCondCertGeneratorTestCase:generateROOT1UseCondCert:";  
    generateROOT1UseCondCert();
    method = "UseCondCertGeneratorTestCase:generateROOT2UseCondCert::";  
    generateROOT2UseCondCert();
    method = "UseCondCertGeneratorTestCase:generateSUB1UseCondCert:";  
    generateSUB1UseCondCert();
    method = "UseCondCertGeneratorTestCase:generateSUB3UseCondCert:";  
    generateSUB3UseCondCert();
    method = "UseCondCertGeneratorTestCase:generateSUB4UseCondCert:";  
    generateSUB4UseCondCert();
    method = "UseCondCertGeneratorTestCase:generateROOT3UseCondCert:";  
    generateROOT3UseCondCert1();
    method = "UseCondCertGeneratorTestCase:generateROOT3UseCondCert1:";  
    generateROOT3UseCondCert2();
    method = "UseCondCertGeneratorTestCase:generateROOT4UseCondCert:";  
    generateROOT4UseCondCert();
  } catch (Exception& ex) {
    CppUnit::Message m(method + ex.getMessage());

    throw CppUnit::Exception(m, CPPUNIT_SOURCELINE());
  } catch (Error& err) {
    CppUnit::Message m(method + err.getMessage());

    throw CppUnit::Exception(m, CPPUNIT_SOURCELINE());
  }
}

void
UCGTC::generateROOT1UseCondCert() {
#if DEBUG
  cout << "\nGenerating ROOT1 UseCondition Certificate ..." << endl;
#endif

  UtcTime now;
  UtcTime after(365 * 24 * 60 * 60);

  DistinguishedName ca(CAIA_DN);
  DistinguishedName issuer(SHI_DN);
  AkentiPrincipal guarantor(issuer, ca);
  vector<string> actions(1, "READ");

  UseCondCertGenerator gen;

  gen.setVersionNumber(2);
  gen.setUniqueID("UC8");
  gen.setNotBefore(now);
  gen.setNotAfter(after);
  gen.setIssuer(guarantor);
  gen.setResource(ROOT1);
  gen.setLocalScope(false);
  gen.setCritical(false);
  gen.setActions(actions);
  
  KeyFactory kfac = KeyFactory::getInstance();
  PrivateKey privKey  = kfac->initPrivateKeyFromFile(
    "PEM", ID_DIR + SHI_KEY
  );

  gen.sign(privKey, SecurityConstants::RSA_MD5);

  AkentiCertificate cert = gen.generate();
  string fileName = AKCERT_DIR + ROOT1 + ".UC." + Util::createFileName(cert->hashCode()) + ".0";
  ofstream os(fileName.c_str());

  AkentiCertificate::writeToStream(AK_CERT_ENC_TYPE, os, cert);
  os.close();
  CPPUNIT_ASSERT(!os.fail());

{
  ostringstream os;
  os << cert->toText() << endl;
  AkentiCertificate::writeToStream("PEM", os, cert);
  istringstream is(os.str());
  AkentiCertificate certFromStream = AkentiCertificate::initFromStream("PEM", is);
  CPPUNIT_ASSERT(cert == certFromStream);
}

{
  ostringstream os;
  AkentiCertificate::writeToStream("XML", os, cert);
  istringstream is(os.str());
  AkentiCertificate certFromStream = AkentiCertificate::initFromStream("XML", is);
  CPPUNIT_ASSERT(cert == certFromStream);
}

#if DEBUG
  cout << "\nSaved ROOT1 UseCondition Certificate in " << fileName << endl;
#endif
}

void
UCGTC::generateROOT2UseCondCert() {
#if DEBUG
  cout << "\n Generating ROOT2 UseCondition Certificate ..." << endl;
#endif
  UtcTime now;
  UtcTime after(365 * 24 * 60 * 60);

  DistinguishedName ca(CAIA_DN);
  DistinguishedName issuer(SHI_DN);
  AkentiPrincipal guarantor(issuer, ca);

  vector<DistinguishedName> cas(1, ca);
  AttributeInfo at1("OU", "LBL", cas);
  vector<AkentiPrincipal> attrIssuers(1, guarantor);
  AttributeInfo at2(true, "JOB", "SUPERVISOR", attrIssuers);
  vector<AttributeInfo> attrInfos; 

  attrInfos.push_back(at1); 
  attrInfos.push_back(at2);

  Condition condition("OU = LBL \n && \t JOB = \tSUPERVISOR", attrInfos);
  vector<string> actions(1, "READ");

  UseCondCertGenerator gen;

  gen.setVersionNumber(2);
  gen.setUniqueID("UC1");
  gen.setNotBefore(now);
  gen.setNotAfter(after);
  gen.setIssuer(guarantor);
  gen.setResource(ROOT2);
  gen.setLocalScope(false);
  gen.setCritical(true);
  gen.setCondition(condition);
  gen.setActions(actions);

  KeyFactory kfac = KeyFactory::getInstance();
  PrivateKey privKey = kfac->initPrivateKeyFromFile("PEM", ID_DIR + SHI_KEY);

  gen.sign(privKey, SecurityConstants::RSA_MD5);

  AkentiCertificate cert = gen.generate();
  string fileName = AKCERT_DIR + ROOT2 + ".UC." + Util::createFileName(cert->hashCode()) + ".0";
  ofstream os(fileName.c_str());

  AkentiCertificate::writeToStream(AK_CERT_ENC_TYPE, os, cert);
  os.close();
  CPPUNIT_ASSERT(!os.fail());

#if DEBUG
  cout << "\nSaved ROOT2 UseCondition Certificate in " << fileName << endl;
#endif
}

void
UCGTC::generateSUB1UseCondCert() {
#if DEBUG
  cout << "\nGenerating SUB1 UseCondition Certificate ..." << endl;
#endif

  UtcTime now;
  UtcTime after(365 * 24 * 60 * 60);

  DistinguishedName ca(CAIA_DN);
  DistinguishedName issuer(SHI_DN);
  AkentiPrincipal guarantor(issuer, ca);

  vector<DistinguishedName> cas(1, ca);

  DistinguishedName user(USERI_DN);
  string val = user.getValue("CN");

  AttributeInfo at1("CN", val, cas);
  vector<AttributeInfo> attrInfos;

  attrInfos.push_back(at1);

  Condition condition("CN = " + val, attrInfos);

  vector<string> actions;

  actions.push_back("WRITE");
  actions.push_back("EXECUTE");

  UseCondCertGenerator gen;

  gen.setVersionNumber(2);
  gen.setUniqueID("UC2");
  gen.setNotBefore(now);
  gen.setNotAfter(after);
  gen.setIssuer(guarantor);
  gen.setResource(SUB1);
  gen.setLocalScope(true);
  gen.setCritical(false);
  gen.setCondition(condition);
  gen.setActions(actions);

  KeyFactory kfac = KeyFactory::getInstance();
  PrivateKey privKey = kfac->initPrivateKeyFromFile(
    SecurityConstants::PEM, ID_DIR + SHI_KEY
  );

  gen.sign(privKey, SecurityConstants::RSA_MD5);

  AkentiCertificate cert = gen.generate();

  string fileName = AKCERT_DIR + File(SUB1).getName();

  fileName += ".UC." + Util::createFileName(cert->hashCode()) + ".0";

  ofstream os(fileName.c_str());

  AkentiCertificate::writeToStream(AK_CERT_ENC_TYPE, os, cert);
  os.close();
  CPPUNIT_ASSERT(!os.fail());

#if DEBUG
  cout << "Sucessfully generated SUB1 UseCondition Certificate" << endl;
#endif
}

void
UCGTC::generateSUB3UseCondCert() {
#if DEBUG
  cout << "\n Generating SUB3 UseCondition Certificate ..." << endl;
#endif

  UtcTime now;
  UtcTime after(365 * 24 * 60 * 60);

  DistinguishedName ca(CAIA_DN);
  DistinguishedName issuer(SHI_DN);
  AkentiPrincipal guarantor(issuer, ca);


  vector<AkentiPrincipal> attrIssuers(1, guarantor);

  AttributeInfo at1(true, "JOB", "SUPERVISOR", attrIssuers);
  AttributeInfo at2(true, "JOB", "MANAGER", attrIssuers);
  vector<AttributeInfo> attrInfos; 

  attrInfos.push_back(at1); 
  attrInfos.push_back(at2);

  Condition condition("JOB = SUPERVISOR || JOB = MANAGER", attrInfos);
  vector<string> actions(1,"WRITE");

  UseCondCertGenerator gen;

  gen.setVersionNumber(2);
  gen.setUniqueID("UC4");
  gen.setNotBefore(now);
  gen.setNotAfter(after);
  gen.setIssuer(guarantor);
  gen.setResource(SUB3);
  gen.setLocalScope(false);
  gen.setCritical(true);
  gen.setCondition(condition);
  gen.setActions(actions);

  KeyFactory kfac = KeyFactory::getInstance();
  PrivateKey privKey = kfac->initPrivateKeyFromFile(
    "PEM", ID_DIR + SHI_KEY
  );

  gen.sign(privKey, SecurityConstants::RSA_MD5);

  AkentiCertificate cert = gen.generate();

  string fileName = AKCERT_DIR + File(SUB3).getName();

  fileName += ".UC." + Util::createFileName(cert->hashCode()) + ".0";

  ofstream os(fileName.c_str());

  AkentiCertificate::writeToStream(AK_CERT_ENC_TYPE, os, cert);
  os.close();
  CPPUNIT_ASSERT(!os.fail());

#if DEBUG
  cout << "Saved Use Condition file in " << fileName << endl;
#endif
}

void
UCGTC::generateSUB4UseCondCert() {
#if DEBUG
  cout << "\nGenerating SUB4 UseCondition Certificate ..." << endl;
#endif

  UtcTime now;
  UtcTime after(365 * 24 * 60 * 60);

  DistinguishedName ca(CAIA_DN);
  DistinguishedName issuer(SHI_DN);
  AkentiPrincipal guarantor(issuer, ca);

  vector<AkentiPrincipal> attrIssuers(1, guarantor);
  AttributeInfo at1(true, "JOB", "SUPERVISOR", attrIssuers);
  vector<AttributeInfo> attrInfos; 

  attrInfos.push_back(at1); 

  Condition condition("JOB = SUPERVISOR", attrInfos);
  vector<string> actions;

  actions.push_back("READ");
  actions.push_back("WRITE");
  actions.push_back("EXECUTE");

  UseCondCertGenerator gen;

  gen.setVersionNumber(2);
  gen.setUniqueID("UC5");
  gen.setNotBefore(now);
  gen.setNotAfter(after);
  gen.setIssuer(guarantor);
  gen.setResource(SUB4);
  gen.setLocalScope(true);
  gen.setCritical(true);
  gen.setCondition(condition);
  gen.setActions(actions);

  KeyFactory kfac = KeyFactory::getInstance();
  PrivateKey privKey = kfac->initPrivateKeyFromFile("PEM", ID_DIR + SHI_KEY);

  gen.sign(privKey, SecurityConstants::RSA_MD5);

  AkentiCertificate cert = gen.generate();

  string fileName = AKCERT_DIR + File(SUB4).getName();

  fileName += ".UC." + Util::createFileName(cert->hashCode()) + ".0";

  ofstream os(fileName.c_str());

  AkentiCertificate::writeToStream(AK_CERT_ENC_TYPE, os, cert);
  os.close();
  CPPUNIT_ASSERT(!os.fail());

#if DEBUG
  cout << "Saved SUB4 UseCondition Certificate in " << fileName << endl;
#endif
}

void
UCGTC::generateROOT3UseCondCert1() {
#if DEBUG
  cout << "\nGenerating ROOT3 UseCondition Certificate ..." << endl;
#endif

  UtcTime now;
  UtcTime after(365 * 24 * 60 * 60);

  DistinguishedName ca(CAIA_DN);
  DistinguishedName issuer(SHI_DN);
  AkentiPrincipal guarantor(issuer, ca);

  vector<DistinguishedName> cas(1, ca);
  vector<AkentiPrincipal> attrIssuers(1, guarantor);
  
  DistinguishedName user(USERI_DN);
  string val = user.getValue("OU");

  AttributeInfo at1("OU", val, cas);
  AttributeInfo at2(true, "JOB", "SUPERVISOR", attrIssuers);
  AttributeInfo at3("TIME", "10:00");   
  AttributeInfo at4("TIME", "12:00");  
  vector<AttributeInfo> attrInfos; 

  attrInfos.push_back(at1); 
  attrInfos.push_back(at2);
  attrInfos.push_back(at3);                         
  attrInfos.push_back(at4);                        

  Condition condition(                               
    "OU =" + val + " && JOB = SUPERVISOR && TIME >  10:00 && TIME < 12:00", attrInfos
  );

  vector<string> actions;

  actions.push_back("WRITE"); 
  actions.push_back("EXECUTE"); 

  UseCondCertGenerator gen;

  gen.setVersionNumber(2);
  gen.setUniqueID("UC6");
  gen.setNotBefore(now);
  gen.setNotAfter(after);
  gen.setIssuer(guarantor);
  gen.setResource(ROOT3);
  gen.setLocalScope(false);
  gen.setCritical(true);
  gen.setCondition(condition);
  gen.setActions(actions);
  
  KeyFactory kfac = KeyFactory::getInstance();
  PrivateKey privKey  = kfac->initPrivateKeyFromFile(
    "PEM", ID_DIR + SHI_KEY
  );

  gen.sign(privKey, SecurityConstants::RSA_MD5);

  AkentiCertificate cert = gen.generate();
  string fileName = AKCERT_DIR + ROOT3 + ".UC." + Util::createFileName(cert->hashCode()) + ".0";
  ofstream os(fileName.c_str());

  AkentiCertificate::writeToStream(AK_CERT_ENC_TYPE, os, cert);
  os.close();
  CPPUNIT_ASSERT(!os.fail());

#if DEBUG
 
  cout << "\nSaved ROOT3 UseCondition Certificate in " << fileName << endl;
#endif
}


void
UCGTC::generateROOT3UseCondCert2() {
#if DEBUG
  cout << "\nGenerating ROOT3 UseCondition Certificate ..." << endl;
#endif

  UtcTime now;
  UtcTime after(365 * 24 * 60 * 60);

  DistinguishedName ca(CAIA_DN);
  DistinguishedName issuer(SHI_DN);
  AkentiPrincipal guarantor(issuer, ca);

  vector<DistinguishedName> cas(1, ca);
  vector<AkentiPrincipal> attrIssuers(1, guarantor);
  
  DistinguishedName user(USERI_DN);
  string val = user.getValue("OU");

  AttributeInfo at1("OU", val, cas);
  AttributeInfo at2(true, "JOB", "SUPERVISOR", attrIssuers);
  vector<AttributeInfo> attrInfos; 

  attrInfos.push_back(at1); 
  attrInfos.push_back(at2);

  Condition condition(                                      
    "OU =" + val + " && JOB = SUPERVISOR", attrInfos
  );

  vector<string> actions(1, "READ");

  UseCondCertGenerator gen;
  gen.setVersionNumber(2);
  gen.setUniqueID("UC7");
  gen.setNotBefore(now);
  gen.setNotAfter(after);
  gen.setIssuer(guarantor);
  gen.setResource(ROOT3);
  gen.setLocalScope(false);
  gen.setCritical(true);
  gen.setCondition(condition);
  gen.setActions(actions);
  
  KeyFactory kfac = KeyFactory::getInstance();
  PrivateKey privKey  = kfac->initPrivateKeyFromFile(
    "PEM", ID_DIR + SHI_KEY
  );

  gen.sign(privKey, SecurityConstants::RSA_MD5);

  AkentiCertificate cert = gen.generate();
  string fileName = AKCERT_DIR + ROOT3 + ".UC." + Util::createFileName(cert->hashCode()) + ".1";
  ofstream os(fileName.c_str());

  AkentiCertificate::writeToStream(AK_CERT_ENC_TYPE, os, cert);
  os.close();
  CPPUNIT_ASSERT(!os.fail());

#if DEBUG
  cout << "\nSaved ROOT3 UseCondition Certificate in " << fileName << endl;
#endif
}

void
UCGTC::generateROOT4UseCondCert() {
#if DEBUG
  cout << "\nGenerating ROOT4 UseCondition Certificate ..." << endl;
#endif

  UtcTime now;
  UtcTime after(365 * 24 * 60 * 60);

  DistinguishedName ca(CAIA_DN);
  DistinguishedName issuer(SHI_DN);
  AkentiPrincipal guarantor(issuer, ca);

  vector<DistinguishedName> cas(1, ca);
  vector<AkentiPrincipal> attrIssuers(1, guarantor);
  
  AttributeInfo at1(false, "ExtAuthTester", "10", attrIssuers);
  vector<AttributeInfo> attrInfos; 

  attrInfos.push_back(at1); 

  Condition condition(                                      
    "ExtAuthTester > 10", attrInfos
  );

  vector<string> actions(1, "WRITE");

  UseCondCertGenerator gen;
  gen.setVersionNumber(2);
  gen.setUniqueID("UC8");
  gen.setNotBefore(now);
  gen.setNotAfter(after);
  gen.setIssuer(guarantor);
  gen.setResource(ROOT4);
  gen.setLocalScope(false);
  gen.setCritical(true);
  gen.setCondition(condition);
  gen.setActions(actions);
  
  KeyFactory kfac = KeyFactory::getInstance();
  PrivateKey privKey  = kfac->initPrivateKeyFromFile(
    "PEM", ID_DIR + SHI_KEY
  );

  gen.sign(privKey, SecurityConstants::RSA_MD5);

  AkentiCertificate cert = gen.generate();
  string fileName = AKCERT_DIR + ROOT4 + ".UC." + Util::createFileName(cert->hashCode()) + ".1";
  ofstream os(fileName.c_str());

  AkentiCertificate::writeToStream(AK_CERT_ENC_TYPE, os, cert);
  os.close();
  CPPUNIT_ASSERT(!os.fail());

#if DEBUG
  cout << "\nSaved ROOT4 UseCondition Certificate in " << fileName << endl;
#endif
}


CppUnit::TestSuite*
UCGTC::suite() {
  CppUnit::TestSuite* testSuite = new CppUnit::TestSuite("UseCondCertGeneratorTestCase");

  testSuite->addTest (
    new CppUnit::TestCaller<UseCondCertGeneratorTestCase>(
      "generateUseCondCerts",
      &UseCondCertGeneratorTestCase::generateUseCondCerts
    )
  );

  return testSuite;
}
