#include <iostream>
#include <string>
#include <map>
#include "cppunit/TestCaller.h"
#include "cppunit/Exception.h"
#include "cppunit/Message.h"
#include "cppunit/SourceLine.h"

#include "../shared/TestCaseConstants.hpp"
#include "CertRetrievalTestCase.hpp"
#include "DistinguishedName.hpp"
#include "AkentiCertificate.hpp"
#include "AttributeCertificate.hpp"
#include "UseCondCertificate.hpp"
#include "Certificate.hpp"
#include "CertificateRetrievalFactory.hpp"
#include "Provider.hpp"
#include "MessageDigest.hpp"
#include "Directory.hpp"
#include "URL.hpp"
#include "File.hpp"
#include "Util.hpp"

using namespace Akenti;
using namespace std;

typedef CertRetrievalTestCase CRTC;

void 
CRTC::doTest() {
  string method;

  try {
    method = "CertRetrievalTestCase:testX509LdapRetrieval:";
    testX509LdapRetrieval();
    method = "CertRetrievalTestCase:testX509FileRetrieval:";
    testX509FileRetrieval();
    method = "CertRetrievalTestCase:testUCFileRetrieval:";
    testUCFileRetrieval();
    method = "CertRetrievalTestCase:testATTRFileRetrieval:";
    testATTRFileRetrieval();
    method = "CertRetrievalTestCase:testX509HttpRetrieval:";
    testX509HttpRetrieval();
    method = "CertRetrievalTestCase:testUCHttpRetrieval:";
    testUCHttpRetrieval();
    method = "CertRetrievalTestCase:testATTRHttpRetrieval:";
    testATTRHttpRetrieval();
  } catch(Exception& ex) {
    CppUnit::Message m(method + ex.getMessage());

    throw CppUnit::Exception(m, CPPUNIT_SOURCELINE());
  } catch(Error& err) {
    CppUnit::Message m(method + err.getMessage());

    throw CppUnit::Exception(m, CPPUNIT_SOURCELINE());
  } 
}

void
CRTC::testX509FileRetrieval() {
#if DEBUG
  cout << "\nBEGIN testX509FileRetrieval" << endl;
#endif

  URL url("file:" + File(ID_DIR).getAbsolutePath());
  map<string, string> args;
  
  args["type"] = "X509";
  args["hash"] = Util::createFileName(
    MessageDigest::digest(
      "MD5",
      Provider::getProvider()->toDerString(USERI_DN)
    )
  );

#if DEBUG
  cout << "URL=" << url << endl;
  cout << "type=" << args["type"] << endl;
  cout << "hash=" << args["hash"] << endl;
#endif

  vector<Certificate> certs = CertificateRetrievalFactory::getCertificates(url, args);
  CPPUNIT_ASSERT(certs.size() == 1);
  CPPUNIT_ASSERT(certs[0]->getPrincipal() == USERI_DN);

#if DEBUG
  cout << certs[0]->toText() << endl;
  cout << "END testX509FileRetrieval" << endl;
#endif
}

void
CRTC::testUCFileRetrieval() {
#if DEBUG
  cout << "\nBEGIN testUCFileRetrieval" << endl;
#endif

  URL url("file:" + File(AKCERT_DIR).getAbsolutePath());
  map<string, string> args;

  args["type"] = "UseCondition";
  args["hash"] = Util::createFileName(MessageDigest::digest("MD5", ROOT1));
  args["enc"] = AK_CERT_ENC_TYPE;

#if DEBUG
  cout << "URL=" << url << endl;
  cout << "type=" << args["type"] << endl;
  cout << "hash=" << args["hash"] << endl;
#endif

  vector<AkentiCertificate> certs = CertificateRetrievalFactory::getAkentiCertificates(
    url, args
  );
  CPPUNIT_ASSERT(certs.size() == 1);
  CPPUNIT_ASSERT(certs[0].instanceOf<UseCondCertificate>());

  UseCondCertificate ucCert = (UseCondCertificate&) certs[0];

  CPPUNIT_ASSERT(ucCert->getResource() == ROOT1);

#if DEBUG
  cout << ucCert->toText() << endl;
  cout << "END testUCFileRetrieval" << endl;
#endif
}

void
CRTC::testATTRFileRetrieval() {
#if DEBUG
  cout << "\nBEGIN testATTRFileRetrieval" << endl;
#endif

  URL url("file:" + File(AKCERT_DIR).getAbsolutePath());
  map<string, string> args;

  args["type"] = "Attribute";
  args["hash"] = Util::createFileName(
    MessageDigest::digest(
      "MD5",
      Util::getHashDN(USERI_DN) + Util::getHashDN(CAIA_DN) + "JOB" + "SUPERVISOR"
    )
  );
  args["enc"] = AK_CERT_ENC_TYPE;

#if DEBUG
  cout << "URL=" << url << endl;
  cout << "type=" << args["type"] << endl;
  cout << "hash=" << args["hash"] << endl;
#endif

  vector<AkentiCertificate> certs = CertificateRetrievalFactory::getAkentiCertificates(
    url, args
  );

  CPPUNIT_ASSERT(certs.size() == 1);
  CPPUNIT_ASSERT(certs[0].instanceOf<AttributeCertificate>());

  AttributeCertificate attrCert = (AttributeCertificate&) certs[0];

  CPPUNIT_ASSERT(attrCert->getPrincipal() == AkentiPrincipal(USERI_DN, CAIA_DN));
  CPPUNIT_ASSERT(attrCert->getAttribute() == "JOB");
  CPPUNIT_ASSERT(attrCert->getValue() == "SUPERVISOR");

#if DEBUG
  cout << attrCert->toText() << endl;
  cout << "End testATTRFileRetrieval" << endl;
#endif
}

void
CRTC::testX509LdapRetrieval() {
#if LBL_DEBUG 
#if DEBUG
  cout << "\nBEGIN testX509LdapRetrieval for LBNL-CA" << endl;
#endif

  Directory dir("ldap://idcg-ds.lbl.gov:389");
  map<string, string> args;

  args["base"] = "/o=Grid/O=Lawrence Berkeley National Laboratory";
  args["type"] = "X509";
  args["filter"] = "cn=*Akenti*";
  args["scope"] = "subtree";

#if DEBUG
  cout << "Directory=" << dir.toString() << endl;
  cout << "base=" << args["base"] << endl;
  cout << "type=" << args["type"] << endl;
  cout << "filter=" << args["filter"] << endl;
  cout << "scope=" << args["scope"] << endl;
#endif

  vector<Certificate> certs = CertificateRetrievalFactory::getCertificates(dir, args);

  CPPUNIT_ASSERT(certs.size() != 0);

#if DEBUG
  for (int i = 0; i < certs.size(); ++i) {
    cout << certs[i]->toText() << endl;
  }

#endif

#if DEBUG
  cout << "\nBEGIN testX509LdapRetrieval for doesciencegrid" << endl;
#endif
  {
  Directory dir1("ldap://ldap.doesciencegrid.org:389");
  map<string, string> args;

  args["base"] = "/O=doesciencegrid.org";
  args["type"] = "X509";
  args["filter"] = "cn=*Thompson*";
  args["scope"] = "subtree";

#if DEBUG
  cout << "Directory=" << dir1.toString() << endl;
  cout << "base=" << args["base"] << endl;
  cout << "type=" << args["type"] << endl;
  cout << "filter=" << args["filter"] << endl;
  cout << "scope=" << args["scope"] << endl;
#endif

  vector<Certificate> certs1 = CertificateRetrievalFactory::getCertificates(dir1, args);

  CPPUNIT_ASSERT(certs1.size() != 0);

#if DEBUG
  for (int i = 0; i < certs1.size(); ++i) {
    cout << certs1[i]->toText() << endl;
  }

  cout << "End testX509LdapRetrieval" << endl;
#endif
  }
#endif
}

void
CRTC::testX509HttpRetrieval() {
#if LBL_DEBUG 
#if DEBUG
  cout << "\nBEGIN testX509HttpRetrieval" << endl;
#endif

  URL url("http://www-itg.lbl.gov/Akenti/TestCerts/");

  map<string, string> args;

  args["type"] = "X509";
  args["hash"] = Util::createFileName(
    MessageDigest::digest(
      "MD5",
      Provider::getProvider()->toDerString(USERI_DN)
    )
  );

#if DEBUG
  cout << "URL=" << url << endl;
  cout << "type=" << args["type"] << endl;
  cout << "hash=" << args["hash"] << endl;
#endif

  vector<Certificate> certs = CertificateRetrievalFactory::getCertificates(url, args);

  CPPUNIT_ASSERT(certs.size() == 1);
  CPPUNIT_ASSERT(certs[0]->getPrincipal() == USERI_DN);

#if DEBUG
  cout << certs[0]->toText() << endl;
  cout << "End testX509HttpRetrieval" << endl;
#endif
#endif
}

void
CRTC::testUCHttpRetrieval() {
#if LBL_DEBUG
#if DEBUG
  cout << "\nBEGIN testUCHttpRetrieval" << endl;
#endif

  URL url("http://www-itg.lbl.gov/Akenti/TestCerts/");
  map<string, string> args;

  args["type"] = "UseCondition";
  args["hash"] = Util::createFileName(MessageDigest::digest("MD5", ROOT2));

#if DEBUG
  cout << "URL=" << url << endl;
  cout << "type=" << args["type"] << endl;
  cout << "hash=" << args["hash"] << endl;
#endif

  vector<AkentiCertificate> certs = CertificateRetrievalFactory::getAkentiCertificates(
    url, args
  );

  CPPUNIT_ASSERT(certs.size() == 1);

#if DEBUG
  cout << certs[0]->toText() << endl;
  cout << "END testUCHttpRetrieval" << endl;
#endif
#endif
}

void
CRTC::testATTRHttpRetrieval() {
#if LBL_DEBUG
#if DEBUG
  cout << "\nBEGIN testATTRHttpRetrieval" << endl;
#endif
  URL url("http://www-itg.lbl.gov/Akenti/TestCerts/");
  map<string, string> args;

  args["type"] = "Attribute";
  args["hash"] = Util::createFileName(
    MessageDigest::digest(
      "MD5",
      Util::getHashDN(USERI_DN) + Util::getHashDN(CAIA_DN) + "JOB" + "SUPERVISOR"
    )
  );

#if DEBUG
  cout << "URL=" << url << endl;
  cout << "type=" << args["type"] << endl;
  cout << "hash=" << args["hash"] << endl;
#endif

  vector<AkentiCertificate> certs = CertificateRetrievalFactory::getAkentiCertificates(
    url, args
  );

  CPPUNIT_ASSERT(certs.size() == 1);

#if DEBUG
  cout << certs[0]->toText() << endl;
  cout << "End testATTRHttpRetrieval" << endl;
#endif
#endif
}

CppUnit::TestSuite*
CRTC::suite() {
  CppUnit::TestSuite* testSuite = new CppUnit::TestSuite("CertRetrievalTestCase");

  testSuite->addTest(
    new CppUnit::TestCaller<CertRetrievalTestCase>(
     "doTest", &CertRetrievalTestCase::doTest
    )
  );

  return testSuite;
}
