#include <unistd.h>
#include <iostream>

#include "cppunit/TestCaller.h"
#include "cppunit/Exception.h"
#include "cppunit/Message.h"
#include "cppunit/SourceLine.h"
#include "../shared/TestCaseConstants.hpp"

#include <memory>
#include "AKRuntime.hpp"
#include "PolicyCollector.hpp"
#include "UseCondCollector.hpp"
#include "PolicyAuthorizer.hpp"
#include "CheckPolCertCollector.hpp"
#include "CheckUCCertCollector.hpp"
#include "CertInfo.hpp"
#include "Util.hpp"
#include "Properties.hpp"
#include "AkentiCertificate.hpp"
#include "SecurityConstants.hpp" 
#include "CertificateCache.hpp"
#include "AKCertHolder.hpp"
#include "MessageDigest.hpp"
#include "EngComponentTestCase.hpp"
#include "Exception.hpp"
#include "Error.hpp"

typedef EngComponentTestCase ECTC;

using namespace Akenti;
using namespace std;

void 
ECTC::doTest() {
  string method;

  try {
    AKRuntime* akRuntime1 = AKRuntime::createAKRuntime(); 
    auto_ptr<AKRuntime> runtimeHolder1(akRuntime1);

    akRuntime1->setResource(SUB4);
    akRuntime1->setResourceAltName(SUB4);

    AKRuntime* akRuntime = AKRuntime::createAKRuntime(); 
    auto_ptr<AKRuntime> runtimeHolder(akRuntime);

    akRuntime->setResource(SUB4);
    akRuntime->setResourceAltName(SUB4);

    method = "EngComponentTestCase:runPolicyColletorTest:";
    runPolicyColletorTest(akRuntime);
    method = "EngComponentTestCase:runUseCondColletorTest:";
    runUseCondColletorTest(akRuntime);

    method = "EngComponentTestCase:runCheckPolicyCollectorTest:";
    runCheckPolicyCollectorTest(akRuntime1);
    method = "EngComponentTestCase:runCheckUseCondCollectorTest:";
    runCheckUseCondCollectorTest(akRuntime1);

    method = "EngComponentTestCase:runPolicyAuthorizerTest:";
    runPolicyAuthorizerTest(akRuntime);
  } catch(Exception& ex) {
    CppUnit::Message m(method + ex.getMessage());

    throw CppUnit::Exception(m, CPPUNIT_SOURCELINE());
  } catch(Error& err) {
    CppUnit::Message m(method + err.getMessage());

    throw CppUnit::Exception(m, CPPUNIT_SOURCELINE());
  }
}

void
ECTC::runPolicyColletorTest(AKRuntime* akRuntime) {
  PolicyCollector policyCollector(akRuntime);

  CPPUNIT_ASSERT(policyCollector.collectPolicies());
  CPPUNIT_ASSERT(policyCollector.verifyPolicies()); 
  CPPUNIT_ASSERT(akRuntime->numOfPolicies() == 4);   
  CPPUNIT_ASSERT(akRuntime->getLevel() == 4);
}

void
ECTC::runCheckPolicyCollectorTest(AKRuntime* akRuntime) {
  CheckPolCertCollector checkPolicyCollector(akRuntime);

  CPPUNIT_ASSERT(checkPolicyCollector.collectPolicies());
  CPPUNIT_ASSERT(checkPolicyCollector.verifyPolicies()); 

  const vector <CertInfo>& cinfoVec = (akRuntime->getCertInfoCollector()).getCertInfoVec ();
  CPPUNIT_ASSERT (cinfoVec.size() == 4); // number of Policy certificates
}


void
ECTC::runUseCondColletorTest(AKRuntime* akRuntime) {
  UseCondCollector ucCollector(akRuntime);

  CPPUNIT_ASSERT(ucCollector.collectUseCondCertificates());
  CPPUNIT_ASSERT(akRuntime->getLevel() == 4);       
  CPPUNIT_ASSERT(akRuntime->numOfPolicies() == 4); 

  vector<URL> urls = akRuntime->getIdentityURLS(1);              

  CPPUNIT_ASSERT(urls.size() == 1);
}

void
ECTC::runCheckUseCondCollectorTest(AKRuntime* akRuntime) {
  CheckUCCertCollector checkUCCollector(akRuntime);
  CPPUNIT_ASSERT(checkUCCollector.collectUseCondCertificates());

  const vector <CertInfo>& cinfoVec = (akRuntime->getCertInfoCollector()).getCertInfoVec ();
  
  // number of Policy + Use Condition certificates 
  CPPUNIT_ASSERT (cinfoVec.size() == 8);  
}

void
ECTC::runPolicyAuthorizerTest(AKRuntime* akRuntime) {
  AkentiPrincipal user(USERI_DN, CAIA_DN);

  akRuntime->setAkentiPrincipal(user);

  PolicyAuthorizer authorizer(akRuntime);
  bool ret = authorizer.authorize();                       

  CPPUNIT_ASSERT(ret);

#if DEBUG
  cout << authorizer.getAkentiMessage().toString() << endl;

  if (ret) {                                                  
    PolicyContext context = akRuntime->getPolicyContext();
    const vector<string>& actions = context->getActions();

    cout << "Actions:" << actions[0];

    for (int i = 1; i < actions.size(); ++i) {
      cout << ", " << actions[i];
    }
   
    cout << endl;
  }
#endif 
}

CppUnit::TestSuite*
ECTC::suite() {
  CppUnit::TestSuite* testSuite = new CppUnit::TestSuite("EngComponentTestCase");

  testSuite->addTest(
    new CppUnit::TestCaller<EngComponentTestCase>( 
      "doTest", 
      &EngComponentTestCase::doTest
    )
  );

  return testSuite;
}
