#include <unistd.h>

#include "cppunit/TestCaller.h"
#include "cppunit/Exception.h"
#include "cppunit/Message.h"
#include "cppunit/SourceLine.h"

#include "../../shared/TestCaseConstants.hpp"

#include "UtcTime.hpp"
#include "Util.hpp"
#include "Properties.hpp"
#include "AkentiCertificate.hpp"
#include "SecurityConstants.hpp" 
#include "CertificateCache.hpp"
#include "AKCertHolder.hpp"
#include "MessageDigest.hpp"
#include "AkentiInit.hpp"
#include "CacheTestCase.hpp"
#include "Exception.hpp"
#include "Error.hpp"

typedef CacheTestCase CTC;

using namespace Akenti;
using namespace std;

void 
CTC::doTest() {
  string method;

  try {
    method = "CacheTestCase:runCacheTest1:";
    runCacheTest1();
    method = "CacheTestCase:runCacheTest2:";
    runCacheTest2();
    method = "CacheTestCase:runCacheTest3:";
    runCacheTest3();
    method = "CacheTestCase:runCacheTest4:";
    runCacheTest4();
  } catch(Exception& ex) {
    CppUnit::Message m(method + ex.getMessage());

    throw CppUnit::Exception(m, CPPUNIT_SOURCELINE());
  } catch(Error& err) {
    CppUnit::Message m(method + err.getMessage());

    throw CppUnit::Exception(m, CPPUNIT_SOURCELINE());
  }
}

void
CTC::runCacheTest1() {
  string hash = MessageDigest::digest("MD5", ROOT1);
  // XXX need to remove "../" later
  string fileName = "../" + AKCERT_DIR + ROOT1 + ".UC." + Util::createFileName(hash) + ".0";

  AkentiCertificate cert = AkentiCertificate::initFromFile(
    AK_CERT_ENC_TYPE, fileName  
  );

  CertificateCache cache;
  UtcTime now;
  UtcTime after(now + 30);	
  int level = 1;
  string type = cert->getType();

  // Test duplicates 
  cache.addCertificate(cert, level, now, after);
  cache.addCertificate(cert, level, now, after); 
  cache.addCertificate(cert, level, now, after); 

  UtcTime notBefore(-1 * 10);
  vector<AKCertHolder> holders = cache.getAKCertHolders(hash, type, notBefore); 

  CPPUNIT_ASSERT(holders.size() == 1);		
  CPPUNIT_ASSERT(holders[0]->getLevel() == level);		
  CPPUNIT_ASSERT(holders[0]->getStartCacheTime() == now);		
  CPPUNIT_ASSERT(holders[0]->getEndCacheTime() == after);		
  
  // Pretend cert did not verify
  holders[0]->setLevel(-1);
  holders = cache.getAKCertHolders(hash, type, notBefore); 
  CPPUNIT_ASSERT(holders.size() == 0);
}

void
CTC::runCacheTest2() {
  string hash = MessageDigest::digest("MD5", ROOT1);
  // XXX need to remove "../" later
  string fileName = "../" + AKCERT_DIR + ROOT1 + ".UC." + Util::createFileName(hash) + ".0";

  AkentiCertificate cert = AkentiCertificate::initFromFile(
    AK_CERT_ENC_TYPE, fileName
  );

  CertificateCache cache;
  UtcTime now;
  UtcTime after(now + 30);	
  string type = cert->getType();

  AKCertHolder holder1 = cache.addCertificate(cert, 1, now, after);

  holder1->setLevel(2);
  cache.addAKCertHolder(holder1);

  UtcTime notBefore(-1 * 10);
  vector<AKCertHolder> holders = cache.getAKCertHolders(hash, type, notBefore); 

  CPPUNIT_ASSERT(holders.size() == 1);		
  CPPUNIT_ASSERT(holders[0]->getLevel() == 2);
  now = UtcTime();
  after = after + 30;

  AKCertHolder holder2 = new AKCertHolderImpl(cert, 3, now, after);
  
  cache.addAKCertHolder(holder2);
  notBefore = UtcTime(now + (-1 * 10));
  holders = cache.getAKCertHolders(hash, type, notBefore); 
  CPPUNIT_ASSERT(holders.size() == 1);		
  CPPUNIT_ASSERT(holders[0]->getLevel() == 3);
  CPPUNIT_ASSERT(holders[0]->getStartCacheTime() == now);		
  CPPUNIT_ASSERT(holders[0]->getEndCacheTime() == after);		
  holders[0]->setLevel(-1);
  holders = cache.getAKCertHolders(hash, type, notBefore); 
  CPPUNIT_ASSERT(holders.size() == 0);
}

#include "ExternalCache.hpp"
#include "FileExternalCacheImpl.hpp"

void
CTC::runCacheTest3() {
  ExternalCache extCache(new FileExternalCacheImpl);
  ExternalCache::setExternalCache(extCache);

  string hash = MessageDigest::digest("MD5", ROOT1);
  string fileName = "../" + AKCERT_DIR + ROOT1 + ".UC." + Util::createFileName(hash) + ".0";

  AkentiCertificate cert = AkentiCertificate::initFromFile(
    AK_CERT_ENC_TYPE, fileName  
  );

  CertificateCache cache;
  UtcTime now;
  UtcTime after(now + 30);	
  int level = 1;
  string type = cert->getType();

  // Test duplicates 
  cache.addCertificate(cert, level, now, after);
  cache.addCertificate(cert, level, now, after); 
  cache.addCertificate(cert, level, now, after); 

  UtcTime notBefore(-1 * 10);
  vector<AKCertHolder> holders = cache.getAKCertHolders(hash, type, notBefore); 

  CPPUNIT_ASSERT(holders.size() == 1);		
  CPPUNIT_ASSERT(holders[0]->getLevel() == level);		
  CPPUNIT_ASSERT(holders[0]->getStartCacheTime() == now);		
  CPPUNIT_ASSERT(holders[0]->getEndCacheTime() == after);		
  CPPUNIT_ASSERT(cache.updateExternalCache());
}

void
CTC::runCacheTest4() {
  string hash = MessageDigest::digest("MD5", ROOT1);
  CertificateCache cache;
  string type = "UseCondition";
  UtcTime notBefore(-1 * 10);
  vector<AKCertHolder> holders = cache.getAKCertHolders(hash, type, notBefore); 

  //XXX will fix this later
  CPPUNIT_ASSERT(holders.size() == 1);		
  CPPUNIT_ASSERT(!cache.updateExternalCache());
}

CppUnit::TestSuite*
CTC::suite() {
  CppUnit::TestSuite* testSuite = new CppUnit::TestSuite("CacheTestCase");

  testSuite->addTest(
    new CppUnit::TestCaller<CacheTestCase>( 
      "doTest", 
      &CacheTestCase::doTest
    )
  );

  return testSuite;
}
