#include <iostream>
#include <vector>
#include "cppunit/TestCaller.h"
#include "cppunit/Exception.h"
#include "cppunit/Message.h"
#include "cppunit/SourceLine.h"
#include "Xerces-Includes.hpp"
#include "CheckAccessResponse.hpp"
#include "CheckAccessResponseTestCase.hpp"
#include "AkentiMessage.hpp"
#include "../shared/TestCaseConstants.hpp"
#include "CapabilityCertGenerator.hpp"
#include "UtcTime.hpp"
#include "DistinguishedName.hpp"
#include "AttributeInfo.hpp"
#include "ConditionalActions.hpp"
#include "KeyFactory.hpp"
#include "KeyPair.hpp"
#include "PrivateKey.hpp"
#include "PublicKey.hpp"
#include "AkentiResponses.hpp"
#include "SecurityConstants.hpp"
#include "Exception.hpp"
#include "Error.hpp"

using namespace Akenti;
using namespace std;

typedef CheckAccessResponseTestCase CARTC;

void 
CARTC::doTest() {
  string method;

  try {
    method = "CheckAccessResponseTestCase:doTest:test1:";
    test1();
  } catch (Exception& ex) {
    CppUnit::Message m(method + ex.getMessage());

    throw CppUnit::Exception(m, CPPUNIT_SOURCELINE());
  } catch (Error& err) {
    CppUnit::Message m(method + err.getMessage());

    throw CppUnit::Exception(m, CPPUNIT_SOURCELINE());
  }
}

void
CARTC::test1() {
#if DEBUG
  cout << "Entering test1 ..." << endl;
#endif

  int id = 1;
  AkentiMessage amsg(104, "Gammasphere status okay");
  CapabilityCertificate cc = genCapabilityCert();
  CheckAccessResponse car;

  car.setID(id);
  car.setAkentiMessage(amsg);
  car.setCapabilityCertificate(cc);

  CPPUNIT_ASSERT(id == car.getID());
  CPPUNIT_ASSERT(amsg.getCode() == car.getAkentiMessage().getCode());
  CPPUNIT_ASSERT(amsg.getMessage() == car.getAkentiMessage().getMessage());
  CPPUNIT_ASSERT(cc == car.getCapabilityCertificate());

  AkentiResponses resps;

  resps.addResponse(car);
  resps.addResponse(car);
  
  string rep = resps.toString();

#if DEBUG
  cout << rep << endl; 
#endif

  DOMElement* element = resps.toXML();

  resps.fromXML(element);
  delete(element->getOwnerDocument());
  CPPUNIT_ASSERT(rep == resps.toString());

  car = (CheckAccessResponse&) resps.getResponseAt(1);
  CPPUNIT_ASSERT(id == car.getID());
  CPPUNIT_ASSERT(amsg.getCode() == car.getAkentiMessage().getCode());
  CPPUNIT_ASSERT(amsg.getMessage() == car.getAkentiMessage().getMessage());
  CPPUNIT_ASSERT(cc == car.getCapabilityCertificate());

#if DEBUG
  cout << "Exiting test1 ..." << endl;
#endif
}

CapabilityCertificate
CARTC::genCapabilityCert() {
#if DEBUG
  cout << "\nGenerating Capability Certificate ..." << endl;
#endif 

  UtcTime now;
  UtcTime after(365 * 24 * 60 * 60);

  DistinguishedName ca(CAIA_DN);
  DistinguishedName issuer(SHI_DN);
  AkentiPrincipal guarantor(issuer, ca);

  DistinguishedName user(USERI_DN);
  AkentiPrincipal principal(user, ca);
  string act1("read");
  string act2("write");
  string act3("execute");

  vector<string> unCondActs;

  unCondActs.push_back(act1);
  unCondActs.push_back(act2);

  vector<string> condAct1Acts;
  vector<string> condAct2Acts;

  condAct1Acts.push_back(act3);
  condAct2Acts.push_back(act1);
  condAct2Acts.push_back(act2);

  vector<DistinguishedName> cas(1, ca);
  vector<AkentiPrincipal> attrIssuers(1, guarantor);

  string val = user.getValue("OU");
  string val2 = user.getValue("CN");

  AttributeInfo at1("OU", val, cas);
  AttributeInfo at2(true, "JOB", "SUPERVISOR", attrIssuers);
  AttributeInfo at3("CN", val2, cas);

  vector<AttributeInfo> attrInfos;
  vector<AttributeInfo> attrInfos2;

  attrInfos.push_back(at1); 
  attrInfos.push_back(at2);
  attrInfos2.push_back(at3);

  Condition cond1("OU =" + val + " && JOB = SUPERVISOR", attrInfos);
  Condition cond2("CN = " + val2, attrInfos2);

  vector<ConditionalActions> condActs;

  ConditionalActions condAct1(cond1, condAct1Acts);
  ConditionalActions condAct2(cond2, condAct2Acts, true);

  condActs.push_back(condAct1);
  condActs.push_back(condAct2);

  CapabilityCertGenerator gen;

  gen.setVersionNumber(2);
  gen.setUniqueID("1");
  gen.setIssuer(guarantor);
  gen.setNotBefore(now);
  gen.setNotAfter(after);
  gen.setResource(ROOT1);
  gen.setPrincipal(principal);
  gen.setUnConditionalActions(unCondActs);
  gen.setConditionalActions(condActs);

  KeyFactory kfac = KeyFactory::getInstance();
  KeyPair pair = kfac->generateRSAKeyPair();

  gen.setPublicKey(pair.getPublicKey());
  gen.sign(pair.getPrivateKey(), SecurityConstants::RSA_MD5);

  return gen.generate();

#if DEBUG
   cout << "done generating capability cert" << endl;
#endif

}

CppUnit::TestSuite*
CARTC::suite() {
  CppUnit::TestSuite* testSuite = new CppUnit::TestSuite("CheckAccessResponseTestCase");

  testSuite->addTest (
    new CppUnit::TestCaller<CheckAccessResponseTestCase> ( 
      "doTest", 
      &CheckAccessResponseTestCase::doTest
    )
  );

  return testSuite;
}
