#include <iostream>
#include <unistd.h>
#include "Provider.hpp"
#include "OpenSSLProviderImpl.hpp"
#include "CertificateFactory.hpp"
#include "Certificate.hpp"
#include "PublicKey.hpp"
#include "Exception.hpp"

using namespace Akenti;
using namespace std;

/**
  */

namespace VerifyX509Cert {

void printPurpose() {
  cerr << "Purpose: Use this program to verify an x509 cert." << endl
       << endl
       << "Check 1: time validity" << endl
       << "Check 2: signature (optional)" << endl
       << endl;
}

void printUsage () {
  cerr << "Usage: VerifyX509Cert args" << endl
       << "  -x arg    - x509 cert" << endl
       << "  [-i arg]  - ca's x509 cert" << endl
       << "  [-v]      - verbose" << endl
       << "  [-h]      - prints this message" << endl;
}
 
int
main (int argc, char *argv[]) {
  string caCertFile;
  string subjectCertFile;
  bool verbose = false;
  int opt;
   
  if (argc < 2) {
    printPurpose();
    printUsage();
    exit(2);
  }

  while ((opt = getopt(argc, argv, "x:i:vh")) != EOF) {
    switch(opt){
      case 'x':
        subjectCertFile = optarg;
        break;
      case 'i':
        caCertFile = optarg;
        break;
      case 'v':
        verbose = true;
        break;
      case 'h':
      default:
        printUsage();
        exit(2);
    }
  }

  if (!subjectCertFile.length()) {
    cerr << "Must provide an X509 cert." << endl;
    printUsage();
    exit(2);
  }

  try { 
    if (verbose) {
      cout << "Installing security provider..." << endl;
    }

    Provider po = Provider(new OpenSSLProviderImpl);
    
    po->loadDigestAlgorithms();
    po->loadSignatureAlgorithms();
    Provider::setProvider(po);

    if (verbose) {
      cout << "Reading subject cert from " << subjectCertFile << endl;
    }

    CertificateFactory cfac = CertificateFactory::getInstance();
    Certificate cert = cfac->initCertificateFromFile("PEM", subjectCertFile);

    if (verbose) {
      cout << cert->toText() << endl;
      cout << "Checking time validity ..." << endl;
    }

    UtcTime now;
    bool ret = cert->checkValidity(now);

    if (ret) {
      cout << "cert is valid as of " << now.toString("YYYY-MM-DDTHH:MM:SSZ") << endl;
    } else {
      cout << "cert is not valid as of " << now.toString("YYYY-MM-DDTHH:MM:SSZ") << endl;
    }

    if (!caCertFile.length()) {
      exit(!ret);
    }

    if (verbose) {
      cout << "Reading caCert from " <<  caCertFile << endl;
    }

    Certificate caCert  = cfac->initCertificateFromFile("PEM", caCertFile);

    if (verbose) {
      cout << caCert->toText() << endl;
    }

    if (!caCert->checkValidity(now)) {
      cout << "issuer's cert is not valid as of " << now.toString("YYYY-MM-DDTHH:MM:SSZ") << endl;
      exit(1);
    } else if (cert->verify(caCert->getSubjectPublicKey())) {
      cout << "signature verifies ok"  << endl;
      exit(!ret);
    } else {
      cout << "signature is invalid"  << endl;
      exit(1);
    }
  } catch (Exception& ex){
    cout << "Exception:" << ex.getMessage() << endl;
    exit(2);
  } catch (Error& err) {
    cerr << "Error:Report to Akenti team:" << err.getMessage() << endl;
    exit(3);
  }
}

}
