#include "SSLSocket.hpp"
#include "Provider.hpp"
#include "OpenSSLProviderImpl.hpp"
#include "Socket.hpp"
#include "InputStream.hpp"
#include "OutputStream.hpp"
#include "RequestResponseGenerator.hpp"
#include "AkentiServiceRequest.hpp"
#include "AkentiServiceResponse.hpp"
#include "ProtocolConstants.hpp"
#include "CertificateFactory.hpp"
#include "Certificate.hpp"
#include "CertificateChain.hpp"
#include "AkentiCertificate.hpp"
#include "CapabilityCertificate.hpp"
#include "ConditionalActions.hpp"
#include "PublicKey.hpp"
#include "Principal.hpp"
#include "URL.hpp"
#include "File.hpp"
#include "URLConnection.hpp"
#include <iostream>
#include <sstream>

using std::cout;
using std::string;
using std::vector;
using std::pair;
using namespace Akenti;

static void interpretResponse(const string& command,
			      const Certificate& peerCert,
			      const AkentiServiceResponse& response);
/**
   host    - Server to be contacted [localhost] \n",
   port	   - Port at which the Server listens [8443] \n",
   cert	   - Subject Certificate [./client.cert] (optional) \n",
   key	   - Subject Key [./client.key] (optional) \n",
   cafile  - CA Certificate that client accepts [./ca.pem] (optional) \n",
   capath  - path of CA Certificates that client accepts [./] (optional) \n",
   resource - Resource to be accessed\n",
*/

extern "C"
int CheckAccess(char *inhost, int port, char *inclientkey, char *inclientcert,
		char *incacert, char *incapath, char *inresource) {

  string host(inhost);
  string cacert(incacert);
  string capath(incapath);
  string clientkey(inclientkey);
  string clientcert(inclientcert);
  string command("CheckAccess");
  vector< pair<string,string> > attrvalpairs;

  SSLSocket s;

  try {
    Provider po = Provider(new OpenSSLProviderImpl);
    Provider::setProvider(po);

    attrvalpairs.push_back(pair<string,string>("Resource",inresource));

    SSLContext clientContext = po->createSSLContext(SSLConstants::TLS1, true);

    clientContext->setVerify(1);

    clientContext->setCertificateFile(
      clientcert,
      SSLConstants::PEM
    );

    clientContext->setPrivateKeyFile(
      clientkey,
      SSLConstants::PEM
    );

    clientContext->setCAFile(cacert);
    clientContext->setCAPath(capath);

    po->setDefaultSSLClientContext(clientContext);

    const CertificateFactory& fac = CertificateFactory::getCertificateFactory();

    CertificateChain clientx509 = 
      fac->initCertificateChainFromFile(
        "PEM", 
        clientcert
     );

    Principal subjectdn = clientx509->getPrincipal();
    Principal cadn = clientx509->getGuarantor();

    attrvalpairs.push_back(pair<string,string>("SubjectDN", subjectdn.getName()));
    attrvalpairs.push_back(pair<string,string>("CaDN", cadn.getName()));

    InputStream is;
    OutputStream os;

    s = SSLSocket(host, port);
    s.connect();

    is = s.getInputStream();
    os = s.getOutputStream();

    cout << is.get() << endl;
    cout << os.get() << endl;

    if (!(is) || !(os)) {
      return 1;
    }

    RequestResponseGenerator reqResponseGen;

    AkentiServiceRequest request = 
      reqResponseGen.generateRequest(
        ProtocolConstants::SATP1_0, 
	command, 
        attrvalpairs);

    os << request;

    AkentiServiceResponse response;

    cout << "READING RESPONSE" << endl;
    is >> response;

    interpretResponse(request->getCommand(), s.getPeerCertificate(), response);

 } catch (exception& ex) {
   cout << "Exception Occurred :" << ex.what() << endl;
   return 1;
 }
 return 0;
}

void 
interpretResponse(const string& command,
		  const Certificate& peerCert,
		  const AkentiServiceResponse& response) {

  cout << "RESPONSE OBTAINED FROM THE SERVER" << endl;
  cout << response->paramString() << endl;
  cout << endl;
  cout << endl;
  cout << endl;
  cout << "RESPONSE IN DETAIL" << endl;
  cout << "__________________" << endl;

  if (command == "CheckAccess") {

    string certstr = response->getAttribute("Capability Certificate");

    if (certstr.length() == 0) {
      cout << "No Capability certificate obtained" << endl;
      return;
    }
   
    istringstream instream(certstr);

    AkentiCertificate cert;

    instream >> cert;

    CapabilityCertificate capcert(cert);

    cout << "VERSION : " << capcert->getVersionNumber() << endl;
    cout << "ID : " << capcert->getUniqueID() << endl;
    cout << "ISSUER : " << capcert->getIssuer().paramString() << endl;
    cout << "VALID NOT BEFORE : " << capcert->getNotBefore().paramString() << endl;
    cout << "VALID NOT AFTER : " << capcert->getNotAfter().paramString() << endl;

    cout << "RESOURCE :" << capcert->getResource() << endl;
    cout << "UNCONDITIONAL ACTIONS :" << endl;

    vector<string> uncondActions = capcert->getUnConditionalActions();

    if (uncondActions.size() > 0) {
      cout << uncondActions[0];
    }

    for (int i=1; i < uncondActions.size(); i++) {
      cout << ",";
      cout << uncondActions[i];
    }

    cout << endl;

    cout << "CONDITIONAL ACTIONS :" << endl;
    
    vector<ConditionalActions> condActions = capcert->getConditionalActions();

    if (condActions.size() > 0) {
      cout << condActions[0].paramString();
    }

    for (int i=1; i < condActions.size(); i++) {
      cout << ",";
      cout << condActions[i].paramString();
    }

    cout << endl;

    cout << "VERIFIED : ";

    if (capcert->verify(peerCert->getSubjectPublicKey())) {
      cout << "yes" << endl;
    } else {
      cout << "no" << endl;
    }
  }
}
