#ifndef AK_CERT_HOLDER_HPP
#define AK_CERT_HOLDER_HPP

#include <string> 
#include <vector> 
#include "AkentiCertificate.hpp"
#include "UtcTime.hpp"

using std::string;
using std::vector;

namespace Akenti {

class AKCertHolderImpl;
class CertificateStatusListener;

/**
  * Class <code>AKCertHolder</code> serves as the wrapper class
  * for <code>AKCertHolder</code> objects. It has a
  * <i>reference-count</i> for the implementation object that it holds.
  *
  * @type AKCertHolder
  *
  * @see ref_ptr
  * @see AKCertHolderImpl 
  */ 
typedef ref_ptr<AKCertHolderImpl> AKCertHolder;

/**
  * The <i>AKCertHolder</i> holds an <i>AkentiCertficate</i>, the level of the 
  * trust verification, and caching start/end times.
  *
  * Components can register their interests in changes to the info held in 
  * <i>AKCertHolder</i>.
  *
  * @see ref_ptr
  * @see AkentiCertificate;
  * @see CertificateStatusListener;
  * @see UtcTime
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

class AKCertHolderImpl {

public:

  /**
    */
  AKCertHolderImpl(const AkentiCertificate& cert = AkentiCertificate(), 
                  int level = 0,
                  const UtcTime& startCacheTime = UtcTime(),
                  const UtcTime& endCacheTime = UtcTime());
  /**
    */ 
  void addCertificateStatusListener(CertificateStatusListener* l);

  /**
    */
  void removeCertificateStatusListener(CertificateStatusListener* l);

  /**
    */
  const AkentiCertificate& getCertificate() const;

  /**
    */
  bool setLevel(int level);
  
  /**
    */
  int getLevel() const;

  /**
    */
  bool setStartCacheTime(const UtcTime& startCacheTime);

  /**
    */
  const UtcTime& getStartCacheTime() const;

  /**
    */
  bool setEndCacheTime(const UtcTime& endCacheTime);

  /**
    */
  const UtcTime& getEndCacheTime() const;

  /**
    */
  bool checkValidity(const UtcTime& utc);

  /**
    */
  string paramString() const;

private:

  AKCertHolderImpl(const AKCertHolderImpl&);

  void operator=(const AKCertHolderImpl&);

  AkentiCertificate cert;      // cached certificate 

  int level;                   // -1 = did not verify, 0 = not yet, 1+ = verified
  
  UtcTime startCacheTime;      // when the cert was cached

  UtcTime endCacheTime;        // cache time expires 

  vector<CertificateStatusListener*> statusListeners;

};

}; // end of name space

#endif
