#ifndef AKENTI_CERTIFICATE_HPP
#define AKENTI_CERTIFICATE_HPP

#include <iosfwd> 
#include "ref_ptr.hpp"
#include "AkentiCertificateImpl.hpp"

namespace Akenti {

class AkentiInputStream;
class AkentiOutputStream;

/**
  * AkentiCertificate is a simple reference-count wrapper aroubd 
  * <i>AkentiCertificateImpl<i>.
  *
  * @see AkentiCertificateImpl
  * @see ref_ptr 
  *
  * @version 1.1 00/05/02
  * @author Abdelilah Essiari
  * @author Srilekha Mudumbai
  **/
class AkentiCertificate : public ref_ptr<AkentiCertificateImpl> {

public:

  /**
    * Contructs this object.
    */
  AkentiCertificate();

  /**
   * Constructs a <i>AkentiCertificate</i> object given 
   * a <i>AkentiCertificateImpl</i> object. 
   *
   * @param impl wrapper <i>AkentiCertificateImpl</i> object
   * @see ref_ptr
   * @see AkentiCertificateImpl
   **/
  AkentiCertificate(AkentiCertificateImpl* impl);

  /**
    * Destructor.
    */
  virtual ~AkentiCertificate();

  /**
    * Compares two <i>AkentiCertificate</i> for equality.
    */
  bool operator==(const AkentiCertificate& cert) const;
  
  /**
    * Writes the certificate to an output stream.
    *
    * Note: only the PEM encoding is supported.
    */
  void writeEncoding(const string& encType, ostream& os) const;

  /**
    * Returns a string containing a short Identification of the certificate.
    * Used for debugging.
    */
  string toLogString() const;

  /**
    * Returns a string containing the contents of the certificate.
    * Used for debugging.
    */
  string paramString() const;

  /**
    * Writes an <i>AkentiCertificate</i>to an outputstream.
    *
    * Note: only the PEM encoding is supported.
    */
  static void 
  writeToStream(const string& encType,
                ostream& os, 
                const AkentiCertificate& cert);
  /**
    * Initialize an <i>AkentiCertificate</i> from an input stream. 
    *
    * Note: only the PEM encoding is supported.
    */
  static AkentiCertificate 
  initFromStream(const string& enc, istream& is);

  /**
    * Writes an <i>AkentiCertificate</i>to a file.
    *
    * Note: Currently, only the PEM encoding is supported.
    **/
  static void 
  writeToFile(const string& encType,
              const string& filename, 
              const AkentiCertificate& cert);
  /**
    * Initialize an <i>AkentiCertificate</i> from a file. 
    *
    * @param encType PEM or XML
    * @param filename name/path of the file 
    */ 
  static AkentiCertificate 
  initFromFile(const string& encType, 
               const string& filename);

  /**
    * Initialize an <i>AkentiCertificate</i> from a file. 
    *
    * Note this function is less efficient since it has
    *      to dtermine the encoding type.
    *
    * @param filename name/path of the file 
    * @see initFromFile(const string&, const string&)
    */
  static AkentiCertificate 
  initFromFile(const string& filename);

  /**
    */
  friend AkentiOutputStream& 
  operator<<(AkentiOutputStream& os, const AkentiCertificate& cert);

  /**
    */
  friend AkentiInputStream& 
  operator>>(AkentiInputStream& is, AkentiCertificate& cert);

  template<class T> bool instanceOf() {
     return dynamic_cast<typename T::ImplType> (get());
  }

private:

  typedef AkentiCertificateImpl* ImplType;

  /**
    * Writes the canonical form of a certificate to an <i>AkentiOutputStream</i>.
    */
  void writeObject(AkentiOutputStream& aos) const;

  /**
    * Reads a certificate from an input stream.
    *
    * Note: only the PEM encoding is supported.
    */
  static AkentiCertificate readEncoding(const string& encType, istream& is);

  /**
    * Initializes from the canonical form.
    */
  static AkentiCertificate readObject(AkentiInputStream& ais);
};

/**
  * Writes the PEM encoding of a certificate to the output stream.
  */
ostream& operator<<(ostream& os, const Akenti::AkentiCertificate& cert);

/**
  * Initializes the PEM encoding of a certificate to the output stream.
  */
istream& operator>>(istream& is, Akenti::AkentiCertificate& cert);

} // end of name space

#endif
