#include <sstream>
#include "Condition.hpp"
#include "AkentiOutputStream.hpp"
#include "AkentiInputStream.hpp"
#include "Expression.hpp"
#include "ExpressionVisitor.hpp"
#include "Exception.hpp"
#include "Error.hpp"
#include "XMLUtil.hpp"
#include "Xerces-Includes.hpp"

namespace Akenti {

typedef Condition COND;

COND::Condition() 
  : constraint("()"), attributeInfos() { 
}

COND::Condition(const string& constraint, 
                const vector<AttributeInfo>& attributeInfos) 
  : constraint(constraint), attributeInfos(attributeInfos)  {
  
}

COND::~Condition() {
}

void 
COND::operator=(const Condition& condition) {
  constraint = condition.constraint;
  attributeInfos = condition.attributeInfos;
}

const string& 
COND::getConstraint() const {
  return constraint;
}

bool 
COND::parseConstraint(Expression& expr) const {
  try {
    expr = Expression(constraint);
    return true;
  } catch(Exception& ex) {
    return false;
  }
}

const vector<AttributeInfo>&
COND::getAttributeInfos() const {
  return attributeInfos;
}

  class Visitor : public ExpressionVisitor {
  
  public:

    Visitor(const vector<AttributeInfo>& attributeInfos) 
      : attributeInfos(attributeInfos) {
    }
    
    void visit(TokenType t, const string& a, const string& v) {
      for (int i = 0; i < attributeInfos.size(); ++i) {
        const AttributeInfo& attr = attributeInfos[i];

        if (attr.hasAttribute(a) && attr.hasValue(v)) {
          return;
        }
      }
      
      throw Exception("Condition:warning:missing attribute:" + a + "," + v);
    }

    bool validateExpression(const Expression& expr) {
      try {
        expr.visit(*this);
        return true;
      } catch(Exception& ex) {
        return false;
      }
    }

  private:

    const vector<AttributeInfo>& attributeInfos;
  };
    
bool
COND::isValid() const {
  Expression expr;

  if (parseConstraint(expr)) {
    Visitor visitor(attributeInfos);

    return visitor.validateExpression(expr);
  }
  
  return false;
}

void 
COND::writeObject(AkentiOutputStream& drain) const {
  if (!constraint.length()) {
    drain  << "()";
  } else {
    drain  << constraint;
  }


  int size = attributeInfos.size();

  drain << ' ' << size;

  for (int i = 0; i < size; ++i) {
    drain << ' ' << attributeInfos[i];
  }
}

void 
COND::readObject(AkentiInputStream& source) {
  attributeInfos.clear();
  source >> constraint;

  int size;

  source >> size;

  if (size < 0) {
    throw Exception(
      "Condition:readObject:reading attributeInfos (size <= 0)"
    );
  }

  for (int i = 0; i < size; ++i) {
    AttributeInfo attribute;

    source >> attribute;
    attributeInfos.push_back(attribute);
  }
}

void 
COND::toXML(DOMElement* cond) const {
  DOMDocument* doc = cond->getOwnerDocument();
  DOMElement* temp = doc->createElement(XMLUtil::convert("Constraint"));

  temp->appendChild(doc->createTextNode(XMLUtil::convert(constraint.c_str())));
  cond->appendChild(temp);

  for (int i = 0; i < attributeInfos.size(); ++i) {
    temp = doc->createElement(XMLUtil::convert("AttributeInfo"));
    attributeInfos[i].toXML(temp);
    cond->appendChild(temp);
  }
}

void 
COND::fromXML(const DOMElement* cond) {
  attributeInfos.clear();

  DOMNode* child = cond->getFirstChild();

  if (child == NULL) {
    throw Exception("missing constraint");
  }

  DOMNode* temp = child->getFirstChild(); 

  if (temp == NULL) {
    throw Exception("missing constraint");
  }

  constraint = XMLUtil::convert(temp->getNodeValue());

  DOMNodeList* attrInfoNodes = cond->getElementsByTagName(XMLUtil::convert("AttributeInfo"));
  int size = attrInfoNodes->getLength();
  
  for (int i = 0; i < size; ++i) {
    temp = attrInfoNodes->item(i);
    
    DOMElement* element = (DOMElement *) temp;
    AttributeInfo info;

    info.fromXML(element);
    attributeInfos.push_back(info);
  }
}

string 
COND::paramString() const {
  std::ostringstream os;

  os << "constraint"
     << '=' 
     << '\"'
     << (constraint.length() == 0 ? "()" : constraint)
     << '\"';
 
  for (int i = 0; i < attributeInfos.size(); ++i) {
    os << ',' << attributeInfos[i].paramString();
  }

  return os.str();
}

}
