#ifndef CONDITIONAL_ACTIONS_HPP 
#define CONDITIONAL_ACTIONS_HPP 

#include <string>
#include <vector>

#include "AkentiObject.hpp"
#include "Condition.hpp"

using std::string; 
using std::vector; 

namespace Akenti {

/**
  * Holds a boolean expression and a vector of <i>AttributeInfo</i>
  * objects that describe the attribute/value pairs contained in 
  * the expression and a list of actions that will be granted if
  * the expression evaluates to true. Furthermore, we provide another
  * binary value <i>critical</i>, that when set to true, implies
  * that the boolean expression must evaluate to true otherwise
  * access must be denied altogether Such conditions come from
  *  UseConditions that have the critical flag set
  * and must be satisified.
  *
  * <pre>
  * ConditionalActions {
  *    Condition condition;
  *    vector&lt;string&gt; actions;
  *    bool critical;                 
  * }
  * </pre>
  *
  * @see AttributeInfo
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  **/

class ConditionalActions : public AkentiObject {

public:

  /** 
    * Default constructor.
    */
  ConditionalActions();
  
  /**
    */   
  ConditionalActions(const Condition& condition, 
                     const vector<string>& actions,
                     bool critical = false);
  /**
    */
  ~ConditionalActions();

  /**
    */
  void operator=(const ConditionalActions conditionalActions);

  /**
    */
  const Condition& getCondition() const;

  /**
    *
    */
  bool parseConstraint(Expression& expr) const;

  /**
    *
    */
  const string& getConstraint() const;

  /**
    */
  const vector<AttributeInfo>& getAttributeInfos() const;

  /**
    */
  const vector<string>& getActions() const;

  /**
    */
  bool isCritical() const;

  /**
    */
  bool isValid() const;
  
  /**
    */
  void writeObject(AkentiOutputStream& drain) const;

  /**
    */
  void readObject(AkentiInputStream& source);

  /**
    * Populates the XML element <i>condActions</i> from this <i>ConditionalActions</i>.
    */
  void toXML(DOMElement* condActions) const;

  /**
    * Initializes this <i>ConditionalActions</i> from the XML element
    * <i>condActions</i>
    */
  void fromXML(const DOMElement* condActions);

  /**
    * Used for debugging.
    *
    * @return a string representation of this object
    *
    * @see AkentiObject
    */
  string paramString() const;

  /**
    * Used for logging.
    */
  string toLogString() const;

private:

  Condition condition;

  vector<string> actions;
  
  bool critical;

};

} // end of name space

#endif
