#include <sstream>
#include "UseCondIssuerGroup.hpp"
#include "AkentiInputStream.hpp"
#include "AkentiOutputStream.hpp"
#include "Exception.hpp"
#include "Error.hpp"
#include "Util.hpp"
#include "Xerces-Includes.hpp"
#include "XMLUtil.hpp"

namespace Akenti {

typedef UseCondIssuerGroup UCIG;

UCIG::UseCondIssuerGroup(const vector<AkentiPrincipal>& issuers, 
                         const vector<Directory>& ucDirs) 
  : issuers(issuers), ucDirs(ucDirs) {
  if (!issuers.size()) {
    throw Error("UseCondIssuerGroup:constructor:illegal argument(1)");
  } else if (!ucDirs.size()) {
    throw Error("UseCondIssuerGroup:constructor:illegal argument(2)");
  }
}

UCIG::~UseCondIssuerGroup() {
}

const vector<AkentiPrincipal>&
UCIG::getUCIssuers() const {
  return issuers;
}

const vector<Directory>&
UCIG::getDirectories() const {
  return ucDirs;
}

bool 
UCIG::containsUCIssuer(const AkentiPrincipal& ap) const {
  bool ret = Util::compare(issuers[0].getName(), ap.getName())
             && 
             Util::compare(issuers[0].getGuarantor(), ap.getGuarantor());

  for (int i = 1; !ret && i < issuers.size(); ++i) {
    ret = Util::compare(issuers[i].getName(), ap.getName())
          && 
          Util::compare(issuers[i].getGuarantor(), ap.getGuarantor());
    ;
  }
  
  return ret;
}

void 
UCIG::writeObject(AkentiOutputStream& drain) const {
  int size = issuers.size();

  drain << size;

  for (int i = 0; i < size; ++i) {
    drain << ' ' << issuers[i];
  }

  size = ucDirs.size();

  drain << ' ' << size; 
  
  for (int i = 0; i < size; ++i) {
    drain << ' ' << ucDirs[i];
  }
}

void
UCIG::readObject(AkentiInputStream& source) {
  issuers.clear();
  ucDirs.clear();

  int size;

  source >> size;

  if (size <= 0) {
    throw Exception( "UseCondIssuerGroup:readObject: number of stakeholders <= 0");
  }

  for (int i = 0; i < size; ++i) {
    AkentiPrincipal ap;

    source >> ap;
    issuers.push_back(ap);
  }

  source >> size;

  if (size <= 0) {
    throw Exception( "UseCondIssuerGroup:readObject:number of ucDirs <= 0"); 
  }
  
  for (int i = 0; i < size; ++i) {
    Directory dir;

    source >> dir;
    ucDirs.push_back(dir);
  }
}

void 
UCIG::toXML(DOMElement* ucGroup) const {
  DOMDocument *doc = ucGroup->getOwnerDocument();

  for (int i= 0; i < issuers.size(); i++) {
    DOMElement *temp = doc->createElement(XMLUtil::convert("Principal"));
    
    issuers[i].toXML(temp);
    ucGroup->appendChild(temp);
  }
  
  for (int i=0; i < ucDirs.size(); i++) {
    DOMElement *temp = doc->createElement(XMLUtil::convert("URL"));

    ucDirs[i].toXML(temp);
    ucGroup->appendChild(temp);
  }
}

void 
UCIG::fromXML(const DOMElement* ucGroup) {
  issuers.clear();
  ucDirs.clear();

  DOMNodeList *issuerList = ucGroup->getElementsByTagName(XMLUtil::convert("Principal"));
  int size = issuerList->getLength();

  if (!size) {
    throw Exception("UseCondIssuerGroup:fromXML:number of stakeholders = 0");
  }
  
  for (int i = 0; i < size; ++i) {
    DOMNode *temp = issuerList->item(i);
 
    AkentiPrincipal ap;

    ap.fromXML((DOMElement*) temp);
    issuers.push_back(ap);
  }
   
  DOMNodeList *ucDirList = ucGroup->getElementsByTagName(XMLUtil::convert("URL"));
  size = ucDirList->getLength();
  
  if (!size) {
    throw Exception("UseCondIssuerGroup:fromXML:number of ucDirs = 0");
  }

  for (int i = 0; i < size; ++i) {
    DOMNode *temp = ucDirList->item(i);
    Directory dir;

    dir.fromXML((DOMElement*) temp);
    ucDirs.push_back(dir);
  }
}

string
UCIG::paramString() const {
  vector<AkentiPrincipal> issuers = getUCIssuers();
  int size = issuers.size();
  std::ostringstream os;

  os << '[' << size;
  
  for (int i = 0; i < size; ++i) {
    os << ',' << issuers[i].paramString();
  }

  size = ucDirs.size();
  os << ',' << size;

  for (int i = 0; i < size; ++i) {
    os << ',' << ucDirs[i].paramString();
  }

  os << ']';
  return os.str();
}

}
