#ifndef AK_IO_HANDLER_HPP
#define AK_IO_HANDLER_HPP

#include "ref_ptr.hpp"

namespace Akenti {

class AbstractIOHandlerImpl;

/**
  * Class <code>IOHandler</code> serves as the wrapper class
  * for <code>AbstractIOHandlerImpl</code> objects. It has a
  * <i>reference-count</i> for the implementation object that it holds.
  * 
  * @type IOHandler
  *
  * @see ref_ptr
  * @see AbstractIOHandlerImpl
  */
typedef ref_ptr<AbstractIOHandlerImpl> IOHandler;

/**
  * This is an abstract class <i>IOHandler</i> that defines the 
  * methods for a descriptor such as determining if the input/output
  * descriptor is open or closed and  read and write methods for the 
  * descriptor.   
  *
  * @version 1.1 00/11/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */
class AbstractIOHandlerImpl {

public:

  /**
    * Destroys this object. 
    */
  virtual ~AbstractIOHandlerImpl() {
  }

  /**
    * Gets the descriptor.
    *
    * @return the descriptor
    */
  virtual int getDescriptor() const = 0;

  /**
    * Determines if the descriptor is open.
    * 
    * @return true if the descriptor is open, otherwise returns false
    */
  virtual bool isOpen() const = 0;

  /**
    * Determines if the descriptor is closed. 
    * 
    * @return an true if the descriptor is closed, otherwise returns false
    */
  virtual bool isClosed() const = 0;

  /**
    * Read from the descriptor, a specified number of bytes into buffer. 
    * 
    * @param  buf the buffer to read into
    * @param  num specified number of bytes   
    * @return the actual number of bytes being read
    */
  virtual int read(char* buf, int num) = 0;

  /**
    * Write to the descriptor, a specified number of bytes from the buffer.
    *
    * @param buf the buffer to write from
    * @param num specified number of bytes 
    * @return the actual bytes being written 
    */
  virtual int write(const char* buf, int num) = 0;

  /**
    * Closes the descriptor.
    *
    * @return number after being closed  
    */
  virtual int close() = 0;

  /**
    * Returns a string representation for IOHandler which is used for
    * debugging purposes.
    *
    * @return debugging info
    */
  virtual string paramString() const = 0; 

};

 /**
   * This class  wraps a file descriptor and provides methods to get information
   * about the descriptor and to read and write from the descriptor.
   */


class IOHandlerImpl : public AbstractIOHandlerImpl {

public:

  /**
    * Constructs this object.
    *
    * @param fd the file descriptor
    */
  IOHandlerImpl(int fd);

  /**
    * Destroys this object.
    */
  ~IOHandlerImpl();

  /**
    * Gets the descriptor. 
    *
    * @return the descriptor 
    */
  int getDescriptor() const;

  /**
    * Determines if the descriptor is open
    *
    * @return an indication if descriptor is open or not
    */
  bool isOpen() const;

  /**
    * To check if the file descriptor is closed. 
    *
    * @return indicates that the file descriptor is closed 
    */
  bool isClosed() const ;

  /**
    * To read from the descriptor, a specified number of bytes into buffer.
    *
    * @param  buf the buffer to read into
    * @param  num specified number of bytes
    * @return the actual bytes being read
    */
  int read(char* buf, int num);

  /**
    * To write to the descriptor, a specified number of bytes from the buffer.
    *
    * @param buf the buffer to write from
    * @param num specified number of bytes to write
    * @return the actual bytes being written
    */
  int write(const char* buf, int num);


  /**
    * Closes the descriptor.
    *
    * @return status after being closed
    */
  int close();

  /**
    * Returns a string representation for IOHandler which is used  
    * debugging purposes.
    *
    * @return debugging info
    */
  string paramString() const;

protected:

  int fd;

};

} // end of namespace Akenti

#endif
