#ifndef OPEN_SSL_CERTIFICATE_GENERATOR_IMPL_HPP
#define OPEN_SSL_CERTIFICATE_GENERATOR_IMPL_HPP

#include "CertificateGeneratorImpl.hpp" 
#include "DistinguishedName.hpp"
#include "PublicKey.hpp"

namespace Akenti {

/**
  * This class is the subclass of <i>CertificateGeneratorImpl</i>,it
  * implements the methods used to initialize and generate 
  * a <i>Certificate</i> which  contains an OpenSSLX509CertficateImpl. 
  * Once a certificate has been generated and signed it can no longer be
  * modified.
  * The methods includes setting the subject, subject's
  * publicKey, issuer, validity, serial number, version number. It generates 
  * and signs the Certificate with a specified algorithm and privateKey. Also,
  * it will display the Certificate as a string. The certificate generation 
  * is done  by calls to the openssl libraries. 
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

class OpenSSLCertificateGeneratorImpl : public CertificateGeneratorImpl {

public:

  /**
    * Constructs this object.
    */
  OpenSSLCertificateGeneratorImpl();
 
  /**
    * Deletes this object.
    */
  ~OpenSSLCertificateGeneratorImpl();
  
  /**
    * Sets the subject of this certificate.
    *
    * @param subject DistinguishedName is the  subject of this certificate
    */
  void 
  setSubject(const DistinguishedName& subject);

  /**
    * Sets the public key of this certificate.
    * 
    * @param subjectKey PublicKey contains an OpenSSLPublicKeyImpl that contains
    * the public key of the subject 
    */
  void 
  setSubjectPublicKey(const PublicKey& subjectKey);

  /**
    * Sets the issuer of this certificate.
    * 
    * @param issuer DistinguishedName is the issuer of this certificate
    */
  void 
  setIssuer(const DistinguishedName& issuer);

  /**
    * Sets the number of days that the certificate will be valid
    * 
    * @param days int the number of days that the certificate will be valid
    */
  void 
  setValidity(int days);

  /**
    * Sets the serial number of this certificate.
    *
    * @param serialNumber long the serial number
    */
  void 
  setSerialNumber(long serialNumber);

  /**
    * Sets the version number of this certificate.
    *
    * @param versionNumber long the version number
    */
  void 
  setVersionNumber(long versionNumber);

  /**
    * Signs the certificate using the signature algorithm specified
    * by <i>algorithm</i> and the guarantor's private key specified
    * by <i>privateKey</i>. the privateKey  should reference an 
    * <i>OpenSSLPrivateKeyImpl</i>.
    *
    * @param algorithm string the specified algorithm
    * @param privateKey PrivateKey the private key of the guarantor 
    * @return the generated certificate
    * @exception SecurityException:  unrecognized public key
    * @exception SecurityException:  unrecognized private key
    * @exception SecurityException:  signature failed
    * @see generate
    */
  Certificate 
  generate(const string& algorithm, const PrivateKey& privateKey);
  
  /**
    * Returns string of items of this certificate for correctness check.
    */
  string 
  paramString() const;

private:

  DistinguishedName subject;

  PublicKey subjectKey;

  DistinguishedName issuer;

  int days;

  long serialNumber;

  long versionNumber;

};

} // end namespace Akenti

#endif
