#ifndef OPEN_SSL_PUBLIC_KEY_IMPL_HPP
#define OPEN_SSL_PUBLIC_KEY_IMPL_HPP

#include <openssl/evp.h>
#include "OpenSSLKeyImpl.hpp"
#include "PublicKeyImpl.hpp"

/**
  * This class is instantiated from two superclasses <i>OpenSSLKeyImpl</i>
  * and <i>PublicKeyImpl</i> and wraps the publicKey methods of 
  * <i>OpenSSLKeyImpl<i>. It inherits a EVP_PKEY structure from 
  * OpenSSLKeyImpl.
  *
  * @see OpenSSLKeyImpl
  * @see PublicKeyImpl
  * @see KeyPair
  * @see KeyPair#getPublicKey()
  * @see KeyFactory
  * @see KeyFactoryImpl#getKeyEncoded(const string&, const PublicKey&)
  * @see KeyFactoryImpl#initPublicKey(const string&, const string&, 
  *   const string&)
  * @see KeyFactoryImpl#generateRSAKeyPair(int, unsigned long)
  * @see KeyFactoryImpl#generateDSAKeyPair(int, const string&)
  * @see Signature
  * @see SignatureImpl#initVerify(const PublicKey&)
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

namespace Akenti {

class OpenSSLPublicKeyImpl : private OpenSSLKeyImpl, public PublicKeyImpl {

public:

  /**
    * Returns the standard algorithm name for this key. For
    * example, "DSA" would indicate that this key is a DSA key.
    * "RSA" would indicate that this key is a RSA key
    *
    * @return the name of the algorithm associated with this key.
    */
  string getAlgorithm() const;

  /**
    * Returns the standard format name such as "DER" 
    * to indicate using which format to encode the key
    */ 
  string getFormat() const;

  /**
    * Returns the size of the key
    */
  int getSize() const;

  /**
    * Encrypt the data using standard algorithm "RSA"
    *
    * @param string& data plain-text string to be  encrypted
    * @eturn the encrypted string
    */  
  string encrypt(const string& data) const;

  /**
    * Decrypts the encrypted string 
    *
    * @param string& encrypted cipher-text to be decrypted
    * @return the decrypted string
    */ 
  string decrypt(const string& encrypted) const;

  /**
    * Returns a string representation of this key for debugging
    * purposes.
    *
    * @return debugging info
    */
  string paramString() const;

  /** 
    * This should discourage clients not to mess around with 
    * implementation details.
    */
  EVP_PKEY *getEVP_PKEY();

  /** 
    * The constructor is used by OpenSSLKeyFactoryImpl.
    */
  OpenSSLPublicKeyImpl(EVP_PKEY* evp);

private:

};

} // end namespace Akenti

#endif
