#include "OpenSSLSignatureImpl.hpp"
#include "OpenSSLPublicKeyImpl.hpp"
#include "OpenSSLPrivateKeyImpl.hpp"
#include "Error.hpp"
#include "SecurityException.hpp"

typedef Akenti::OpenSSLSignatureImpl OSSI;

OSSI::OpenSSLSignatureImpl(const string& algo) : alg(algo){
  context.digest = getEVP_MD(algo);
}

OSSI::~OpenSSLSignatureImpl() {
}

string
OSSI::getAlgorithm() const {
  /* Oh well ... */
  return alg;
}

void
OSSI::resetEngine() {
  EVP_DigestInit(&context, context.digest);
}

void
OSSI::doUpdate(const string &data, int offset, int len) {
  EVP_DigestUpdate(&context, data.c_str() + offset, len);  
}

string 
OSSI::doSign() {
  OpenSSLPrivateKeyImpl *impl = dynamic_cast<OpenSSLPrivateKeyImpl *> (privateKey.get());

  if (!impl) {
    throw Error("OpenSSLSignatureImpl:initSign:bad key");
  }

  EVP_PKEY* evp = impl->getEVP_PKEY();
  unsigned int size = EVP_PKEY_size(evp);
  string signature(size, '\0');

  if (!EVP_SignFinal(&context, (unsigned char *) signature.data(), &size, evp)) {
    throw SecurityException("OpenSSLSignatureImpl:sign:failed");
  }

  return signature.substr(0, size);
}

bool 
OSSI::doVerify(const string &signature) {
  OpenSSLPublicKeyImpl *impl = dynamic_cast<OpenSSLPublicKeyImpl *> (publicKey.get());

  if (!impl) {
    throw Error("OpenSSLSignatureImpl:verify:unrecognized key");
  }

  EVP_PKEY* evp = impl->getEVP_PKEY();

  return  EVP_VerifyFinal( 
    &context, 
    (unsigned char *) signature.data(), 
    signature.length(), 
    evp 
  ) == 1;
}

bool 
OSSI::isSignatureAlgorithmSupported(const string& alg) {
  return OBJ_txt2nid((char *) alg.c_str()) != NID_undef;
}

const EVP_MD *
OSSI::getEVP_MD(const string& alg) {
  int nid = OBJ_txt2nid((char *) alg.c_str());

  if (nid == NID_undef) {
    throw SecurityException(
      "OpenSSLSignatureImpl:getEVP_MD:No Such Algorithm " + alg
    );
  }

  return EVP_get_digestbynid(nid);
}

string 
OSSI::paramString() const {
  string params;
  
  params += '[';
  params += "alg";
  params += '=';
  params += '"';
  params += getAlgorithm();
  params += '"';
  params += ']';
  return params;
}

