#ifndef OPEN_SSL_SIGNATURE_IMPL_HPP
#define OPEN_SSL_SIGNATURE_IMPL_HPP

#include <openssl/evp.h>
#include "SignatureImpl.hpp"

namespace Akenti {

/**
  * This class implements the digital signature interface defined in 
  * SignatureImpl by holding a ENV_MD_CTX object and wrapping the appropriate 
  * openssl functions.
  * 
  * Typically the clients will not deal with an instance of this class 
  * directly. Instead they interact with the <i>Signature</i> reference-count 
  * pointer.
  *
  * @see SignatureImpl
  * @see Error
  * @see Provider#getProvider()
  * @see ProviderImpl#getSignature(const string&)
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */


class OpenSSLSignatureImpl : public SignatureImpl {

public:

  /**
    *
    */
  OpenSSLSignatureImpl(const string& alg);

  /**
    *
    */
  ~OpenSSLSignatureImpl();

  /**
    * Returns a string that identifies the algorithm, independent of
    * implementation details. The name should be a standard
    * name such as "RSA-MD5" or "DSA-SHA1"
    *
    * @return name of the algorithm
    */
  string getAlgorithm() const;

  /**
    *
    */
  static bool isSignatureAlgorithmSupported(const string& alg);

  /**
    *
    */
  static const EVP_MD *getEVP_MD(const string& alg); 

protected:

  /**
    * Resets the digest engine.
    */
  void resetEngine();

  /**
    * Updates the data to be signed or verified.
    *
    * @param data the string of bytes to use for the update
    * @param offset start index
    * @param len the number of bytes to use starting at <i>offset</i>
    */
  void doUpdate(const string &data, int offset, int len);

  /**
    * Returns the digital signature of all the data that has been
    * accumlated from update calls.
    *
    * @exception Error if the private key is uninitialized
    * @exception SecurityException if signature fails
    */
  string doSign();

  /**
    * Returns true if signature verifies, false otherwise. All the data
    * that has been accumulated by update calls is signed by the public
    * key and the results are compared to the input signature. If they
    * match verify is true.
    *
    * @param signature digital signature to be verified
    * @exception Error if <i>initVerify</i> has not been called
    *                  or if the public key is uninitialized
    * @see doVerify(const string&)
    */
  bool doVerify(const string &signature);

  /**
    * Returns a string representation for OpenSSLSignatureImpl which is used for
    * debugging purposes.
    *
    * @return debugging info
    */
  string paramString() const;

private:

  EVP_MD_CTX context;

  /**
    * It would be nice to able to compute this.
    * There must be a way, but I am giving up ...
    * Srilekha, could u look into this?
    * @see getAlgorithm()
    */
  const string alg;

};

} // end namespace Akenti

#endif
