#ifndef AK_DISTINGUISHED_NAME_HPP
#define AK_DISTINGUISHED_NAME_HPP

#include <algorithm>
#include <vector>
#include <string>

using std::string;
using std::vector;
using std::pair;

namespace Akenti {

/**
  * This is a class to hold a Distinguished Name as a set of name and value
  * pairs. A Distinguished Name is a multi-component name used in X5.09
  * certifificate to uniquely identify an entity. Examples of attributes
  * are listed below.
  * <pre>
  * commonName                   "CN"
  * user ID                      "UID"
  * countryName                  "C"
  * localityName                 "L"
  * Email Address                "E"
  * stateOrProvinceName          "ST"
  * organizationName             "O"
  * organizationalUnitName       "OU"
  * </pre>
  * Some of the attributes may have more than one value.
  * And we allow user defined attributes
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  * @author Xiang Sun
  */

class DistinguishedName {

public:

  /**
    * Constructs this object from the input distinguishedName string.
    * The format of the string is /NAME1=value1/NAME2=value2...
    * 
    * @exception ParsingException if unable to parse  DN 
    */
  DistinguishedName(const string& dn = "/CN=NO_NAME");

  /**
    * Deletes this object.
    */
  ~DistinguishedName();

  /**
    * Checks to see if the pair of <attribute,value> exits. 
    *  
    * @param attribute name(s) or component
    * @param value the value of the name(s) or component 
    * @return ture if pair of <attribute,value> exits, false otherwise
    */
   bool containsAttributeValue(const string& attribute, const string& value) const;

  /**
    * Returns the <i>value</i> of the pair of <attribute,value>
    * specified by <i>attribute</i>.
    * 
    * @param attribute 
    * @exception Exception if the value can not be found for the specified 
    *   attribute
    */
  const string& getValue(const string& attribute) const;

  /**
    * Returns the vector with <i>values</i> corresponding to the 
    * <i>attribute</i>. If there are no  values for the attribute, the vector 
    * will not have any entries.
    * 
    * @param attribute
    * @return a vector of values corresponding to the specified attribute
    * @exception Exception no value found for the attribute
    */
  vector<string> getValues(const string& attribute) ;

  /**
    * Returns true if the specified <i> attribute</i> exists 
    * in this <i>DistinguishName</i>
    */  
  bool containsAttribute(const string& attribute) const;

  /**
    * Returns true if the specified <i>value</i> exists 
    * in the pair of <attribute,value>
    */
  bool containsValue(const string& value) const;

  /**
    * Returns the number of ocurrences of the specified <i>attribute</i>.
    * 
    * @param attribute 
    */ 
  int numOfOcurrences(const string& attribute) const;
  
  /**
    * Returns the size of the <i>vector</i> pairs
    *
    */ 
  int size() const;

  /**
    * Returns true if the specified index is in pairs' range
    *
    */
  bool hasMorePairs(int index) const;

  /**
    * Returns the pair of <attribute,value> by specified index.
    *
    */
  const pair<string, string>& getPairAt(int index) const;

  /**
    */
  string toString() const;

  /**
    * Returns a string representation for DistinguishedName which is used for
    * debugging purposes.
    *
    * @return debugging info
    */
  string paramString() const;

  /**
    * For logging purposes.
    */
  string toLogString() const;

  /**
    * Is true if the DN's are identical: same pairs, same values in the 
    *   same order.
    * 
    * @param dn the DN that is used for comparison with this DN 
    * Returns <i>true</i> if equal. The order is observed
    * but the attribute comparisons are not case sensitive.
    */
  bool operator==(const DistinguishedName& dn) const;

private:

  typedef pair<string, string> Pair;

  vector<Pair> pairs; 

};

} // end of namespace Akenti

#endif

